@@ -3,14 +3,14 @@
 export CFLAGS
 export LDFLAGS
 
-LTV="libtoolize (GNU libtool) 1.4.2"
-ACV="autoconf (GNU Autoconf) 2.53"
-AMV="automake (GNU automake) 1.6.3"
+LTV="libtoolize (GNU libtool) 1.4.3"
+ACV="autoconf (GNU Autoconf) 2.57"
+AMV="automake (GNU automake) 1.7.5"
 USAGE="
 This script documents the versions of the tools I'm using to build rpm:
-	libtool-1.4.2
-	autoconf-2.53
-	automake-1.6.2
+	libtool-1.4.3
+	autoconf-2.57
+	automake-1.7.5
 Simply edit this script to change the libtool/autoconf/automake versions
 checked if you need to, as rpm should build (and has built) with all
 recent versions of libtool/autoconf/automake.
@@ -20,10 +20,21 @@ recent versions of libtool/autoconf/automake.
 [ "`autoconf --version | head -1`" != "$ACV" ] && echo "$USAGE" && exit 1
 [ "`automake --version | head -1 | sed -e 's/1\.4[a-z]/1.4/'`" != "$AMV" ] && echo "$USAGE" && exit 1
 
-(echo "--- popt"; cd popt; ./autogen.sh --noconfigure "$@")
-(echo "--- zlib"; cd zlib; ./autogen.sh --noconfigure "$@")
-(echo "--- beecrypt"; cd beecrypt; ./autogen.sh --noconfigure "$@")
-(echo "--- libelf"; cd libelf; ./autogen.sh --noconfigure "$@")
+if [ -d popt ]; then
+    (echo "--- popt"; cd popt; ./autogen.sh --noconfigure "$@")
+fi
+if [ -d zlib ]; then
+    (echo "--- zlib"; cd zlib; ./autogen.sh --noconfigure "$@")
+fi
+if [ -d beecrypt ]; then
+    (echo "--- beecrypt"; cd beecrypt; ./autogen.sh --noconfigure "$@")
+fi
+if [ -d elfutils ]; then
+    (echo "--- elfutils"; cd elfutils; ./autogen.sh --noconfigure "$@")
+fi
+if [ -d file ]; then
+    (echo "--- file"; cd file; ./autogen.sh --noconfigure "$@")
+fi
 
 echo "--- rpm"
 libtoolize --copy --force
@@ -44,7 +55,12 @@ if [ X"$@" = X  -a "X`uname -s`" = "XLinux" ]; then
 	mandir=/usr/man
 	infodir=/usr/info
     fi
-    ./configure --prefix=/usr --sysconfdir=/etc --localstatedir=/var --infodir=${infodir} --mandir=${mandir} --without-javaglue "$@"
+    if [ -d /usr/lib/nptl ]; then
+	enable_posixmutexes="--enable-posixmutexes"
+    else
+	enable_posixmutexes=
+    fi
+    ./configure --prefix=/usr --sysconfdir=/etc --localstatedir=/var --infodir=${infodir} --mandir=${mandir} ${enable_posixmutexes} "$@"
 else
     ./configure "$@"
 fi
