@@ -1,14 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2001
+ * Copyright (c) 1996-2002
  *	Sleepycat Software.  All rights reserved.
  */
 
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "Id: bt_cursor.c,v 11.123 2001/09/17 16:59:29 bostic Exp ";
+static const char revid[] = "Id: bt_cursor.c,v 11.147 2002/08/13 20:46:07 ubell Exp ";
 #endif /* not lint */
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -18,12 +18,10 @@ static const char revid[] = "Id: bt_cursor.c,v 11.123 2001/09/17 16:59:29 bostic
 #endif
 
 #include "db_int.h"
-#include "db_page.h"
-#include "db_shash.h"
-#include "btree.h"
-#include "lock.h"
-#include "qam.h"
-#include "common_ext.h"
+#include "dbinc/db_page.h"
+#include "dbinc/db_shash.h"
+#include "dbinc/btree.h"
+#include "dbinc/lock.h"
 
 static int  __bam_bulk __P((DBC *, DBT *, u_int32_t));
 static int  __bam_c_close __P((DBC *, db_pgno_t, int *));
@@ -37,7 +35,6 @@ static int  __bam_c_next __P((DBC *, int, int));
 static int  __bam_c_physdel __P((DBC *));
 static int  __bam_c_prev __P((DBC *));
 static int  __bam_c_put __P((DBC *, DBT *, DBT *, u_int32_t, db_pgno_t *));
-static void __bam_c_reset __P((BTREE_CURSOR *));
 static int  __bam_c_search __P((DBC *,
 		db_pgno_t, const DBT *, u_int32_t, int *));
 static int  __bam_c_writelock __P((DBC *));
@@ -82,8 +79,7 @@ static int  __bam_isopd __P((DBC *, db_pgno_t *));
 		ret = 0;						\
 	if ((ret) == 0 && STD_LOCKING(dbc))				\
 		ret = __db_lget(dbc,					\
-		    LOCK_ISSET(lock) ? LCK_COUPLE : 0,			\
-		    lpgno, mode, 0, &(lock));				\
+		    LCK_COUPLE_ALWAYS, lpgno, mode, 0, &(lock));	\
 	if ((ret) == 0)							\
 		ret = __mpf->get(__mpf, &(fpgno), 0, &(pagep));		\
 }
@@ -161,12 +157,12 @@ static int  __bam_isopd __P((DBC *, db_pgno_t *));
 
 /* If on-page item is a deleted record. */
 #undef	IS_DELETED
-#define	IS_DELETED(page, indx)						\
-	B_DISSET(GET_BKEYDATA(page,					\
+#define	IS_DELETED(dbp, page, indx)					\
+	B_DISSET(GET_BKEYDATA(dbp, page,				\
 	    (indx) + (TYPE(page) == P_LBTREE ? O_INDX : 0))->type)
 #undef	IS_CUR_DELETED
 #define	IS_CUR_DELETED(dbc)						\
-	IS_DELETED((dbc)->internal->page, (dbc)->internal->indx)
+	IS_DELETED((dbc)->dbp, (dbc)->internal->page, (dbc)->internal->indx)
 
 /*
  * Test to see if two cursors could point to duplicates of the same key.
@@ -178,30 +174,14 @@ static int  __bam_isopd __P((DBC *, db_pgno_t *));
  */
 #undef	IS_DUPLICATE
 #define	IS_DUPLICATE(dbc, i1, i2)					\
-	    (((PAGE *)(dbc)->internal->page)->inp[i1] ==		\
-	     ((PAGE *)(dbc)->internal->page)->inp[i2])
+	    (P_INP((dbc)->dbp,((PAGE *)(dbc)->internal->page))[i1] ==	\
+	     P_INP((dbc)->dbp,((PAGE *)(dbc)->internal->page))[i2])
 #undef	IS_CUR_DUPLICATE
 #define	IS_CUR_DUPLICATE(dbc, orig_pgno, orig_indx)			\
 	(F_ISSET(dbc, DBC_OPD) ||					\
 	    (orig_pgno == (dbc)->internal->pgno &&			\
 	    IS_DUPLICATE(dbc, (dbc)->internal->indx, orig_indx)))
 
-/*
- * __bam_c_reset --
- *	Initialize internal cursor structure.
- */
-static void
-__bam_c_reset(cp)
-	BTREE_CURSOR *cp;
-{
-	cp->csp = cp->sp;
-	LOCK_INIT(cp->lock);
-	cp->lock_mode = DB_LOCK_NG;
-	cp->recno = RECNO_OOB;
-	cp->order = INVALID_ORDER;
-	cp->flags = 0;
-}
-
 /*
  * __bam_c_init --
  *	Initialize the access private portion of a cursor
@@ -213,26 +193,15 @@ __bam_c_init(dbc, dbtype)
 	DBC *dbc;
 	DBTYPE dbtype;
 {
-	BTREE *t;
-	BTREE_CURSOR *cp;
-	DB *dbp;
+	DB_ENV *dbenv;
 	int ret;
-	u_int32_t minkey;
 
-	dbp = dbc->dbp;
+	dbenv = dbc->dbp->dbenv;
 
 	/* Allocate/initialize the internal structure. */
-	if (dbc->internal == NULL) {
-		if ((ret = __os_malloc(dbp->dbenv,
-		    sizeof(BTREE_CURSOR), &cp)) != 0)
-			return (ret);
-		dbc->internal = (DBC_INTERNAL *)cp;
-
-		cp->sp = cp->csp = cp->stack;
-		cp->esp = cp->stack + sizeof(cp->stack) / sizeof(cp->stack[0]);
-	} else
-		cp = (BTREE_CURSOR *)dbc->internal;
-	__bam_c_reset(cp);
+	if (dbc->internal == NULL && (ret =
+	    __os_malloc(dbenv, sizeof(BTREE_CURSOR), &dbc->internal)) != 0)
+		return (ret);
 
 	/* Initialize methods. */
 	dbc->c_close = __db_c_close;
@@ -260,18 +229,6 @@ __bam_c_init(dbc, dbtype)
 		dbc->c_am_writelock = __bam_c_writelock;
 	}
 
-	/*
-	 * The btree leaf page data structures require that two key/data pairs
-	 * (or four items) fit on a page, but other than that there's no fixed
-	 * requirement.  The btree off-page duplicates only require two items,
-	 * to be exact, but requiring four for them as well seems reasonable.
-	 *
-	 * Recno uses the btree bt_ovflsize value -- it's close enough.
-	 */
-	t = dbp->bt_internal;
-	minkey = F_ISSET(dbc, DBC_OPD) ? 2 : t->bt_minkey;
-	cp->ovflsize = B_MINKEY_TO_OVFLSIZE(minkey, dbp->pgsize);
-
 	return (0);
 }
 
@@ -285,12 +242,13 @@ int
 __bam_c_refresh(dbc)
 	DBC *dbc;
 {
+	BTREE *t;
 	BTREE_CURSOR *cp;
 	DB *dbp;
 
 	dbp = dbc->dbp;
+	t = dbp->bt_internal;
 	cp = (BTREE_CURSOR *)dbc->internal;
-	__bam_c_reset(cp);
 
 	/*
 	 * If our caller set the root page number, it's because the root was
@@ -298,11 +256,32 @@ __bam_c_refresh(dbc)
 	 * pull it out of our internal information.
 	 */
 	if (cp->root == PGNO_INVALID)
-		cp->root = ((BTREE *)dbp->bt_internal)->bt_root;
+		cp->root = t->bt_root;
+
+	LOCK_INIT(cp->lock);
+	cp->lock_mode = DB_LOCK_NG;
+
+	cp->sp = cp->csp = cp->stack;
+	cp->esp = cp->stack + sizeof(cp->stack) / sizeof(cp->stack[0]);
+
+	/*
+	 * The btree leaf page data structures require that two key/data pairs
+	 * (or four items) fit on a page, but other than that there's no fixed
+	 * requirement.  The btree off-page duplicates only require two items,
+	 * to be exact, but requiring four for them as well seems reasonable.
+	 *
+	 * Recno uses the btree bt_ovflsize value -- it's close enough.
+	 */
+	cp->ovflsize = B_MINKEY_TO_OVFLSIZE(
+	    dbp,  F_ISSET(dbc, DBC_OPD) ? 2 : t->bt_minkey, dbp->pgsize);
+
+	cp->recno = RECNO_OOB;
+	cp->order = INVALID_ORDER;
+	cp->flags = 0;
 
 	/* Initialize for record numbers. */
 	if (F_ISSET(dbc, DBC_OPD) ||
-	    dbc->dbtype == DB_RECNO || F_ISSET(dbp, DB_BT_RECNUM)) {
+	    dbc->dbtype == DB_RECNO || F_ISSET(dbp, DB_AM_RECNUM)) {
 		F_SET(cp, C_RECNUM);
 
 		/*
@@ -311,7 +290,7 @@ __bam_c_refresh(dbc)
 		 * mutable record numbers.
 		 */
 		if ((F_ISSET(dbc, DBC_OPD) && dbc->dbtype == DB_RECNO) ||
-		    F_ISSET(dbp, DB_BT_RECNUM | DB_RE_RENUMBER))
+		    F_ISSET(dbp, DB_AM_RECNUM | DB_AM_RENUMBER))
 			F_SET(cp, C_RENUMBER);
 	}
 
@@ -429,7 +408,7 @@ __bam_c_close(dbc, root_pgno, rmroot)
 		 */
 		if ((ret = mpf->get(mpf, &cp->pgno, 0, &h)) != 0)
 			goto err;
-		root_pgno = GET_BOVERFLOW(h, cp->indx + O_INDX)->pgno;
+		root_pgno = GET_BOVERFLOW(dbp, h, cp->indx + O_INDX)->pgno;
 		if ((ret = mpf->put(mpf, h, 0)) != 0)
 			goto err;
 
@@ -472,8 +451,7 @@ lock:	cp_c = (BTREE_CURSOR *)dbc_c->internal;
 	 * info in __db_c_get--the OPD is also a WRITEDUP.
 	 */
 	if (CDB_LOCKING(dbp->dbenv)) {
-		DB_ASSERT(!F_ISSET(dbc, DBC_OPD) || F_ISSET(dbc, DBC_WRITEDUP));
-		if (!F_ISSET(dbc, DBC_WRITER)) {
+		if (F_ISSET(dbc, DBC_WRITEDUP | DBC_WRITECURSOR)) {
 			if ((ret = dbp->dbenv->lock_get(
 			    dbp->dbenv, dbc->locker, DB_LOCK_UPGRADE,
 			    &dbc->lock_dbt, DB_LOCK_WRITE, &dbc->mylock)) != 0)
@@ -617,7 +595,7 @@ __bam_c_destroy(dbc)
 	DBC *dbc;
 {
 	/* Discard the structures. */
-	__os_free(dbc->dbp->dbenv, dbc->internal, sizeof(BTREE_CURSOR));
+	__os_free(dbc->dbp->dbenv, dbc->internal);
 
 	return (0);
 }
@@ -729,9 +707,8 @@ __bam_c_del(dbc)
 	}
 
 	/* Log the change. */
-	if (DB_LOGGING(dbc)) {
-		if ((ret = __bam_cdel_log(dbp->dbenv,
-		    dbc->txn, &LSN(cp->page), 0, dbp->log_fileid,
+	if (DBC_LOGGING(dbc)) {
+		if ((ret = __bam_cdel_log(dbp, dbc->txn, &LSN(cp->page), 0,
 		    PGNO(cp->page), &LSN(cp->page), cp->indx)) != 0)
 			goto err;
 	} else
@@ -739,9 +716,9 @@ __bam_c_del(dbc)
 
 	/* Set the intent-to-delete flag on the page. */
 	if (TYPE(cp->page) == P_LBTREE)
-		B_DSET(GET_BKEYDATA(cp->page, cp->indx + O_INDX)->type);
+		B_DSET(GET_BKEYDATA(dbp, cp->page, cp->indx + O_INDX)->type);
 	else
-		B_DSET(GET_BKEYDATA(cp->page, cp->indx)->type);
+		B_DSET(GET_BKEYDATA(dbp, cp->page, cp->indx)->type);
 
 	/* Mark the page dirty. */
 	ret = mpf->set(mpf, cp->page, DB_MPOOL_DIRTY);
@@ -868,6 +845,7 @@ __bam_c_get(dbc, key, data, flags, pgnop)
 		if (F_ISSET(dbc, DBC_OPD)) {
 			if ((ret = __bam_c_search(
 			    dbc, PGNO_INVALID, data, flags, &exact)) != 0)
+				goto err;
 			if (flags == DB_GET_BOTH) {
 				if (!exact) {
 					ret = DB_NOTFOUND;
@@ -971,7 +949,7 @@ __bam_c_get(dbc, key, data, flags, pgnop)
 	case DB_SET_RANGE:
 		newopd = 1;
 		if ((ret = __bam_c_search(dbc,
-		     PGNO_INVALID, key, flags, &exact)) != 0)
+		    PGNO_INVALID, key, flags, &exact)) != 0)
 			goto err;
 
 		/*
@@ -996,8 +974,15 @@ __bam_c_get(dbc, key, data, flags, pgnop)
 	if (newopd && pgnop != NULL)
 		(void)__bam_isopd(dbc, pgnop);
 
-	/* Don't return the key, it was passed to us */
-	if (flags == DB_SET)
+	/*
+	 * Don't return the key, it was passed to us (this is true even if the
+	 * application defines a compare function returning equality for more
+	 * than one key value, since in that case which actual value we store
+	 * in the database is undefined -- and particularly true in the case of
+	 * duplicates where we only store one key value).
+	 */
+	if (flags == DB_GET_BOTH ||
+	    flags == DB_GET_BOTH_RANGE || flags == DB_SET)
 		F_SET(key, DB_DBT_ISSET);
 
 err:	/*
@@ -1005,8 +990,8 @@ err:	/*
 	 * moved, clear the delete flag, DBcursor->c_get never references
 	 * a deleted key, if it moved at all.
 	 */
-	if (F_ISSET(cp, C_DELETED)
-	    && (cp->pgno != orig_pgno || cp->indx != orig_indx))
+	if (F_ISSET(cp, C_DELETED) &&
+	    (cp->pgno != orig_pgno || cp->indx != orig_indx))
 		F_CLR(cp, C_DELETED);
 
 	return (ret);
@@ -1026,10 +1011,10 @@ __bam_get_prev(dbc)
 
 	if (__bam_isopd(dbc, &pgno)) {
 		cp = (BTREE_CURSOR *)dbc->internal;
-		if ((ret = __db_c_newopd(dbc, pgno, &cp->opd)) != 0)
+		if ((ret = __db_c_newopd(dbc, pgno, cp->opd, &cp->opd)) != 0)
 			return (ret);
 		if ((ret = cp->opd->c_am_get(cp->opd,
-		     &key, &data, DB_LAST, NULL)) != 0)
+		    &key, &data, DB_LAST, NULL)) != 0)
 			return (ret);
 	}
 
@@ -1049,7 +1034,7 @@ __bam_bulk(dbc, data, flags)
 	BOVERFLOW *bo;
 	BTREE_CURSOR *cp;
 	PAGE *pg;
-	db_indx_t indx, pg_keyoff;
+	db_indx_t *inp, indx, pg_keyoff;
 	int32_t  *endp, key_off, *offp, *saveoffp;
 	u_int8_t *dbuf, *dp, *np;
 	u_int32_t key_size, size, space;
@@ -1106,6 +1091,7 @@ next_pg:
 	indx = cp->indx;
 	pg = cp->page;
 
+	inp = P_INP(dbc->dbp, pg);
 	/* The current page is not yet in the buffer. */
 	need_pg = 1;
 
@@ -1116,10 +1102,10 @@ next_pg:
 	 */
 	pg_keyoff = 0;
 	if (is_key == 0)
-		pg_keyoff = pg->inp[indx];
+		pg_keyoff = inp[indx];
 
 	do {
-		if (IS_DELETED(pg, indx)) {
+		if (IS_DELETED(dbc->dbp, pg, indx)) {
 			if (dbc->dbtype != DB_RECNO)
 				continue;
 
@@ -1149,18 +1135,18 @@ next_pg:
 		 * key on the page.  If its already there
 		 * then we just point to it.
 		 */
-		if (is_key && pg_keyoff != pg->inp[indx]) {
-			bk = GET_BKEYDATA(pg, indx);
+		if (is_key && pg_keyoff != inp[indx]) {
+			bk = GET_BKEYDATA(dbc->dbp, pg, indx);
 			if (B_TYPE(bk->type) == B_OVERFLOW) {
 				bo = (BOVERFLOW *)bk;
 				size = key_size = bo->tlen;
 				if (key_size > space)
 					goto get_key_space;
 				if ((ret = __bam_bulk_overflow(dbc,
-				     bo->tlen, bo->pgno, np)) != 0)
+				    bo->tlen, bo->pgno, np)) != 0)
 					return (ret);
 				space -= key_size;
-				key_off = np - dbuf;
+				key_off = (int32_t)(np - dbuf);
 				np += key_size;
 			} else {
 				if (need_pg) {
@@ -1205,9 +1191,9 @@ get_key_space:
 					np += size;
 				}
 				key_size = bk->len;
-				key_off = pg->inp[indx] - HOFFSET(pg)
-				     + dp - dbuf + SSZA(BKEYDATA, data);
-				pg_keyoff = pg->inp[indx];
+				key_off = (int32_t)(inp[indx] - HOFFSET(pg)
+				    + dp - dbuf + SSZA(BKEYDATA, data));
+				pg_keyoff = inp[indx];
 			}
 		}
 
@@ -1229,7 +1215,7 @@ get_key_space:
 		 * If we have an off page dup, then copy as many
 		 * as will fit into the buffer.
 		 */
-		bk = GET_BKEYDATA(pg, indx + adj - 1);
+		bk = GET_BKEYDATA(dbc->dbp, pg, indx + adj - 1);
 		if (B_TYPE(bk->type) == B_DUPLICATE) {
 			bo = (BOVERFLOW *)bk;
 			if (is_key) {
@@ -1271,7 +1257,7 @@ get_key_space:
 				*offp-- = key_size;
 			} else if (rec_key)
 				*offp-- = cp->recno;
-			*offp-- = np - dbuf;
+			*offp-- = (int32_t)(np - dbuf);
 			np += size;
 			*offp-- = size;
 		} else {
@@ -1288,8 +1274,8 @@ back_up:
 						indx -= adj;
 					else {
 						if ((ret =
-						    __bam_get_prev(dbc)) != 0
-						    && ret != DB_NOTFOUND)
+						    __bam_get_prev(dbc)) != 0 &&
+						    ret != DB_NOTFOUND)
 							return (ret);
 						indx = cp->indx;
 						pg = cp->page;
@@ -1302,8 +1288,9 @@ get_space:
 					 * buffer or if we are doing a DBP->get
 					 * did we get all of the data.
 					 */
-					if (offp >= (is_key ? &endp[-1] : endp)
-					     || F_ISSET(dbc, DBC_TRANSIENT)) {
+					if (offp >=
+					    (is_key ? &endp[-1] : endp) ||
+					    F_ISSET(dbc, DBC_TRANSIENT)) {
 						data->size = ALIGN(size +
 						    data->ulen - space,
 						    sizeof(u_int32_t));
@@ -1325,23 +1312,23 @@ get_space:
 				*offp-- = key_size;
 			} else if (rec_key)
 				*offp-- = cp->recno;
-			*offp-- = pg->inp[indx + adj - 1] - HOFFSET(pg)
-			     + dp - dbuf + SSZA(BKEYDATA, data);
+			*offp-- = (int32_t)(inp[indx + adj - 1] - HOFFSET(pg)
+			    + dp - dbuf + SSZA(BKEYDATA, data));
 			*offp-- = bk->len;
 		}
 		if (dbc->dbtype == DB_RECNO)
 			cp->recno++;
 		else if (no_dup) {
-			while (indx + adj < NUM_ENT(pg)
-			    && pg_keyoff == pg->inp[indx + adj])
+			while (indx + adj < NUM_ENT(pg) &&
+			    pg_keyoff == inp[indx + adj])
 				indx += adj;
 		}
 	/*
 	 * Stop when we either run off the page or we
 	 * move to the next key and we are not returning mulitple keys.
 	 */
-	} while ((indx += adj) < NUM_ENT(pg)
-	     && (next_key || pg_keyoff == pg->inp[indx]));
+	} while ((indx += adj) < NUM_ENT(pg) &&
+	    (next_key || pg_keyoff == inp[indx]));
 
 	/* If we are off the page then try to the next page. */
 	if (ret == 0 && next_key && indx >= NUM_ENT(pg)) {
@@ -1361,7 +1348,7 @@ get_space:
 	 */
 
 	if (ret == 0 &&
-	    F_ISSET(dbc, DBC_TRANSIENT) && pg_keyoff == pg->inp[indx]) {
+	    F_ISSET(dbc, DBC_TRANSIENT) && pg_keyoff == inp[indx]) {
 		data->size = (data->ulen - space) + size;
 		return (ENOMEM);
 	}
@@ -1370,12 +1357,15 @@ get_space:
 	 * If we are not fetching keys, we may have stepped to the
 	 * next key.
 	 */
-	if (next_key || pg_keyoff == pg->inp[indx])
+	if (next_key || pg_keyoff == inp[indx])
 		cp->indx = indx;
 	else
 		cp->indx = indx - P_INDX;
 
-	*offp = (u_int32_t) -1;
+	if (rec_key == 1)
+		*offp = (u_int32_t) RECNO_OOB;
+	else
+		*offp = (u_int32_t) -1;
 	return (0);
 }
 
@@ -1421,13 +1411,14 @@ __bam_bulk_duplicates(dbc, pgno, dbuf, keyoff, offpp, dpp, spacep, no_dup)
 	u_int32_t *spacep;
 	int no_dup;
 {
+	DB *dbp;
 	BKEYDATA *bk;
 	BOVERFLOW *bo;
 	BTREE_CURSOR *cp;
 	DBC *opd;
 	DBT key, data;
 	PAGE *pg;
-	db_indx_t indx;
+	db_indx_t indx, *inp;
 	int32_t *offp;
 	u_int32_t size, space;
 	u_int8_t *dp, *np;
@@ -1435,11 +1426,12 @@ __bam_bulk_duplicates(dbc, pgno, dbuf, keyoff, offpp, dpp, spacep, no_dup)
 
 	ret = 0;
 
+	dbp = dbc->dbp;
 	cp = (BTREE_CURSOR *)dbc->internal;
 	opd = cp->opd;
 
 	if (opd == NULL) {
-		if ((ret = __db_c_newopd(dbc, pgno, &opd)) != 0)
+		if ((ret = __db_c_newopd(dbc, pgno, NULL, &opd)) != 0)
 			return (ret);
 		cp->opd = opd;
 		if ((ret = opd->c_am_get(opd,
@@ -1467,13 +1459,14 @@ __bam_bulk_duplicates(dbc, pgno, dbuf, keyoff, offpp, dpp, spacep, no_dup)
 			break;
 		pg = cp->page;
 		indx = cp->indx;
+		inp = P_INP(dbp, pg);
 		/* We need to copy the page to the buffer. */
 		need_pg = 1;
 
 		do {
-			if (IS_DELETED(pg, indx))
+			if (IS_DELETED(dbp, pg, indx))
 				goto contin;
-			bk = GET_BKEYDATA(pg, indx);
+			bk = GET_BKEYDATA(dbp, pg, indx);
 			space -= 2 * sizeof(*offp);
 			/* Allocate space for key if needed. */
 			if (first == 0 && keyoff != NULL)
@@ -1507,7 +1500,7 @@ __bam_bulk_duplicates(dbc, pgno, dbuf, keyoff, offpp, dpp, spacep, no_dup)
 				    bo->tlen, bo->pgno, np)) != 0)
 					return (ret);
 				space -= size;
-				*offp-- = np - dbuf;
+				*offp-- = (int32_t)(np - dbuf);
 				np += size;
 			} else {
 				if (need_pg) {
@@ -1532,8 +1525,8 @@ __bam_bulk_duplicates(dbc, pgno, dbuf, keyoff, offpp, dpp, spacep, no_dup)
 					*offp-- = keyoff[-1];
 				}
 				size = bk->len;
-				*offp-- = pg->inp[indx] - HOFFSET(pg)
-				     + dp - dbuf + SSZA(BKEYDATA, data);
+				*offp-- = (int32_t)(inp[indx] - HOFFSET(pg)
+				    + dp - dbuf + SSZA(BKEYDATA, data));
 			}
 			*offp-- = size;
 			first = 0;
@@ -1638,7 +1631,7 @@ __bam_getbothc(dbc, data)
 		cp->page = NULL;
 
 		return (__bam_c_search(dbc,
-		     PGNO_INVALID, data, DB_GET_BOTH, &exact));
+		    PGNO_INVALID, data, DB_GET_BOTH, &exact));
 	}
 
 	/*
@@ -1717,7 +1710,7 @@ __bam_getboth_finddatum(dbc, data, flags)
 	if (base == (top - P_INDX)) {
 		if  ((ret = __bam_cmp(dbp, data,
 		    cp->page, cp->indx + O_INDX, dbp->dup_compare, &cmp)) != 0)
-		       return (ret);
+			return (ret);
 		return (cmp == 0 ||
 		    (cmp < 0 && flags == DB_GET_BOTH_RANGE) ? 0 : DB_NOTFOUND);
 	}
@@ -1843,7 +1836,7 @@ split:	ret = stack = 0;
 
 			/* Disallow "sorted" duplicate duplicates. */
 			if (exact) {
-				if (IS_DELETED(cp->page, cp->indx)) {
+				if (IS_DELETED(dbp, cp->page, cp->indx)) {
 					iiop = DB_CURRENT;
 					break;
 				}
@@ -1919,8 +1912,8 @@ split:	ret = stack = 0;
 		 */
 		for (;; cp->indx += P_INDX) {
 			if ((ret = __bam_cmp(dbp, data, cp->page,
-			    cp->indx + O_INDX, dbp->dup_compare, &cmp)) !=0)
-				return (ret);
+			    cp->indx + O_INDX, dbp->dup_compare, &cmp)) != 0)
+				goto err;
 			if (cmp < 0) {
 				iiop = DB_BEFORE;
 				break;
@@ -1928,7 +1921,7 @@ split:	ret = stack = 0;
 
 			/* Disallow "sorted" duplicate duplicates. */
 			if (cmp == 0) {
-				if (IS_DELETED(cp->page, cp->indx)) {
+				if (IS_DELETED(dbp, cp->page, cp->indx)) {
 					iiop = DB_CURRENT;
 					break;
 				}
@@ -1937,8 +1930,8 @@ split:	ret = stack = 0;
 			}
 
 			if (cp->indx + P_INDX >= NUM_ENT(cp->page) ||
-			    ((PAGE *)cp->page)->inp[cp->indx] !=
-			    ((PAGE *)cp->page)->inp[cp->indx + P_INDX]) {
+			    P_INP(dbp, ((PAGE *)cp->page))[cp->indx] !=
+			    P_INP(dbp, ((PAGE *)cp->page))[cp->indx + P_INDX]) {
 				iiop = DB_AFTER;
 				break;
 			}
@@ -2055,7 +2048,7 @@ __bam_c_rget(dbc, data)
 	    1, &recno, &exact)) != 0)
 		goto err;
 
-	ret = __db_retcopy(dbp, data,
+	ret = __db_retcopy(dbp->dbenv, data,
 	    &recno, sizeof(recno), &dbc->rdata->data, &dbc->rdata->ulen);
 
 	/* Release the stack. */
@@ -2115,7 +2108,7 @@ __bam_c_first(dbc)
 		if (ISLEAF(cp->page))
 			break;
 
-		pgno = GET_BINTERNAL(cp->page, 0)->pgno;
+		pgno = GET_BINTERNAL(dbc->dbp, cp->page, 0)->pgno;
 	}
 
 	/* If we want a write lock instead of a read lock, get it now. */
@@ -2160,8 +2153,8 @@ __bam_c_last(dbc)
 		if (ISLEAF(cp->page))
 			break;
 
-		pgno =
-		    GET_BINTERNAL(cp->page, NUM_ENT(cp->page) - O_INDX)->pgno;
+		pgno = GET_BINTERNAL(dbc->dbp, cp->page,
+		    NUM_ENT(cp->page) - O_INDX)->pgno;
 	}
 
 	/* If we want a write lock instead of a read lock, get it now. */
@@ -2332,7 +2325,7 @@ __bam_c_search(dbc, root_pgno, key, flags, exactp)
 	BTREE_CURSOR *cp;
 	DB *dbp;
 	PAGE *h;
-	db_indx_t indx;
+	db_indx_t indx, *inp;
 	db_pgno_t bt_lpgno;
 	db_recno_t recno;
 	u_int32_t sflags;
@@ -2412,6 +2405,7 @@ fast_search:	/*
 		if (ret != 0)
 			goto fast_miss;
 
+		inp = P_INP(dbp, h);
 		/*
 		 * It's okay if the page type isn't right or it's empty, it
 		 * just means that the world changed.
@@ -2450,7 +2444,7 @@ fast_search:	/*
 			if (flags == DB_KEYLAST)
 				goto fast_hit;
 			for (;
-			    indx > 0 && h->inp[indx - P_INDX] == h->inp[indx];
+			    indx > 0 && inp[indx - P_INDX] == inp[indx];
 			    indx -= P_INDX)
 				;
 			goto fast_hit;
@@ -2477,7 +2471,7 @@ try_begin:	if (h->prev_pgno == PGNO_INVALID) {
 				goto fast_hit;
 			for (;
 			    indx < (db_indx_t)(NUM_ENT(h) - P_INDX) &&
-			    h->inp[indx] == h->inp[indx + P_INDX];
+			    inp[indx] == inp[indx + P_INDX];
 			    indx += P_INDX)
 				;
 			goto fast_hit;
@@ -2524,12 +2518,15 @@ search:		if ((ret = __bam_search(dbc, root_pgno,
 	/*
 	 * If we inserted a key into the first or last slot of the tree,
 	 * remember where it was so we can do it more quickly next time.
+	 * If there are duplicates and we are inserting into the last slot,
+	 * the cursor will point _to_ the last item, not after it, which
+	 * is why we subtract P_INDX below.
 	 */
 	if (TYPE(cp->page) == P_LBTREE &&
 	    (flags == DB_KEYFIRST || flags == DB_KEYLAST))
 		t->bt_lpgno =
 		    (NEXT_PGNO(cp->page) == PGNO_INVALID &&
-		    cp->indx >= NUM_ENT(cp->page)) ||
+		    cp->indx >= NUM_ENT(cp->page) - P_INDX) ||
 		    (PREV_PGNO(cp->page) == PGNO_INVALID &&
 		    cp->indx == 0) ? cp->pgno : PGNO_INVALID;
 	return (0);
@@ -2567,7 +2564,7 @@ __bam_c_physdel(dbc)
 	 * space will never be reused unless the exact same key is specified.
 	 */
 	if (delete_page &&
-	    !F_ISSET(dbc, DBC_OPD) && F_ISSET(dbp, DB_BT_REVSPLIT))
+	    !F_ISSET(dbc, DBC_OPD) && F_ISSET(dbp, DB_AM_REVSPLITOFF))
 		delete_page = 0;
 
 	/*
@@ -2693,10 +2690,10 @@ __bam_c_physdel(dbc)
 		 */
 		switch (TYPE(h)) {
 		case P_IBTREE:
-			pgno = GET_BINTERNAL(h, 0)->pgno;
+			pgno = GET_BINTERNAL(dbp, h, 0)->pgno;
 			break;
 		case P_IRECNO:
-			pgno = GET_RINTERNAL(h, 0)->pgno;
+			pgno = GET_RINTERNAL(dbp, h, 0)->pgno;
 			break;
 		default:
 			return (__db_pgfmt(dbp->dbenv, PGNO(h)));
@@ -2763,7 +2760,7 @@ __bam_c_getstack(dbc)
 	/* Get a write-locked stack for the page. */
 	exact = 0;
 	ret = __bam_search(dbc, PGNO_INVALID,
-	     &dbt, S_KEYFIRST, 1, NULL, &exact);
+	    &dbt, S_KEYFIRST, 1, NULL, &exact);
 
 err:	/* Discard the key and the page. */
 	if ((t_ret = mpf->put(mpf, h, 0)) != 0 && ret == 0)
@@ -2787,7 +2784,8 @@ __bam_isopd(dbc, pgnop)
 	if (TYPE(dbc->internal->page) != P_LBTREE)
 		return (0);
 
-	bo = GET_BOVERFLOW(dbc->internal->page, dbc->internal->indx + O_INDX);
+	bo = GET_BOVERFLOW(dbc->dbp,
+	    dbc->internal->page, dbc->internal->indx + O_INDX);
 	if (B_TYPE(bo->type) == B_DUPLICATE) {
 		*pgnop = bo->pgno;
 		return (1);
