@@ -1,7 +1,7 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2001
+ * Copyright (c) 1996-2002
  *	Sleepycat Software.  All rights reserved.
  */
 
@@ -9,9 +9,9 @@
 
 #ifndef lint
 static const char copyright[] =
-    "Copyright (c) 1996-2001\nSleepycat Software Inc.  All rights reserved.\n";
+    "Copyright (c) 1996-2002\nSleepycat Software Inc.  All rights reserved.\n";
 static const char revid[] =
-    "Id: db_stat.c,v 11.84 2001/11/03 18:43:46 bostic Exp ";
+    "Id: db_stat.c,v 11.125 2002/08/08 15:26:15 bostic Exp ";
 #endif
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -35,21 +35,18 @@ static const char revid[] =
 #endif
 
 #include "db_int.h"
-#include "db_page.h"
-#include "db_shash.h"
-#include "lock.h"
-#include "mp.h"
-#include "clib_ext.h"
+#include "dbinc/db_page.h"
 
 #define	PCT(f, t, pgsize)						\
     ((t) == 0 ? 0 :							\
     (((double)(((t) * (pgsize)) - (f)) / ((t) * (pgsize))) * 100))
 
-typedef enum { T_NOTSET, T_DB, T_ENV, T_LOCK, T_LOG, T_MPOOL, T_TXN } test_t;
+typedef enum { T_NOTSET,
+    T_DB, T_ENV, T_LOCK, T_LOG, T_MPOOL, T_REP, T_TXN } test_t;
 
 int	 db_stat_argcheck __P((char *, const char *));
 int	 db_stat_btree_stats __P((DB_ENV *, DB *, DB_BTREE_STAT *, int));
-int	 db_stat_db_init __P((DB_ENV *, char *, test_t));
+int	 db_stat_db_init __P((DB_ENV *, char *, test_t, u_int32_t, int *));
 void	 db_stat_dl __P((const char *, u_long));
 void	 db_stat_dl_bytes __P((const char *, u_long, u_long, u_long));
 int	 db_stat_env_stats __P((DB_ENV *, u_int32_t));
@@ -60,6 +57,7 @@ int	 db_stat_main __P((int, char *[]));
 int	 db_stat_mpool_stats __P((DB_ENV *, char *, u_int32_t));
 void	 db_stat_prflags __P((u_int32_t, const FN *));
 int	 db_stat_queue_stats __P((DB_ENV *, DB *, int));
+int	 db_stat_rep_stats __P((DB_ENV *, u_int32_t));
 int	 db_stat_txn_compare __P((const void *, const void *));
 int	 db_stat_txn_stats __P((DB_ENV *, u_int32_t));
 int	 db_stat_usage __P((void));
@@ -91,27 +89,28 @@ db_stat_main(argc, argv)
 	DB_BTREE_STAT *sp;
 	DB *alt_dbp, *dbp;
 	test_t ttype;
-	int ch, checked, d_close, e_close, exitval, fast, flags, nflag, ret;
-	char *db, *home, *internal, *subdb;
+	u_int32_t cache;
+	int ch, checked, d_close, e_close, exitval, fast, flags;
+	int nflag, private, resize, ret;
+	char *db, *home, *internal, *passwd, *subdb;
 
 	if ((ret = db_stat_version_check(progname)) != 0)
 		return (ret);
 
 	dbp = NULL;
 	ttype = T_NOTSET;
-	nflag = 0;
-	fast = 0;
-	d_close = e_close = exitval = 0;
-	db = home = internal = subdb = NULL;
-	flags = 0;
+	cache = MEGABYTE;
+	d_close = e_close = exitval = fast = flags = nflag = private = 0;
+	db = home = internal = passwd = subdb = NULL;
+
 	__db_getopt_reset = 1;
-	while ((ch = getopt(argc, argv, "C:cd:efh:lM:mNs:tVZ")) != EOF)
+	while ((ch = getopt(argc, argv, "C:cd:efh:lM:mNP:rs:tVZ")) != EOF)
 		switch (ch) {
 		case 'C':
 			if (ttype != T_NOTSET)
 				goto argcombo;
 			ttype = T_LOCK;
-			if (!db_stat_argcheck(internal = optarg, "Acflmo"))
+			if (!db_stat_argcheck(internal = optarg, "Aclmop"))
 				return (db_stat_usage());
 			break;
 		case 'c':
@@ -145,7 +144,7 @@ db_stat_main(argc, argv)
 			if (ttype != T_NOTSET)
 				goto argcombo;
 			ttype = T_MPOOL;
-			if (!db_stat_argcheck(internal = optarg, "Ahlm"))
+			if (!db_stat_argcheck(internal = optarg, "Ahm"))
 				return (db_stat_usage());
 			break;
 		case 'm':
@@ -156,6 +155,20 @@ db_stat_main(argc, argv)
 		case 'N':
 			nflag = 1;
 			break;
+		case 'P':
+			passwd = strdup(optarg);
+			memset(optarg, 0, strlen(optarg));
+			if (passwd == NULL) {
+				fprintf(stderr, "%s: strdup: %s\n",
+				    progname, strerror(errno));
+				return (EXIT_FAILURE);
+			}
+			break;
+		case 'r':
+			if (ttype != T_NOTSET)
+				goto argcombo;
+			ttype = T_REP;
+			break;
 		case 's':
 			if (ttype != T_DB && ttype != T_NOTSET)
 				goto argcombo;
@@ -205,7 +218,7 @@ argcombo:			fprintf(stderr,
 	 * Create an environment object and initialize it for error
 	 * reporting.
 	 */
-	if ((ret = db_env_create(&dbenv, 0)) != 0) {
+retry:	if ((ret = db_env_create(&dbenv, 0)) != 0) {
 		fprintf(stderr,
 		    "%s: db_env_create: %s\n", progname, db_strerror(ret));
 		goto shutdown;
@@ -220,14 +233,20 @@ argcombo:			fprintf(stderr,
 			dbenv->err(dbenv, ret, "set_flags: DB_NOLOCKING");
 			goto shutdown;
 		}
-		if ((ret = dbenv->set_flags(dbenv, DB_NOPANIC, 0)) != 0) {
+		if ((ret = dbenv->set_flags(dbenv, DB_NOPANIC, 1)) != 0) {
 			dbenv->err(dbenv, ret, "set_flags: DB_NOPANIC");
 			goto shutdown;
 		}
 	}
 
+	if (passwd != NULL &&
+	    (ret = dbenv->set_encrypt(dbenv, passwd, DB_ENCRYPT_AES)) != 0) {
+		dbenv->err(dbenv, ret, "set_passwd");
+		goto shutdown;
+	}
+
 	/* Initialize the environment. */
-	if (db_stat_db_init(dbenv, home, ttype) != 0)
+	if (db_stat_db_init(dbenv, home, ttype, cache, &private) != 0)
 		goto shutdown;
 
 	switch (ttype) {
@@ -239,13 +258,29 @@ argcombo:			fprintf(stderr,
 			dbenv->err(dbenv, ret, "db_create");
 			goto shutdown;
 		}
+		d_close = 1;
 
-		if ((ret =
-		    dbp->open(dbp, db, subdb, DB_UNKNOWN, DB_RDONLY, 0)) != 0) {
-			dbp->err(dbp, ret, "open: %s", db);
+		if ((ret = dbp->open(dbp,
+		    NULL, db, subdb, DB_UNKNOWN, DB_RDONLY, 0)) != 0) {
+			dbp->err(dbp, ret, "DB->open: %s", db);
 			goto shutdown;
 		}
 
+		/* Check if cache is too small for this DB's pagesize. */
+		if (private) {
+			if ((ret =
+			    __db_util_cache(dbenv, dbp, &cache, &resize)) != 0)
+				goto shutdown;
+			if (resize) {
+				(void)dbp->close(dbp, 0);
+				d_close = 0;
+
+				(void)dbenv->close(dbenv, 0);
+				e_close = 0;
+				goto retry;
+			}
+		}
+
 		/*
 		 * See if we can open this db read/write to update counts.
 		 * If its a master-db then we cannot.  So check to see,
@@ -253,10 +288,9 @@ argcombo:			fprintf(stderr,
 		 */
 		checked = 0;
 		if (subdb == NULL && dbp->type == DB_BTREE) {
-			if ((ret =
-			dbp->stat(dbp, &sp, DB_FAST_STAT)) != 0) {
-				dbp->err(dbp, ret, "dbp->stat");
-				return (EXIT_FAILURE);
+			if ((ret = dbp->stat(dbp, &sp, DB_FAST_STAT)) != 0) {
+				dbp->err(dbp, ret, "DB->stat");
+				goto shutdown;
 			}
 			checked = 1;
 		}
@@ -268,16 +302,21 @@ argcombo:			fprintf(stderr,
 				dbenv->err(dbenv, ret, "db_create");
 				goto shutdown;
 			}
-			if ((ret = dbp->open(alt_dbp,
-			    db, subdb, DB_UNKNOWN, 0, 0)) == 0) {
-				(void)dbp->close(dbp, 0);
-				dbp = alt_dbp;
+			if ((ret = dbp->open(alt_dbp, NULL,
+			    db, subdb, DB_UNKNOWN, 0, 0)) != 0) {
+				dbenv->err(dbenv,
+				   ret, "DB->open: %s:%s", db, subdb);
+				(void)alt_dbp->close(alt_dbp, 0);
+				goto shutdown;
 			}
+
+			(void)dbp->close(dbp, 0);
+			dbp = alt_dbp;
+
 			/* Need to run again to update counts */
 			checked = 0;
 		}
 
-		d_close = 1;
 		switch (dbp->type) {
 		case DB_BTREE:
 		case DB_RECNO:
@@ -294,33 +333,37 @@ argcombo:			fprintf(stderr,
 				goto shutdown;
 			break;
 		case DB_UNKNOWN:
-			abort();		/* Impossible. */
-			/* NOTREACHED */
+			dbenv->errx(dbenv, "Unknown database type.");
+			goto shutdown;
 		}
 		break;
 	case T_ENV:
 		if (db_stat_env_stats(dbenv, flags))
-			exitval = 1;
+			goto shutdown;
 		break;
 	case T_LOCK:
 		if (db_stat_lock_stats(dbenv, internal, flags))
-			exitval = 1;
+			goto shutdown;
 		break;
 	case T_LOG:
 		if (db_stat_log_stats(dbenv, flags))
-			exitval = 1;
+			goto shutdown;
 		break;
 	case T_MPOOL:
 		if (db_stat_mpool_stats(dbenv, internal, flags))
-			exitval = 1;
+			goto shutdown;
+		break;
+	case T_REP:
+		if (db_stat_rep_stats(dbenv, flags))
+			goto shutdown;
 		break;
 	case T_TXN:
 		if (db_stat_txn_stats(dbenv, flags))
-			exitval = 1;
+			goto shutdown;
 		break;
 	case T_NOTSET:
-		abort();			/* Impossible. */
-		/* NOTREACHED */
+		dbenv->errx(dbenv, "Unknown statistics flag.");
+		goto shutdown;
 	}
 
 	if (0) {
@@ -442,7 +485,7 @@ db_stat_btree_stats(dbenv, dbp, msp, fast)
 	if (msp != NULL)
 		sp = msp;
 	else if ((ret = dbp->stat(dbp, &sp, fast)) != 0) {
-		dbp->err(dbp, ret, "dbp->stat");
+		dbp->err(dbp, ret, "DB->stat");
 		return (1);
 	}
 
@@ -497,6 +540,8 @@ db_stat_btree_stats(dbenv, dbp, msp, fast)
 
 	db_stat_dl("Number of pages on the free list.\n", (u_long)sp->bt_free);
 
+	free(sp);
+
 	return (0);
 }
 
@@ -521,7 +566,7 @@ db_stat_hash_stats(dbenv, dbp, fast)
 	COMPQUIET(dbenv, NULL);
 
 	if ((ret = dbp->stat(dbp, &sp, fast)) != 0) {
-		dbp->err(dbp, ret, "dbp->stat");
+		dbp->err(dbp, ret, "DB->stat");
 		return (1);
 	}
 
@@ -529,7 +574,6 @@ db_stat_hash_stats(dbenv, dbp, fast)
 	printf("%lu\tHash version number.\n", (u_long)sp->hash_version);
 	db_stat_prflags(sp->hash_metaflags, fn);
 	db_stat_dl("Underlying database page size.\n", (u_long)sp->hash_pagesize);
-	db_stat_dl("Specified number of elements.\n", (u_long)sp->hash_nelem);
 	db_stat_dl("Specified fill factor.\n", (u_long)sp->hash_ffactor);
 	db_stat_dl("Number of keys in the database.\n", (u_long)sp->hash_nkeys);
 	db_stat_dl("Number of data items in the database.\n", (u_long)sp->hash_ndata);
@@ -559,6 +603,8 @@ db_stat_hash_stats(dbenv, dbp, fast)
 
 	db_stat_dl("Number of pages on the free list.\n", (u_long)sp->hash_free);
 
+	free(sp);
+
 	return (0);
 }
 
@@ -578,7 +624,7 @@ db_stat_queue_stats(dbenv, dbp, fast)
 	COMPQUIET(dbenv, NULL);
 
 	if ((ret = dbp->stat(dbp, &sp, fast)) != 0) {
-		dbp->err(dbp, ret, "dbp->stat");
+		dbp->err(dbp, ret, "DB->stat");
 		return (1);
 	}
 
@@ -600,7 +646,9 @@ db_stat_queue_stats(dbenv, dbp, fast)
 	    PCT(sp->qs_pgfree, sp->qs_pages, sp->qs_pagesize));
 	printf("%lu\tFirst undeleted record.\n", (u_long)sp->qs_first_recno);
 	printf(
-	    "%lu\tLast allocated record number.\n", (u_long)sp->qs_cur_recno);
+	    "%lu\tNext available record number.\n", (u_long)sp->qs_cur_recno);
+
+	free(sp);
 
 	return (0);
 }
@@ -632,34 +680,47 @@ db_stat_lock_stats(dbenv, internal, flags)
 		return (1);
 	}
 
-	printf("%lu Last allocated locker ID.\n", (u_long)sp->st_lastid);
+	db_stat_dl("Last allocated locker ID.\n", (u_long)sp->st_id);
+	db_stat_dl("Current maximum unused locker ID.\n", (u_long)sp->st_cur_maxid);
 	db_stat_dl("Number of lock modes.\n", (u_long)sp->st_nmodes);
 	db_stat_dl("Maximum number of locks possible.\n", (u_long)sp->st_maxlocks);
 	db_stat_dl("Maximum number of lockers possible.\n", (u_long)sp->st_maxlockers);
-	db_stat_dl("Maximum number of objects possible.\n", (u_long)sp->st_maxobjects);
-	db_stat_dl("Current locks.\n", (u_long)sp->st_nlocks);
-	db_stat_dl("Maximum number of locks so far.\n", (u_long)sp->st_maxnlocks);
-	db_stat_dl("Current number of lockers.\n", (u_long)sp->st_nlockers);
-	db_stat_dl("Maximum number  lockers so far.\n", (u_long)sp->st_maxnlockers);
-	db_stat_dl("Current number lock objects.\n", (u_long)sp->st_nobjects);
-	db_stat_dl("Maximum number of lock objects so far.\n",
+	db_stat_dl("Maximum number of lock objects possible.\n",
+	    (u_long)sp->st_maxobjects);
+	db_stat_dl("Number of current locks.\n", (u_long)sp->st_nlocks);
+	db_stat_dl("Maximum number of locks at any one time.\n",
+	    (u_long)sp->st_maxnlocks);
+	db_stat_dl("Number of current lockers.\n", (u_long)sp->st_nlockers);
+	db_stat_dl("Maximum number of lockers at any one time.\n",
+	    (u_long)sp->st_maxnlockers);
+	db_stat_dl("Number of current lock objects.\n", (u_long)sp->st_nobjects);
+	db_stat_dl("Maximum number of lock objects at any one time.\n",
 	    (u_long)sp->st_maxnobjects);
-	db_stat_dl("Number of lock requests.\n", (u_long)sp->st_nrequests);
-	db_stat_dl("Number of lock releases.\n", (u_long)sp->st_nreleases);
-	db_stat_dl("Number of lock requests that would have waited.\n",
+	db_stat_dl("Total number of locks requested.\n", (u_long)sp->st_nrequests);
+	db_stat_dl("Total number of locks released.\n", (u_long)sp->st_nreleases);
+	db_stat_dl(
+  "Total number of lock requests failing because DB_LOCK_NOWAIT was set.\n",
 	    (u_long)sp->st_nnowaits);
-	db_stat_dl("Number of lock conflicts.\n", (u_long)sp->st_nconflicts);
+	db_stat_dl(
+  "Total number of locks not immediately available due to conflicts.\n",
+	    (u_long)sp->st_nconflicts);
 	db_stat_dl("Number of deadlocks.\n", (u_long)sp->st_ndeadlocks);
-	db_stat_dl("Number of transaction timeouts.\n", (u_long)sp->st_ntxntimeouts);
-	db_stat_dl("Number of lock timeouts.\n", (u_long)sp->st_nlocktimeouts);
+	db_stat_dl("Lock timeout value.\n", (u_long)sp->st_locktimeout);
+	db_stat_dl("Number of locks that have timed out.\n",
+	    (u_long)sp->st_nlocktimeouts);
+	db_stat_dl("Transaction timeout value.\n", (u_long)sp->st_txntimeout);
+	db_stat_dl("Number of transactions that have timed out.\n",
+	    (u_long)sp->st_ntxntimeouts);
 
-	db_stat_dl_bytes("Lock region size",
+	db_stat_dl_bytes("The size of the lock region.",
 	    (u_long)0, (u_long)0, (u_long)sp->st_regsize);
 	db_stat_dl("The number of region locks granted after waiting.\n",
 	    (u_long)sp->st_region_wait);
 	db_stat_dl("The number of region locks granted without waiting.\n",
 	    (u_long)sp->st_region_nowait);
 
+	free(sp);
+
 	return (0);
 }
 
@@ -685,13 +746,15 @@ db_stat_log_stats(dbenv, flags)
 	db_stat_dl_bytes("Log record cache size",
 	    (u_long)0, (u_long)0, (u_long)sp->st_lg_bsize);
 	printf("%#o\tLog file mode.\n", sp->st_mode);
-	if (sp->st_lg_max % MEGABYTE == 0)
-		printf("%luMb\tLog file size.\n",
-		    (u_long)sp->st_lg_max / MEGABYTE);
-	else if (sp->st_lg_max % 1024 == 0)
-		printf("%luKb\tLog file size.\n", (u_long)sp->st_lg_max / 1024);
+	if (sp->st_lg_size % MEGABYTE == 0)
+		printf("%luMb\tCurrent log file size.\n",
+		    (u_long)sp->st_lg_size / MEGABYTE);
+	else if (sp->st_lg_size % 1024 == 0)
+		printf("%luKb\tCurrent log file size.\n",
+		    (u_long)sp->st_lg_size / 1024);
 	else
-		printf("%lu\tLog file size.\n", (u_long)sp->st_lg_max);
+		printf("%lu\tCurrent log file size.\n",
+		    (u_long)sp->st_lg_size);
 	db_stat_dl_bytes("Log bytes written",
 	    (u_long)0, (u_long)sp->st_w_mbytes, (u_long)sp->st_w_bytes);
 	db_stat_dl_bytes("Log bytes written since last checkpoint",
@@ -707,8 +770,6 @@ db_stat_log_stats(dbenv, flags)
 
 	db_stat_dl("Max commits in a log flush.\n", (u_long)sp->st_maxcommitperflush);
 	db_stat_dl("Min commits in a log flush.\n", (u_long)sp->st_mincommitperflush);
-	db_stat_dl("Number of log flushes containing a transaction commit.\n",
-	    (u_long)sp->st_flushcommit);
 
 	db_stat_dl_bytes("Log region size",
 	    (u_long)0, (u_long)0, (u_long)sp->st_regsize);
@@ -717,6 +778,8 @@ db_stat_log_stats(dbenv, flags)
 	db_stat_dl("The number of region locks granted without waiting.\n",
 	    (u_long)sp->st_region_nowait);
 
+	free(sp);
+
 	return (0);
 }
 
@@ -751,14 +814,15 @@ db_stat_mpool_stats(dbenv, internal, flags)
 	db_stat_dl_bytes("Total cache size",
 	    (u_long)gsp->st_gbytes, (u_long)0, (u_long)gsp->st_bytes);
 	db_stat_dl("Number of caches.\n", (u_long)gsp->st_ncache);
-	db_stat_dl("Pool individual cache size.\n", (u_long)gsp->st_regsize);
+	db_stat_dl_bytes("Pool individual cache size",
+	    (u_long)0, (u_long)0, (u_long)gsp->st_regsize);
+	db_stat_dl("Requested pages mapped into the process' address space.\n",
+	    (u_long)gsp->st_map);
 	db_stat_dl("Requested pages found in the cache", (u_long)gsp->st_cache_hit);
 	if (gsp->st_cache_hit + gsp->st_cache_miss != 0)
 		printf(" (%.0f%%)", ((double)gsp->st_cache_hit /
 		    (gsp->st_cache_hit + gsp->st_cache_miss)) * 100);
 	printf(".\n");
-	db_stat_dl("Requested pages mapped into the process' address space.\n",
-	    (u_long)gsp->st_map);
 	db_stat_dl("Requested pages not found in the cache.\n",
 	    (u_long)gsp->st_cache_miss);
 	db_stat_dl("Pages created in the cache.\n", (u_long)gsp->st_page_create);
@@ -769,11 +833,13 @@ db_stat_mpool_stats(dbenv, internal, flags)
 	    (u_long)gsp->st_ro_evict);
 	db_stat_dl("Dirty pages forced from the cache.\n",
 	    (u_long)gsp->st_rw_evict);
-	db_stat_dl("Dirty buffers written by trickle-sync thread.\n",
+	db_stat_dl("Dirty pages written by trickle-sync thread.\n",
 	    (u_long)gsp->st_page_trickle);
-	db_stat_dl("Current clean buffer count.\n",
+	db_stat_dl("Current total page count.\n",
+	    (u_long)gsp->st_pages);
+	db_stat_dl("Current clean page count.\n",
 	    (u_long)gsp->st_page_clean);
-	db_stat_dl("Current dirty buffer count.\n",
+	db_stat_dl("Current dirty page count.\n",
 	    (u_long)gsp->st_page_dirty);
 	db_stat_dl("Number of hash buckets used for page location.\n",
 	    (u_long)gsp->st_hash_buckets);
@@ -783,15 +849,33 @@ db_stat_mpool_stats(dbenv, internal, flags)
 	    (u_long)gsp->st_hash_longest);
 	db_stat_dl("Total number of hash buckets examined for page location.\n",
 	    (u_long)gsp->st_hash_examined);
+	db_stat_dl("The number of hash bucket locks granted without waiting.\n",
+	    (u_long)gsp->st_hash_nowait);
+	db_stat_dl("The number of hash bucket locks granted after waiting.\n",
+	    (u_long)gsp->st_hash_wait);
+	db_stat_dl("The maximum number of times any hash bucket lock was waited for.\n",
+	    (u_long)gsp->st_hash_max_wait);
 	db_stat_dl("The number of region locks granted without waiting.\n",
 	    (u_long)gsp->st_region_nowait);
 	db_stat_dl("The number of region locks granted after waiting.\n",
 	    (u_long)gsp->st_region_wait);
+	db_stat_dl("The number of page allocations.\n",
+	    (u_long)gsp->st_alloc);
+	db_stat_dl("The number of hash buckets examined during allocations\n",
+	    (u_long)gsp->st_alloc_buckets);
+	db_stat_dl("The max number of hash buckets examined for an allocation\n",
+	    (u_long)gsp->st_alloc_max_buckets);
+	db_stat_dl("The number of pages examined during allocations\n",
+	    (u_long)gsp->st_alloc_pages);
+	db_stat_dl("The max number of pages examined for an allocation\n",
+	    (u_long)gsp->st_alloc_max_pages);
 
 	for (; fsp != NULL && *fsp != NULL; ++fsp) {
 		printf("%s\n", DB_LINE);
 		printf("Pool File: %s\n", (*fsp)->file_name);
 		db_stat_dl("Page size.\n", (u_long)(*fsp)->st_pagesize);
+		db_stat_dl("Requested pages mapped into the process' address space.\n",
+		    (u_long)(*fsp)->st_map);
 		db_stat_dl("Requested pages found in the cache",
 		    (u_long)(*fsp)->st_cache_hit);
 		if ((*fsp)->st_cache_hit + (*fsp)->st_cache_miss != 0)
@@ -799,8 +883,6 @@ db_stat_mpool_stats(dbenv, internal, flags)
 			    ((*fsp)->st_cache_hit + (*fsp)->st_cache_miss)) *
 			    100);
 		printf(".\n");
-		db_stat_dl("Requested pages mapped into the process' address space.\n",
-		    (u_long)(*fsp)->st_map);
 		db_stat_dl("Requested pages not found in the cache.\n",
 		    (u_long)(*fsp)->st_cache_miss);
 		db_stat_dl("Pages created in the cache.\n",
@@ -811,6 +893,123 @@ db_stat_mpool_stats(dbenv, internal, flags)
 		    (u_long)(*fsp)->st_page_out);
 	}
 
+	free(gsp);
+
+	return (0);
+}
+
+/*
+ * rep_stats --
+ *	Display replication statistics.
+ */
+int
+db_stat_rep_stats(dbenv, flags)
+	DB_ENV *dbenv;
+	u_int32_t flags;
+{
+	DB_REP_STAT *sp;
+	int is_client, ret;
+	const char *p;
+
+	if ((ret = dbenv->rep_stat(dbenv, &sp, flags)) != 0) {
+		dbenv->err(dbenv, ret, NULL);
+		return (1);
+	}
+
+	is_client = 0;
+	switch (sp->st_status) {
+	case DB_REP_MASTER:
+		printf("Environment configured as a replication master.\n");
+		break;
+	case DB_REP_CLIENT:
+		printf("Environment configured as a replication client.\n");
+		is_client = 1;
+		break;
+	case DB_REP_LOGSONLY:
+		printf("Environment configured as a logs-only replica.\n");
+		is_client = 1;
+		break;
+	default:
+		printf("Environment not configured for replication.\n");
+		break;
+	}
+
+	printf("%lu/%lu\t%s\n",
+	    (u_long)sp->st_next_lsn.file, (u_long)sp->st_next_lsn.offset,
+	    is_client ? "Next LSN expected." : "Next LSN to be used.");
+	p = sp->st_waiting_lsn.file == 0 ?
+	    "Not waiting for any missed log records." :
+	    "LSN of first missed log record being waited for.";
+	printf("%lu/%lu\t%s\n",
+	    (u_long)sp->st_waiting_lsn.file, (u_long)sp->st_waiting_lsn.offset,
+	    p);
+
+	db_stat_dl("Number of duplicate master conditions detected.\n",
+	    (u_long)sp->st_dupmasters);
+	if (sp->st_env_id != DB_EID_INVALID)
+		db_stat_dl("Current environment ID.\n", (u_long)sp->st_env_id);
+	else
+		printf("No current environment ID.\n");
+	db_stat_dl("Current environment priority.\n", (u_long)sp->st_env_priority);
+	db_stat_dl("Current generation number.\n", (u_long)sp->st_gen);
+	db_stat_dl("Number of duplicate log records received.\n",
+	    (u_long)sp->st_log_duplicated);
+	db_stat_dl("Number of log records currently queued.\n",
+	    (u_long)sp->st_log_queued);
+	db_stat_dl("Maximum number of log records ever queued at once.\n",
+	    (u_long)sp->st_log_queued_max);
+	db_stat_dl("Total number of log records queued.\n",
+	    (u_long)sp->st_log_queued_total);
+	db_stat_dl("Number of log records received and appended to the log.\n",
+	    (u_long)sp->st_log_records);
+	db_stat_dl("Number of log records missed and requested.\n",
+	    (u_long)sp->st_log_requested);
+	if (sp->st_master != DB_EID_INVALID)
+		db_stat_dl("Current master ID.\n", (u_long)sp->st_master);
+	else
+		printf("No current master ID.\n");
+	db_stat_dl("Number of times the master has changed.\n",
+	    (u_long)sp->st_master_changes);
+	db_stat_dl("Number of messages received with a bad generation number.\n",
+	    (u_long)sp->st_msgs_badgen);
+	db_stat_dl("Number of messages received and processed.\n",
+	    (u_long)sp->st_msgs_processed);
+	db_stat_dl("Number of messages ignored due to pending recovery.\n",
+	    (u_long)sp->st_msgs_recover);
+	db_stat_dl("Number of failed message sends.\n",
+	    (u_long)sp->st_msgs_send_failures);
+	db_stat_dl("Number of messages sent.\n", (u_long)sp->st_msgs_sent);
+	db_stat_dl("Number of new site messages received.\n", (u_long)sp->st_newsites);
+	db_stat_dl("Transmission limited.\n", (u_long)sp->st_nthrottles);
+	db_stat_dl("Number of outdated conditions detected.\n",
+	    (u_long)sp->st_outdated);
+	db_stat_dl("Number of transactions applied.\n", (u_long)sp->st_txns_applied);
+
+	db_stat_dl("Number of elections held.\n", (u_long)sp->st_elections);
+	db_stat_dl("Number of elections won.\n", (u_long)sp->st_elections_won);
+
+	if (sp->st_election_status == 0)
+		printf("No election in progress.\n");
+	else {
+		db_stat_dl("Current election phase.\n", (u_long)sp->st_election_status);
+		db_stat_dl("Election winner.\n",
+		    (u_long)sp->st_election_cur_winner);
+		db_stat_dl("Election generation number.\n",
+		    (u_long)sp->st_election_gen);
+		printf("%lu/%lu\tMaximum LSN of election winner.\n",
+		    (u_long)sp->st_election_lsn.file,
+		    (u_long)sp->st_election_lsn.offset);
+		db_stat_dl("Number of sites expected to participate in elections.\n",
+		    (u_long)sp->st_election_nsites);
+		db_stat_dl("Election priority.\n", (u_long)sp->st_election_priority);
+		db_stat_dl("Election tiebreaker value.\n",
+		    (u_long)sp->st_election_tiebreaker);
+		db_stat_dl("Votes received this election round.\n",
+		    (u_long)sp->st_election_votes);
+	}
+
+	free(sp);
+
 	return (0);
 }
 
@@ -837,12 +1036,6 @@ db_stat_txn_stats(dbenv, flags)
 	    "No checkpoint LSN." : "File/offset for last checkpoint LSN.";
 	printf("%lu/%lu\t%s\n",
 	    (u_long)sp->st_last_ckp.file, (u_long)sp->st_last_ckp.offset, p);
-	p = sp->st_pending_ckp.file == 0 ?
-	    "No pending checkpoint LSN." :
-	    "File/offset for last pending checkpoint LSN.";
-	printf("%lu/%lu\t%s\n",
-	    (u_long)sp->st_pending_ckp.file,
-	    (u_long)sp->st_pending_ckp.offset, p);
 	if (sp->st_time_ckp == 0)
 		printf("0\tNo checkpoint timestamp.\n");
 	else
@@ -869,7 +1062,7 @@ db_stat_txn_stats(dbenv, flags)
 	qsort(sp->st_txnarray,
 	    sp->st_nactive, sizeof(sp->st_txnarray[0]), db_stat_txn_compare);
 	for (i = 0; i < sp->st_nactive; ++i) {
-		printf("\tid: %lx; initial LSN file/offset %lu/%lu",
+		printf("\tid: %lx; begin LSN: file/offset %lu/%lu",
 		    (u_long)sp->st_txnarray[i].txnid,
 		    (u_long)sp->st_txnarray[i].lsn.file,
 		    (u_long)sp->st_txnarray[i].lsn.offset);
@@ -880,6 +1073,8 @@ db_stat_txn_stats(dbenv, flags)
 			    (u_long)sp->st_txnarray[i].parentid);
 	}
 
+	free(sp);
+
 	return (0);
 }
 
@@ -928,10 +1123,8 @@ db_stat_dl_bytes(msg, gbytes, mbytes, bytes)
 	u_long gbytes, mbytes, bytes;
 {
 	const char *sep;
-	u_long sbytes;
-	int showbytes;
 
-	sbytes = bytes;
+	/* Normalize the values. */
 	while (bytes >= MEGABYTE) {
 		++mbytes;
 		bytes -= MEGABYTE;
@@ -942,33 +1135,23 @@ db_stat_dl_bytes(msg, gbytes, mbytes, bytes)
 	}
 
 	sep = "";
-	showbytes = 0;
 	if (gbytes > 0) {
 		printf("%luGB", gbytes);
 		sep = " ";
-		showbytes = 1;
 	}
 	if (mbytes > 0) {
 		printf("%s%luMB", sep, mbytes);
 		sep = " ";
-		showbytes = 1;
 	}
 	if (bytes >= 1024) {
 		printf("%s%luKB", sep, bytes / 1024);
 		bytes %= 1024;
 		sep = " ";
-		showbytes = 1;
 	}
 	if (bytes > 0)
 		printf("%s%luB", sep, bytes);
-	else
-		if (!showbytes)
-			printf("%s%lu", sep, bytes);
 
-	printf("\t%s", msg);
-	if (showbytes)
-		printf(" (%lu bytes)", sbytes);
-	printf(".\n");
+	printf("\t%s.\n", msg);
 }
 
 /*
@@ -997,11 +1180,14 @@ db_stat_prflags(flags, fnp)
  *	Initialize the environment.
  */
 int
-db_stat_db_init(dbenv, home, ttype)
+db_stat_db_init(dbenv, home, ttype, cache, is_private)
 	DB_ENV *dbenv;
 	char *home;
 	test_t ttype;
+	u_int32_t cache;
+	int *is_private;
 {
+	u_int32_t oflags;
 	int ret;
 
 	/*
@@ -1014,6 +1200,7 @@ db_stat_db_init(dbenv, home, ttype)
 	 * pool to look at a database can be safely construed as operator
 	 * error, I think.
 	 */
+	*is_private = 0;
 	if ((ret =
 	    dbenv->open(dbenv, home, DB_JOINENV | DB_USE_ENVIRON, 0)) == 0)
 		return (0);
@@ -1033,13 +1220,17 @@ db_stat_db_init(dbenv, home, ttype)
 	 * trying to look at databases in directories other than the current
 	 * one.
 	 */
-	if ((ret = dbenv->set_cachesize(dbenv, 0, MEGABYTE, 1)) != 0) {
+	if ((ret = dbenv->set_cachesize(dbenv, 0, cache, 1)) != 0) {
 		dbenv->err(dbenv, ret, "set_cachesize");
 		return (1);
 	}
-	if ((ret = dbenv->open(dbenv, home,
-    DB_CREATE | DB_INIT_LOG | DB_INIT_MPOOL | DB_PRIVATE | DB_USE_ENVIRON,
-	    0)) == 0)
+	*is_private = 1;
+	oflags = DB_CREATE | DB_PRIVATE | DB_USE_ENVIRON;
+	if (ttype == T_DB)
+		oflags |= DB_INIT_MPOOL;
+	if (ttype == T_LOG)
+		oflags |= DB_INIT_LOG;
+	if ((ret = dbenv->open(dbenv, home, oflags, 0)) == 0)
 		return (0);
 
 	/* An environment is required. */
@@ -1065,8 +1256,9 @@ db_stat_argcheck(arg, ok_args)
 int
 db_stat_usage()
 {
-	fprintf(stderr, "usage: db_stat %s\n",
-"[-celmNtVZ] [-C Acflmo]\n\t[-d file [-f] [-s database]] [-h home] [-M Ahlm]");
+	fprintf(stderr, "%s\n\t%s\n",
+	    "usage: db_stat [-celmNrtVZ] [-C Aclmop]",
+	    "[-d file [-f] [-s database]] [-h home] [-M Ahlm] [-P password]");
 	return (EXIT_FAILURE);
 }
 
