@@ -1,13 +1,13 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2001
+ * Copyright (c) 1997-2002
  *	Sleepycat Software.  All rights reserved.
  */
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "Id: java_Db.c,v 11.57 2001/10/04 21:25:24 bostic Exp ";
+static const char revid[] = "Id: java_Db.c,v 11.80 2002/08/29 14:22:23 margo Exp ";
 #endif /* not lint */
 
 #include <jni.h>
@@ -15,10 +15,11 @@ static const char revid[] = "Id: java_Db.c,v 11.57 2001/10/04 21:25:24 bostic Ex
 #include <string.h>
 
 #include "db_int.h"
-#include "db_page.h"
-#include "btree.h"
-#include "db_ext.h"
+#include "dbinc/db_page.h"
+#include "dbinc/btree.h"
+#include "dbinc_auto/db_ext.h"
 #include "java_util.h"
+#include "java_stat_auto.h"
 #include "com_sleepycat_db_Db.h"
 
 /* This struct is used in Db.verify and its callback */
@@ -30,18 +31,21 @@ struct verify_callback_struct {
 	jmethodID writemid;
 };
 
-JAVADB_WO_ACCESS_METHOD(Db, jint, flags, DB, flags)
-JAVADB_WO_ACCESS_METHOD(Db, jint, h_1ffactor, DB, h_ffactor)
-JAVADB_WO_ACCESS_METHOD(Db, jint, h_1nelem, DB, h_nelem)
-JAVADB_WO_ACCESS_METHOD(Db, jint, lorder, DB, lorder)
-JAVADB_WO_ACCESS_METHOD(Db, jint, re_1delim, DB, re_delim)
-JAVADB_WO_ACCESS_METHOD(Db, jint, re_1len, DB, re_len)
-JAVADB_WO_ACCESS_METHOD(Db, jint, re_1pad, DB, re_pad)
-JAVADB_WO_ACCESS_METHOD(Db, jint, q_1extentsize, DB, q_extentsize)
-JAVADB_WO_ACCESS_METHOD(Db, jint, bt_1maxkey, DB, bt_maxkey)
-JAVADB_WO_ACCESS_METHOD(Db, jint, bt_1minkey, DB, bt_minkey)
+JAVADB_GET_FLD(Db, jint, flags_1raw, DB, flags)
 
-/* This only gets called once ever, at the beginning of execution
+JAVADB_SET_METH(Db, jint, flags, DB, flags)
+JAVADB_SET_METH(Db, jint, h_1ffactor, DB, h_ffactor)
+JAVADB_SET_METH(Db, jint, h_1nelem, DB, h_nelem)
+JAVADB_SET_METH(Db, jint, lorder, DB, lorder)
+JAVADB_SET_METH(Db, jint, re_1delim, DB, re_delim)
+JAVADB_SET_METH(Db, jint, re_1len, DB, re_len)
+JAVADB_SET_METH(Db, jint, re_1pad, DB, re_pad)
+JAVADB_SET_METH(Db, jint, q_1extentsize, DB, q_extentsize)
+JAVADB_SET_METH(Db, jint, bt_1maxkey, DB, bt_maxkey)
+JAVADB_SET_METH(Db, jint, bt_1minkey, DB, bt_minkey)
+
+/*
+ * This only gets called once ever, at the beginning of execution
  * and can be used to initialize unchanging methodIds, fieldIds, etc.
  */
 JNIEXPORT void JNICALL Java_com_sleepycat_db_Db_one_1time_1init
@@ -69,23 +73,25 @@ JNIEXPORT void JNICALL Java_com_sleepycat_db_Db__1init
 		set_private_dbobj(jnienv, name_DB, jthis, db);
 		dbinfo = dbji_construct(jnienv, jthis, flags);
 		set_private_info(jnienv, name_DB, jthis, dbinfo);
-		db->cj_internal = dbinfo;
+		db->api_internal = dbinfo;
 	}
 }
 
 JNIEXPORT void JNICALL Java_com_sleepycat_db_Db__1associate
-    (JNIEnv *jnienv, /*Db*/ jobject jthis, /*Db*/ jobject jsecondary,
-     /*DbSecondaryKeyCreate*/ jobject jcallback, jint flags)
+    (JNIEnv *jnienv, /*Db*/ jobject jthis, /* DbTxn */ jobject jtxn,
+     /*Db*/ jobject jsecondary, /*DbSecondaryKeyCreate*/ jobject jcallback,
+     jint flags)
 {
-	DB *db;
+	DB *db, *secondary;
 	DB_JAVAINFO *second_info;
-	DB *secondary;
+	DB_TXN *txn;
 
 	db = get_DB(jnienv, jthis);
+	txn = get_DB_TXN(jnienv, jtxn);
 	secondary = get_DB(jnienv, jsecondary);
 
-	second_info = (DB_JAVAINFO*)secondary->cj_internal;
-	dbji_set_assoc_object(second_info, jnienv, db, secondary,
+	second_info = (DB_JAVAINFO*)secondary->api_internal;
+	dbji_set_assoc_object(second_info, jnienv, db, txn, secondary,
 			      jcallback, flags);
 
 }
@@ -102,21 +108,22 @@ JNIEXPORT jint JNICALL Java_com_sleepycat_db_Db__1close
 	if (!verify_non_null(jnienv, db))
 		return (0);
 
-	/* Null out the private data to indicate the DB is invalid.
+	/*
+	 * Null out the private data to indicate the DB is invalid.
 	 * We do this in advance to help guard against multithreading
 	 * issues.
 	 */
 	set_private_dbobj(jnienv, name_DB, jthis, 0);
 
 	err = db->close(db, flags);
-	if (err != DB_INCOMPLETE)
-		verify_return(jnienv, err, 0);
+	verify_return(jnienv, err, 0);
 	dbji_dealloc(dbinfo, jnienv);
 
 	return (err);
 }
 
-/* We are being notified that the parent DbEnv has closed.
+/*
+ * We are being notified that the parent DbEnv has closed.
  * Zero out the pointer to the DB, since it is no longer
  * valid, to prevent mistakes.  The user will get a null
  * pointer exception if they try to use this Db again.
@@ -137,7 +144,7 @@ JNIEXPORT void JNICALL Java_com_sleepycat_db_Db_append_1recno_1changed
 	if (!verify_non_null(jnienv, db))
 		return;
 
-	dbinfo = (DB_JAVAINFO*)db->cj_internal;
+	dbinfo = (DB_JAVAINFO*)db->api_internal;
 	dbji_set_append_recno_object(dbinfo, jnienv, db, jcallback);
 }
 
@@ -151,7 +158,7 @@ JNIEXPORT void JNICALL Java_com_sleepycat_db_Db_bt_1compare_1changed
 	if (!verify_non_null(jnienv, db))
 		return;
 
-	dbinfo = (DB_JAVAINFO*)db->cj_internal;
+	dbinfo = (DB_JAVAINFO*)db->api_internal;
 	dbji_set_bt_compare_object(dbinfo, jnienv, db, jbtcompare);
 }
 
@@ -165,7 +172,7 @@ JNIEXPORT void JNICALL Java_com_sleepycat_db_Db_bt_1prefix_1changed
 	if (!verify_non_null(jnienv, db))
 		return;
 
-	dbinfo = (DB_JAVAINFO*)db->cj_internal;
+	dbinfo = (DB_JAVAINFO*)db->api_internal;
 	dbji_set_bt_prefix_object(dbinfo, jnienv, db, jbtprefix);
 }
 
@@ -199,16 +206,15 @@ JNIEXPORT jint JNICALL Java_com_sleepycat_db_Db_del
 		return (0);
 
 	dbtxnid = get_DB_TXN(jnienv, txnid);
-	if (locked_dbt_get(&lkey, jnienv, key, inOp) != 0)
+	if (locked_dbt_get(&lkey, jnienv, db->dbenv, key, inOp) != 0)
 		goto out;
 
 	err = db->del(db, dbtxnid, &lkey.javainfo->dbt, dbflags);
-	if (err != DB_NOTFOUND) {
+	if (!DB_RETOK_DBDEL(err))
 		verify_return(jnienv, err, 0);
-	}
 
  out:
-	locked_dbt_put(&lkey, jnienv);
+	locked_dbt_put(&lkey, jnienv, db->dbenv);
 	return (err);
 }
 
@@ -222,7 +228,7 @@ JNIEXPORT void JNICALL Java_com_sleepycat_db_Db_dup_1compare_1changed
 	if (!verify_non_null(jnienv, db))
 		return;
 
-	dbinfo = (DB_JAVAINFO*)db->cj_internal;
+	dbinfo = (DB_JAVAINFO*)db->api_internal;
 	dbji_set_dup_compare_object(dbinfo, jnienv, db, jdupcompare);
 }
 
@@ -238,7 +244,7 @@ JNIEXPORT void JNICALL Java_com_sleepycat_db_Db_err
 	if (!verify_non_null(jnienv, db))
 		goto out;
 
-	db->err(db, ecode, ls_msg.string);
+	db->err(db, ecode, "%s", ls_msg.string);
 
  out:
 	locked_string_put(&ls_msg, jnienv);
@@ -255,7 +261,7 @@ JNIEXPORT void JNICALL Java_com_sleepycat_db_Db_errx
 	if (!verify_non_null(jnienv, db))
 		goto out;
 
-	db->errx(db, ls_msg.string);
+	db->errx(db, "%s", ls_msg.string);
 
  out:
 	locked_string_put(&ls_msg, jnienv);
@@ -277,6 +283,25 @@ JNIEXPORT jint JNICALL Java_com_sleepycat_db_Db_fd
 	return (return_value);
 }
 
+JNIEXPORT void JNICALL Java_com_sleepycat_db_Db_set_1encrypt
+  (JNIEnv *jnienv, /*Db*/ jobject jthis, jstring jpasswd, jint flags)
+{
+	int err;
+	DB *db;
+	LOCKED_STRING ls_passwd;
+
+	db = get_DB(jnienv, jthis);
+	if (!verify_non_null(jnienv, db))
+		return;
+	if (locked_string_get(&ls_passwd, jnienv, jpasswd) != 0)
+		goto out;
+
+	err = db->set_encrypt(db, ls_passwd.string, flags);
+	verify_return(jnienv, err, 0);
+
+out:	locked_string_put(&ls_passwd, jnienv);
+}
+
 JNIEXPORT void JNICALL Java_com_sleepycat_db_Db_feedback_1changed
   (JNIEnv *jnienv, /*Db*/ jobject jthis, /*DbFeedback*/ jobject jfeedback)
 {
@@ -287,7 +312,7 @@ JNIEXPORT void JNICALL Java_com_sleepycat_db_Db_feedback_1changed
 	if (!verify_non_null(jnienv, db))
 		return;
 
-	dbinfo = (DB_JAVAINFO*)db->cj_internal;
+	dbinfo = (DB_JAVAINFO*)db->api_internal;
 	dbji_set_feedback_object(dbinfo, jnienv, db, jfeedback);
 }
 
@@ -297,6 +322,7 @@ JNIEXPORT jint JNICALL Java_com_sleepycat_db_Db_get
 {
 	int err, op_flags, retry;
 	DB *db;
+	DB_ENV *dbenv;
 	OpKind keyop, dataop;
 	DB_TXN *dbtxnid;
 	LOCKED_DBT lkey, ldata;
@@ -305,6 +331,7 @@ JNIEXPORT jint JNICALL Java_com_sleepycat_db_Db_get
 	db = get_DB(jnienv, jthis);
 	if (!verify_non_null(jnienv, db))
 		goto out3;
+	dbenv = db->dbenv;
 
 	/* Depending on flags, the key may be input/output. */
 	keyop = inOp;
@@ -320,28 +347,30 @@ JNIEXPORT jint JNICALL Java_com_sleepycat_db_Db_get
 
 	dbtxnid = get_DB_TXN(jnienv, txnid);
 
-	if (locked_dbt_get(&lkey, jnienv, key, keyop) != 0)
+	if (locked_dbt_get(&lkey, jnienv, dbenv, key, keyop) != 0)
 		goto out2;
-	if (locked_dbt_get(&ldata, jnienv, data, dataop) != 0)
+	if (locked_dbt_get(&ldata, jnienv, dbenv, data, dataop) != 0)
 		goto out1;
 	for (retry = 0; retry < 3; retry++) {
-		err = db->get(db, dbtxnid, &lkey.javainfo->dbt, &ldata.javainfo->dbt, flags);
+		err = db->get(db,
+		    dbtxnid, &lkey.javainfo->dbt, &ldata.javainfo->dbt, flags);
 
-		/* If we failed due to lack of memory in our DBT arrays,
+		/*
+		 * If we failed due to lack of memory in our DBT arrays,
 		 * retry.
 		 */
 		if (err != ENOMEM)
 			break;
-		if (!locked_dbt_realloc(&lkey, jnienv) &&
-		    !locked_dbt_realloc(&ldata, jnienv))
+		if (!locked_dbt_realloc(&lkey, jnienv, dbenv) &&
+		    !locked_dbt_realloc(&ldata, jnienv, dbenv))
 			break;
 	}
  out1:
-	locked_dbt_put(&ldata, jnienv);
+	locked_dbt_put(&ldata, jnienv, dbenv);
  out2:
-	locked_dbt_put(&lkey, jnienv);
+	locked_dbt_put(&lkey, jnienv, dbenv);
  out3:
-	if (err != 0 && err != DB_NOTFOUND) {
+	if (!DB_RETOK_DBGET(err)) {
 		if (verify_dbt(jnienv, err, &lkey) &&
 		    verify_dbt(jnienv, err, &ldata))
 			verify_return(jnienv, err, 0);
@@ -359,7 +388,7 @@ JNIEXPORT void JNICALL Java_com_sleepycat_db_Db_hash_1changed
 	if (!verify_non_null(jnienv, db))
 		return;
 
-	dbinfo = (DB_JAVAINFO*)db->cj_internal;
+	dbinfo = (DB_JAVAINFO*)db->api_internal;
 	dbji_set_h_hash_object(dbinfo, jnienv, db, jhash);
 }
 
@@ -384,8 +413,9 @@ JNIEXPORT jobject JNICALL Java_com_sleepycat_db_Db_join
 	}
 
 	/* Convert the java array of Dbc's to a C array of DBC's. */
-	for (i=0; i<count; i++) {
-		jobject jobj = (*jnienv)->GetObjectArrayElement(jnienv, curslist, i);
+	for (i = 0; i < count; i++) {
+		jobject jobj =
+		    (*jnienv)->GetObjectArrayElement(jnienv, curslist, i);
 		if (jobj == 0) {
 			/*
 			 * An embedded null in the array is treated
@@ -405,33 +435,36 @@ JNIEXPORT jobject JNICALL Java_com_sleepycat_db_Db_join
 
 	err = db->join(db, newlist, &dbc, flags);
 	verify_return(jnienv, err, 0);
-	__os_free(db->dbenv, newlist, size);
+	__os_free(db->dbenv, newlist);
 
 	return (get_Dbc(jnienv, dbc));
 }
 
 JNIEXPORT void JNICALL Java_com_sleepycat_db_Db_key_1range
-  (JNIEnv *jnienv, /*Db*/ jobject jthis, /*DbTxn*/ jobject jtxn,
+  (JNIEnv *jnienv, /*Db*/ jobject jthis, /*DbTxn*/ jobject txnid,
    /*Dbt*/ jobject jkey, jobject /*DbKeyRange*/ range, jint flags)
 {
 	int err;
-	DB *db = get_DB(jnienv, jthis);
-	DB_TXN *txn = get_DB_TXN(jnienv, jtxn);
+	DB *db;
+	DB_TXN *dbtxnid;
 	LOCKED_DBT lkey;
 	DB_KEY_RANGE result;
 	jfieldID fid;
 	jclass krclass;
 
+	db = get_DB(jnienv, jthis);
+	dbtxnid = get_DB_TXN(jnienv, txnid);
 	if (!verify_non_null(jnienv, db))
 		return;
 	if (!verify_non_null(jnienv, range))
 		return;
-	if (locked_dbt_get(&lkey, jnienv, jkey, inOp) != 0)
+	if (locked_dbt_get(&lkey, jnienv, db->dbenv, jkey, inOp) != 0)
 		goto out;
-	err = db->key_range(db, txn, &lkey.javainfo->dbt, &result, flags);
+	err = db->key_range(db, dbtxnid, &lkey.javainfo->dbt, &result, flags);
 	if (verify_return(jnienv, err, 0)) {
 		/* fill in the values of the DbKeyRange structure */
-		krclass = get_class(jnienv, "DbKeyRange");
+		if ((krclass = get_class(jnienv, "DbKeyRange")) == NULL)
+			return;	/* An exception has been posted. */
 		fid = (*jnienv)->GetFieldID(jnienv, krclass, "less", "D");
 		(*jnienv)->SetDoubleField(jnienv, range, fid, result.less);
 		fid = (*jnienv)->GetFieldID(jnienv, krclass, "equal", "D");
@@ -440,7 +473,7 @@ JNIEXPORT void JNICALL Java_com_sleepycat_db_Db_key_1range
 		(*jnienv)->SetDoubleField(jnienv, range, fid, result.greater);
 	}
  out:
-	locked_dbt_put(&lkey, jnienv);
+	locked_dbt_put(&lkey, jnienv, db->dbenv);
 }
 
 JNIEXPORT jint JNICALL Java_com_sleepycat_db_Db_pget
@@ -449,6 +482,7 @@ JNIEXPORT jint JNICALL Java_com_sleepycat_db_Db_pget
 {
 	int err, op_flags, retry;
 	DB *db;
+	DB_ENV *dbenv;
 	OpKind keyop, rkeyop, dataop;
 	DB_TXN *dbtxnid;
 	LOCKED_DBT lkey, lrkey, ldata;
@@ -457,6 +491,7 @@ JNIEXPORT jint JNICALL Java_com_sleepycat_db_Db_pget
 	db = get_DB(jnienv, jthis);
 	if (!verify_non_null(jnienv, db))
 		goto out4;
+	dbenv = db->dbenv;
 
 	/* Depending on flags, the key may be input/output. */
 	keyop = inOp;
@@ -474,34 +509,35 @@ JNIEXPORT jint JNICALL Java_com_sleepycat_db_Db_pget
 
 	dbtxnid = get_DB_TXN(jnienv, txnid);
 
-	if (locked_dbt_get(&lkey, jnienv, key, keyop) != 0)
+	if (locked_dbt_get(&lkey, jnienv, dbenv, key, keyop) != 0)
 		goto out3;
-	if (locked_dbt_get(&lrkey, jnienv, rkey, rkeyop) != 0)
+	if (locked_dbt_get(&lrkey, jnienv, dbenv, rkey, rkeyop) != 0)
 		goto out2;
-	if (locked_dbt_get(&ldata, jnienv, data, dataop) != 0)
+	if (locked_dbt_get(&ldata, jnienv, dbenv, data, dataop) != 0)
 		goto out1;
 	for (retry = 0; retry < 3; retry++) {
-		err = db->pget(db, dbtxnid, &lkey.javainfo->dbt, &lrkey.javainfo->dbt,
-			       &ldata.javainfo->dbt, flags);
+		err = db->pget(db, dbtxnid, &lkey.javainfo->dbt,
+		    &lrkey.javainfo->dbt, &ldata.javainfo->dbt, flags);
 
-		/* If we failed due to lack of memory in our DBT arrays,
+		/*
+		 * If we failed due to lack of memory in our DBT arrays,
 		 * retry.
 		 */
 		if (err != ENOMEM)
 			break;
-		if (!locked_dbt_realloc(&lkey, jnienv) &&
-		    !locked_dbt_realloc(&lrkey, jnienv) &&
-		    !locked_dbt_realloc(&ldata, jnienv))
+		if (!locked_dbt_realloc(&lkey, jnienv, dbenv) &&
+		    !locked_dbt_realloc(&lrkey, jnienv, dbenv) &&
+		    !locked_dbt_realloc(&ldata, jnienv, dbenv))
 			break;
 	}
  out1:
-	locked_dbt_put(&ldata, jnienv);
+	locked_dbt_put(&ldata, jnienv, dbenv);
  out2:
-	locked_dbt_put(&lrkey, jnienv);
+	locked_dbt_put(&lrkey, jnienv, dbenv);
  out3:
-	locked_dbt_put(&lkey, jnienv);
+	locked_dbt_put(&lkey, jnienv, dbenv);
  out4:
-	if (err != 0 && err != DB_NOTFOUND) {
+	if (!DB_RETOK_DBGET(err)) {
 		if (verify_dbt(jnienv, err, &lkey) &&
 		    verify_dbt(jnienv, err, &lrkey) &&
 		    verify_dbt(jnienv, err, &ldata))
@@ -516,6 +552,7 @@ JNIEXPORT jint JNICALL Java_com_sleepycat_db_Db_put
 {
 	int err;
 	DB *db;
+	DB_ENV *dbenv;
 	DB_TXN *dbtxnid;
 	LOCKED_DBT lkey, ldata;
 	OpKind keyop;
@@ -525,6 +562,7 @@ JNIEXPORT jint JNICALL Java_com_sleepycat_db_Db_put
 	dbtxnid = get_DB_TXN(jnienv, txnid);
 	if (!verify_non_null(jnienv, db))
 		return (0);   /* error will be thrown, retval doesn't matter */
+	dbenv = db->dbenv;
 
 	/*
 	 * For DB_APPEND, the key may be output-only;  for all other flags,
@@ -535,27 +573,61 @@ JNIEXPORT jint JNICALL Java_com_sleepycat_db_Db_put
 	else
 		keyop = inOp;
 
-	if (locked_dbt_get(&lkey, jnienv, key, keyop) != 0)
+	if (locked_dbt_get(&lkey, jnienv, dbenv, key, keyop) != 0)
 		goto out2;
-	if (locked_dbt_get(&ldata, jnienv, data, inOp) != 0)
+	if (locked_dbt_get(&ldata, jnienv, dbenv, data, inOp) != 0)
 		goto out1;
 
 	if (!verify_non_null(jnienv, db))
 		goto out1;
-	err = db->put(db, dbtxnid, &lkey.javainfo->dbt, &ldata.javainfo->dbt, flags);
-	if (err != DB_KEYEXIST) {
+
+	err = db->put(db,
+	    dbtxnid, &lkey.javainfo->dbt, &ldata.javainfo->dbt, flags);
+	if (!DB_RETOK_DBPUT(err))
 		verify_return(jnienv, err, 0);
-	}
+
  out1:
-	locked_dbt_put(&ldata, jnienv);
+	locked_dbt_put(&ldata, jnienv, dbenv);
  out2:
-	locked_dbt_put(&lkey, jnienv);
+	locked_dbt_put(&lkey, jnienv, dbenv);
 	return (err);
 }
 
+JNIEXPORT void JNICALL Java_com_sleepycat_db_Db__1remove
+  (JNIEnv *jnienv, /*Db*/ jobject jthis,
+   jstring file, jstring database, jint flags)
+{
+	int err;
+	DB *db;
+	DB_JAVAINFO *dbinfo;
+	LOCKED_STRING ls_file;
+	LOCKED_STRING ls_database;
+
+	db = get_DB(jnienv, jthis);
+	dbinfo = get_DB_JAVAINFO(jnienv, jthis);
+
+	if (!verify_non_null(jnienv, db))
+		return;
+	if (locked_string_get(&ls_file, jnienv, file) != 0)
+		goto out2;
+	if (locked_string_get(&ls_database, jnienv, database) != 0)
+		goto out1;
+	err = db->remove(db, ls_file.string, ls_database.string, flags);
+
+	set_private_dbobj(jnienv, name_DB, jthis, 0);
+	verify_return(jnienv, err, EXCEPTION_FILE_NOT_FOUND);
+
+ out1:
+	locked_string_put(&ls_database, jnienv);
+ out2:
+	locked_string_put(&ls_file, jnienv);
+
+	dbji_dealloc(dbinfo, jnienv);
+}
+
 JNIEXPORT void JNICALL Java_com_sleepycat_db_Db__1rename
-  (JNIEnv *jnienv, /*Db*/ jobject jthis, jstring file,
-   jstring database, jstring newname, jint flags)
+  (JNIEnv *jnienv, /*Db*/ jobject jthis,
+   jstring file, jstring database, jstring newname, jint flags)
 {
 	int err;
 	DB *db;
@@ -591,66 +663,13 @@ JNIEXPORT void JNICALL Java_com_sleepycat_db_Db__1rename
 	dbji_dealloc(dbinfo, jnienv);
 }
 
-JNIEXPORT void JNICALL Java_com_sleepycat_db_Db__1remove
-  (JNIEnv *jnienv, /*Db*/ jobject jthis, jstring file,
-   jstring database, jint flags)
-{
-	int err;
-	DB *db;
-	DB_JAVAINFO *dbinfo;
-	LOCKED_STRING ls_file;
-	LOCKED_STRING ls_database;
-
-	db = get_DB(jnienv, jthis);
-	dbinfo = get_DB_JAVAINFO(jnienv, jthis);
-
-	if (!verify_non_null(jnienv, db))
-		return;
-	if (locked_string_get(&ls_file, jnienv, file) != 0)
-		goto out2;
-	if (locked_string_get(&ls_database, jnienv, database) != 0)
-		goto out1;
-	err = db->remove(db, ls_file.string, ls_database.string, flags);
-
-	set_private_dbobj(jnienv, name_DB, jthis, 0);
-	verify_return(jnienv, err, EXCEPTION_FILE_NOT_FOUND);
-
- out1:
-	locked_string_put(&ls_database, jnienv);
- out2:
-	locked_string_put(&ls_file, jnienv);
-
-	dbji_dealloc(dbinfo, jnienv);
-}
-
-JNIEXPORT void JNICALL
-  Java_com_sleepycat_db_Db_set_1pagesize
-  (JNIEnv *jnienv, /*Db*/ jobject jthis, jlong value)
-{
-	int err;
-	DB *db;
-
-	db = get_DB(jnienv, jthis);
-	if (verify_non_null(jnienv, db)) {
-		err = db->set_pagesize(db, (u_int32_t)value);
-		verify_return(jnienv, err, 0);
-	}
-}
-
-JNIEXPORT void JNICALL
-  Java_com_sleepycat_db_Db_set_1cachesize
-  (JNIEnv *jnienv, /*Db*/ jobject jthis, jint gbytes, jint bytes,
-   jint ncaches)
-{
-	int err;
-	DB *db;
-
-	db = get_DB(jnienv, jthis);
-	if (verify_non_null(jnienv, db)) {
-		err = db->set_cachesize(db, gbytes, bytes, ncaches);
-		verify_return(jnienv, err, 0);
-	}
-}
+JAVADB_METHOD(Db_set_1pagesize, (JAVADB_ARGS, jlong pagesize), DB,
+    set_pagesize, (c_this, (u_int32_t)pagesize))
+JAVADB_METHOD(Db_set_1cachesize,
+    (JAVADB_ARGS, jint gbytes, jint bytes, jint ncaches), DB,
+    set_cachesize, (c_this, gbytes, bytes, ncaches))
+JAVADB_METHOD(Db_set_1cache_1priority, (JAVADB_ARGS, jint priority), DB,
+    set_cache_priority, (c_this, (DB_CACHE_PRIORITY)priority))
 
 JNIEXPORT void JNICALL
   Java_com_sleepycat_db_Db_set_1re_1source
@@ -664,7 +683,8 @@ JNIEXPORT void JNICALL
 
 		/* XXX does the string from get_c_string ever get freed? */
 		if (re_source != NULL)
-			err = db->set_re_source(db, get_c_string(jnienv, re_source));
+			err = db->set_re_source(db,
+			    get_c_string(jnienv, re_source));
 		else
 			err = db->set_re_source(db, 0);
 
@@ -703,50 +723,11 @@ JNIEXPORT jobject JNICALL Java_com_sleepycat_db_Db_stat
 			bytesize = sizeof(DB_BTREE_STAT);
 			retval = create_default_object(jnienv,
 						       name_DB_BTREE_STAT);
-			dbclass = get_class(jnienv, name_DB_BTREE_STAT);
-
-			/* Set the individual fields */
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_magic", bstp->bt_magic);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_version", bstp->bt_version);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_metaflags", bstp->bt_metaflags);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_nkeys", bstp->bt_nkeys);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_ndata", bstp->bt_ndata);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_pagesize", bstp->bt_pagesize);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_maxkey", bstp->bt_maxkey);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_minkey", bstp->bt_minkey);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_re_len", bstp->bt_re_len);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_re_pad", bstp->bt_re_pad);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_levels", bstp->bt_levels);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_int_pg", bstp->bt_int_pg);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_leaf_pg", bstp->bt_leaf_pg);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_dup_pg", bstp->bt_dup_pg);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_over_pg", bstp->bt_over_pg);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_free", bstp->bt_free);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_int_pgfree", bstp->bt_int_pgfree);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_leaf_pgfree", bstp->bt_leaf_pgfree);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_dup_pgfree", bstp->bt_dup_pgfree);
-			set_int_field(jnienv, dbclass, retval,
-				      "bt_over_pgfree", bstp->bt_over_pgfree);
+			if ((dbclass =
+			    get_class(jnienv, name_DB_BTREE_STAT)) == NULL)
+				break;	/* An exception has been posted. */
 
+			__jv_fill_bt_stat(jnienv, dbclass, retval, bstp);
 			break;
 
 			/* Hash stat structure */
@@ -755,44 +736,11 @@ JNIEXPORT jobject JNICALL Java_com_sleepycat_db_Db_stat
 			bytesize = sizeof(DB_HASH_STAT);
 			retval = create_default_object(jnienv,
 						       name_DB_HASH_STAT);
-			dbclass = get_class(jnienv, name_DB_HASH_STAT);
-
-			/* Set the individual fields */
-			set_int_field(jnienv, dbclass, retval,
-				      "hash_magic", hstp->hash_magic);
-			set_int_field(jnienv, dbclass, retval,
-				      "hash_version", hstp->hash_version);
-			set_int_field(jnienv, dbclass, retval,
-				      "hash_metaflags", hstp->hash_metaflags);
-			set_int_field(jnienv, dbclass, retval,
-				      "hash_nkeys", hstp->hash_nkeys);
-			set_int_field(jnienv, dbclass, retval,
-				      "hash_ndata", hstp->hash_ndata);
-			set_int_field(jnienv, dbclass, retval,
-				      "hash_pagesize", hstp->hash_pagesize);
-			set_int_field(jnienv, dbclass, retval,
-				      "hash_nelem", hstp->hash_nelem);
-			set_int_field(jnienv, dbclass, retval,
-				      "hash_ffactor", hstp->hash_ffactor);
-			set_int_field(jnienv, dbclass, retval,
-				      "hash_buckets", hstp->hash_buckets);
-			set_int_field(jnienv, dbclass, retval,
-				      "hash_free", hstp->hash_free);
-			set_int_field(jnienv, dbclass, retval,
-				      "hash_bfree", hstp->hash_bfree);
-			set_int_field(jnienv, dbclass, retval,
-				      "hash_bigpages", hstp->hash_bigpages);
-			set_int_field(jnienv, dbclass, retval,
-				      "hash_big_bfree", hstp->hash_big_bfree);
-			set_int_field(jnienv, dbclass, retval,
-				      "hash_overflows", hstp->hash_overflows);
-			set_int_field(jnienv, dbclass, retval,
-				      "hash_ovfl_free", hstp->hash_ovfl_free);
-			set_int_field(jnienv, dbclass, retval,
-				      "hash_dup", hstp->hash_dup);
-			set_int_field(jnienv, dbclass, retval,
-				      "hash_dup_free", hstp->hash_dup_free);
+			if ((dbclass =
+			    get_class(jnienv, name_DB_HASH_STAT)) == NULL)
+				break;	/* An exception has been posted. */
 
+			__jv_fill_h_stat(jnienv, dbclass, retval, hstp);
 			break;
 
 		case DB_QUEUE:
@@ -800,35 +748,11 @@ JNIEXPORT jobject JNICALL Java_com_sleepycat_db_Db_stat
 			bytesize = sizeof(DB_QUEUE_STAT);
 			retval = create_default_object(jnienv,
 						       name_DB_QUEUE_STAT);
-			dbclass = get_class(jnienv, name_DB_QUEUE_STAT);
+			if ((dbclass =
+			    get_class(jnienv, name_DB_QUEUE_STAT)) == NULL)
+				break;	/* An exception has been posted. */
 
-			/* Set the individual fields */
-			set_int_field(jnienv, dbclass, retval,
-				      "qs_magic", qstp->qs_magic);
-			set_int_field(jnienv, dbclass, retval,
-				      "qs_version", qstp->qs_version);
-			set_int_field(jnienv, dbclass, retval,
-				      "qs_metaflags", qstp->qs_metaflags);
-			set_int_field(jnienv, dbclass, retval,
-				      "qs_nkeys", qstp->qs_nkeys);
-			set_int_field(jnienv, dbclass, retval,
-				      "qs_ndata", qstp->qs_ndata);
-			set_int_field(jnienv, dbclass, retval,
-				      "qs_pagesize", qstp->qs_pagesize);
-			set_int_field(jnienv, dbclass, retval,
-				      "qs_extentsize", qstp->qs_extentsize);
-			set_int_field(jnienv, dbclass, retval,
-				      "qs_pages", qstp->qs_pages);
-			set_int_field(jnienv, dbclass, retval,
-				      "qs_re_len", qstp->qs_re_len);
-			set_int_field(jnienv, dbclass, retval,
-				      "qs_re_pad", qstp->qs_re_pad);
-			set_int_field(jnienv, dbclass, retval,
-				      "qs_pgfree", qstp->qs_pgfree);
-			set_int_field(jnienv, dbclass, retval,
-				      "qs_first_recno", qstp->qs_first_recno);
-			set_int_field(jnienv, dbclass, retval,
-				      "qs_cur_recno", qstp->qs_cur_recno);
+			__jv_fill_qam_stat(jnienv, dbclass, retval, qstp);
 			break;
 
 			/* That's all the database types we're aware of! */
@@ -841,24 +765,13 @@ JNIEXPORT jobject JNICALL Java_com_sleepycat_db_Db_stat
 			break;
 		}
 		if (bytesize != 0)
-			__os_ufree(db->dbenv, statp, bytesize);
+			__os_ufree(db->dbenv, statp);
 	}
 	return (retval);
 }
 
-JNIEXPORT jint JNICALL Java_com_sleepycat_db_Db_sync
-  (JNIEnv *jnienv, /*Db*/ jobject jthis, jint flags)
-{
-	int err;
-	DB *db = get_DB(jnienv, jthis);
-
-	if (!verify_non_null(jnienv, db))
-		return (0);
-	err = db->sync(db, flags);
-	if (err != DB_INCOMPLETE)
-		verify_return(jnienv, err, 0);
-	return (err);
-}
+JAVADB_METHOD(Db_sync, (JAVADB_ARGS, jint flags), DB,
+    sync, (c_this, flags))
 
 JNIEXPORT jboolean JNICALL Java_com_sleepycat_db_Db_get_1byteswapped
   (JNIEnv *jnienv, /*Db*/ jobject jthis)
@@ -900,24 +813,27 @@ JNIEXPORT jint JNICALL Java_com_sleepycat_db_Db_get_1type
 }
 
 JNIEXPORT void JNICALL Java_com_sleepycat_db_Db__1open
-  (JNIEnv *jnienv, /*Db*/ jobject jthis, jstring file, jstring database,
-   jint type, jint flags, jint mode)
+  (JNIEnv *jnienv, /*Db*/ jobject jthis, /*DbTxn*/ jobject txnid,
+   jstring file, jstring database, jint type, jint flags, jint mode)
 {
 	int err;
 	DB *db;
+	DB_TXN *dbtxnid;
 	LOCKED_STRING ls_file;
 	LOCKED_STRING ls_database;
 
-	/* Java is assumed to be threaded. */
+	/* Java is assumed to be threaded */
 	flags |= DB_THREAD;
 
 	db = get_DB(jnienv, jthis);
+
+	dbtxnid = get_DB_TXN(jnienv, txnid);
 	if (locked_string_get(&ls_file, jnienv, file) != 0)
 		goto out2;
 	if (locked_string_get(&ls_database, jnienv, database) != 0)
 		goto out1;
 	if (verify_non_null(jnienv, db)) {
-		err = db->open(db, ls_file.string, ls_database.string,
+		err = db->open(db, dbtxnid, ls_file.string, ls_database.string,
 			       (DBTYPE)type, flags, mode);
 		verify_return(jnienv, err, EXCEPTION_FILE_NOT_FOUND);
 	}
@@ -978,9 +894,13 @@ static int java_verify_callback(void *handle, const void *str_arg)
 		vc->nbytes = len;
 		vc->bytes = (*jnienv)->NewByteArray(jnienv, len);
 	}
-	(*jnienv)->SetByteArrayRegion(jnienv, vc->bytes, 0, len, (jbyte*)str);
-	(*jnienv)->CallVoidMethod(jnienv, vc->streamobj,
-				  vc->writemid, vc->bytes, 0, len-1);
+
+	if (vc->bytes != NULL) {
+		(*jnienv)->SetByteArrayRegion(jnienv, vc->bytes, 0, len,
+		    (jbyte*)str);
+		(*jnienv)->CallVoidMethod(jnienv, vc->streamobj,
+		    vc->writemid, vc->bytes, 0, len-1);
+	}
 
 	if ((*jnienv)->ExceptionOccurred(jnienv) != NULL)
 		return (EIO);
@@ -1011,7 +931,8 @@ JNIEXPORT void JNICALL Java_com_sleepycat_db_Db_verify
 	vcs.env = jnienv;
 	vcs.streamobj = stream;
 	vcs.nbytes = 100;
-	vcs.bytes = (*jnienv)->NewByteArray(jnienv, vcs.nbytes);
+	if ((vcs.bytes = (*jnienv)->NewByteArray(jnienv, vcs.nbytes)) == NULL)
+		goto out1;
 
 	/* get the method ID for OutputStream.write(byte[], int, int); */
 	streamclass = (*jnienv)->FindClass(jnienv, "java/io/OutputStream");
@@ -1040,7 +961,8 @@ JNIEXPORT void JNICALL Java_com_sleepycat_db_Db__1finalize
 	db = get_DB(jnienv, jthis);
 	DB_ASSERT(dbinfo != NULL);
 
-	/* Note: We can never be sure if the underlying DB is attached to
+	/*
+	 * Note: We can never be sure if the underlying DB is attached to
 	 * a DB_ENV that was already closed.  Sure, that's a user error,
 	 * but it shouldn't crash the VM.  Therefore, we cannot just
 	 * automatically close if the handle indicates we are not yet
