@@ -1,13 +1,13 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2001
+ * Copyright (c) 1997-2002
  *	Sleepycat Software.  All rights reserved.
  */
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "Id: java_DbLsn.c,v 11.9 2001/04/06 00:56:01 dda Exp ";
+static const char revid[] = "Id: java_DbLsn.c,v 11.12 2002/05/07 16:12:41 dda Exp ";
 #endif /* not lint */
 
 #include <jni.h>
@@ -22,7 +22,8 @@ static const char revid[] = "Id: java_DbLsn.c,v 11.9 2001/04/06 00:56:01 dda Exp
 JNIEXPORT void JNICALL Java_com_sleepycat_db_DbLsn_init_1lsn
   (JNIEnv *jnienv, /*DbLsn*/ jobject jthis)
 {
-	/* Note: the DB_LSN object stored in the private_dbobj_
+	/*
+	 * Note: the DB_LSN object stored in the private_dbobj_
 	 * is allocated in get_DbLsn() or get_DB_LSN().
 	 */
 
@@ -37,6 +38,6 @@ JNIEXPORT void JNICALL Java_com_sleepycat_db_DbLsn_finalize
 
 	dblsn = get_DB_LSN(jnienv, jthis);
 	if (dblsn) {
-		(void)__os_free(NULL, dblsn, sizeof(DB_LSN));
+		(void)__os_free(NULL, dblsn);
 	}
 }
