@@ -1,13 +1,13 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2001
+ * Copyright (c) 1996-2002
  *	Sleepycat Software.  All rights reserved.
  */
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "Id: mp_bh.c,v 11.45 2001/07/26 19:53:31 bostic Exp ";
+static const char revid[] = "Id: mp_bh.c,v 11.71 2002/09/04 19:06:45 margo Exp ";
 #endif /* not lint */
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -18,43 +18,41 @@ static const char revid[] = "Id: mp_bh.c,v 11.45 2001/07/26 19:53:31 bostic Exp
 #endif
 
 #include "db_int.h"
-#include "db_shash.h"
-#include "mp.h"
-#include "log.h"
-#include "db_page.h"
+#include "dbinc/db_shash.h"
+#include "dbinc/mp.h"
+#include "dbinc/log.h"
+#include "dbinc/db_page.h"
 
+static int __memp_pgwrite
+	   __P((DB_MPOOL *, DB_MPOOLFILE *, DB_MPOOL_HASH *, BH *));
 static int __memp_upgrade __P((DB_MPOOL *, DB_MPOOLFILE *, MPOOLFILE *));
 
 /*
  * __memp_bhwrite --
- *	Write the page associated with a given bucket header.
+ *	Write the page associated with a given buffer header.
  *
- * PUBLIC: int __memp_bhwrite
- * PUBLIC:     __P((DB_MPOOL *, MPOOLFILE *, BH *, int, int *, int *));
+ * PUBLIC: int __memp_bhwrite __P((DB_MPOOL *,
+ * PUBLIC:      DB_MPOOL_HASH *, MPOOLFILE *, BH *, int));
  */
 int
-__memp_bhwrite(dbmp, mfp, bhp, open_extents, restartp, wrotep)
+__memp_bhwrite(dbmp, hp, mfp, bhp, open_extents)
 	DB_MPOOL *dbmp;
+	DB_MPOOL_HASH *hp;
 	MPOOLFILE *mfp;
 	BH *bhp;
-	int open_extents, *restartp, *wrotep;
+	int open_extents;
 {
 	DB_ENV *dbenv;
 	DB_MPOOLFILE *dbmfp;
 	DB_MPREG *mpreg;
-	int incremented, ret;
+	int local_open, incremented, ret;
 
 	dbenv = dbmp->dbenv;
-
-	if (restartp != NULL)
-		*restartp = 0;
-	if (wrotep != NULL)
-		*wrotep = 0;
-	incremented = 0;
+	local_open = incremented = 0;
 
 	/*
-	 * If the file has been removed or is a closed temporary file, Jump
-	 * right ahead and pretend that we've found the file we want-- the
+	 * If the file has been removed or is a closed temporary file, jump
+	 * right ahead and pretend that we've found the file we want -- the
 	 * page-write function knows how to handle the fact that we don't have
 	 * (or need!) any real file descriptor information.
 	 */
@@ -74,23 +72,33 @@ __memp_bhwrite(dbmp, mfp, bhp, open_extents, restartp, wrotep)
 	    dbmfp != NULL; dbmfp = TAILQ_NEXT(dbmfp, q))
 		if (dbmfp->mfp == mfp) {
 			if (F_ISSET(dbmfp, MP_READONLY) &&
-			    __memp_upgrade(dbmp, dbmfp, mfp)) {
+			    !F_ISSET(dbmfp, MP_UPGRADE) &&
+			    (F_ISSET(dbmfp, MP_UPGRADE_FAIL) ||
+			    __memp_upgrade(dbmp, dbmfp, mfp))) {
 				MUTEX_THREAD_UNLOCK(dbenv, dbmp->mutexp);
-				return (0);
+				return (EPERM);
 			}
 
 			/*
 			 * Increment the reference count -- see the comment in
-			 * memp_fclose().
+			 * __memp_fclose_int().
 			 */
 			++dbmfp->ref;
 			incremented = 1;
 			break;
 		}
 	MUTEX_THREAD_UNLOCK(dbenv, dbmp->mutexp);
+
 	if (dbmfp != NULL)
 		goto found;
 
+	/*
+	 * !!!
+	 * It's the caller's choice if we're going to open extent files.
+	 */
+	if (!open_extents && F_ISSET(mfp, MP_EXTENT))
+		return (EPERM);
+
 	/*
 	 * !!!
 	 * Don't try to attach to temporary files.  There are two problems in
@@ -107,15 +115,12 @@ __memp_bhwrite(dbmp, mfp, bhp, open_extents, restartp, wrotep)
 	 * with resource starvation, and the memp_trickle thread couldn't do
 	 * anything about it.  That's a pretty unlikely scenario, though.
 	 *
-	 * Note that we should never get here when the temporary file
-	 * in question has already been closed in another process, in which
-	 * case it should be marked MP_DEADFILE.
+	 * Note we should never get here when the temporary file in question
+	 * has already been closed in another process, in which case it should
+	 * be marked MP_DEADFILE.
 	 */
-	if (F_ISSET(mfp, MP_TEMP)
-	    || (F_ISSET(mfp, MP_EXTENT) && !open_extents)) {
-		DB_ASSERT(!F_ISSET(mfp, MP_DEADFILE));
-		return (0);
-	}
+	if (F_ISSET(mfp, MP_TEMP))
+		return (EPERM);
 
 	/*
 	 * It's not a page from a file we've opened.  If the file requires
@@ -131,7 +136,7 @@ __memp_bhwrite(dbmp, mfp, bhp, open_extents, restartp, wrotep)
 				break;
 		MUTEX_THREAD_UNLOCK(dbenv, dbmp->mutexp);
 		if (mpreg == NULL)
-			return (0);
+			return (EPERM);
 	}
 
 	/*
@@ -142,25 +147,24 @@ __memp_bhwrite(dbmp, mfp, bhp, open_extents, restartp, wrotep)
 	 * There's no negative cache, so we may repeatedly try and open files
 	 * that we have previously tried (and failed) to open.
 	 */
-	if (dbenv->memp_fcreate(dbenv, &dbmfp, 0) != 0)
-		return (0);
-	if (__memp_fopen_int(dbmfp, mfp,
+	if ((ret = dbenv->memp_fcreate(dbenv, &dbmfp, 0)) != 0)
+		return (ret);
+	if ((ret = __memp_fopen_int(dbmfp, mfp,
 	    R_ADDR(dbmp->reginfo, mfp->path_off),
-	    0, 0, mfp->stat.st_pagesize, 0) != 0) {
+	    0, 0, mfp->stat.st_pagesize)) != 0) {
 		(void)dbmfp->close(dbmfp, 0);
-		return (0);
+		return (ret);
 	}
-	F_SET(dbmfp, MP_FLUSH);
-	if (F_ISSET(mfp, MP_EXTENT))
-		dbmp->extents = 1;
+	local_open = 1;
 
-found:	ret = __memp_pgwrite(dbmp, dbmfp, bhp, restartp, wrotep);
+found:	ret = __memp_pgwrite(dbmp, dbmfp, hp, bhp);
 
-	if (incremented) {
-		MUTEX_THREAD_LOCK(dbenv, dbmp->mutexp);
+	MUTEX_THREAD_LOCK(dbenv, dbmp->mutexp);
+	if (incremented)
 		--dbmfp->ref;
-		MUTEX_THREAD_UNLOCK(dbenv, dbmp->mutexp);
-	}
+	else if (local_open)
+		F_SET(dbmfp, MP_FLUSH);
+	MUTEX_THREAD_UNLOCK(dbenv, dbmp->mutexp);
 
 	return (ret);
 }
@@ -169,11 +173,12 @@ found:	ret = __memp_pgwrite(dbmp, dbmfp, bhp, restartp, wrotep);
  * __memp_pgread --
  *	Read a page from a file.
  *
- * PUBLIC: int __memp_pgread __P((DB_MPOOLFILE *, BH *, int));
+ * PUBLIC: int __memp_pgread __P((DB_MPOOLFILE *, DB_MUTEX *, BH *, int));
  */
 int
-__memp_pgread(dbmfp, bhp, can_create)
+__memp_pgread(dbmfp, mutexp, bhp, can_create)
 	DB_MPOOLFILE *dbmfp;
+	DB_MUTEX *mutexp;
 	BH *bhp;
 	int can_create;
 {
@@ -181,19 +186,21 @@ __memp_pgread(dbmfp, bhp, can_create)
 	DB_ENV *dbenv;
 	DB_MPOOL *dbmp;
 	MPOOLFILE *mfp;
-	size_t len, pagesize;
-	size_t nr;
-	int created, ret;
+	size_t len, nr, pagesize;
+	int ret;
 
 	dbmp = dbmfp->dbmp;
 	dbenv = dbmp->dbenv;
 	mfp = dbmfp->mfp;
 	pagesize = mfp->stat.st_pagesize;
-	created = 0;
 
+	/* We should never be called with a dirty or a locked buffer. */
+	DB_ASSERT(!F_ISSET(bhp, BH_DIRTY | BH_DIRTY_CREATE | BH_LOCKED));
+
+	/* Lock the buffer and swap the hash bucket lock for the buffer lock. */
 	F_SET(bhp, BH_LOCKED | BH_TRASH);
-	MUTEX_LOCK(dbenv, &bhp->mutex, dbenv->lockfhp);
-	R_UNLOCK(dbenv, dbmp->reginfo);
+	MUTEX_LOCK(dbenv, &bhp->mutex);
+	MUTEX_UNLOCK(dbenv, mutexp);
 
 	/*
 	 * Temporary files may not yet have been created.  We don't create
@@ -208,125 +215,99 @@ __memp_pgread(dbmfp, bhp, can_create)
 		db_io.buf = bhp->buf;
 
 		/*
-		 * The page may not exist;  if it doesn't, nr may well be 0,
+		 * The page may not exist; if it doesn't, nr may well be 0,
 		 * but we expect the underlying OS calls not to return an
 		 * error code in this case.
 		 */
 		if ((ret = __os_io(dbenv, &db_io, DB_IO_READ, &nr)) != 0)
 			goto err;
-	} else
-		ret = 0;
+	}
 
 	if (nr < pagesize) {
-		if (can_create)
-			created = 1;
-		else {
-			/*
-			 * Don't output error messages for short reads.  In
-			 * particular, DB recovery processing may request pages
-			 * that have never been written to disk or for which
-			 * only some part have been written to disk, in which
-			 * case we won't find the page.  The caller must know
-			 * how to handle the error.
-			 */
+		/*
+		 * Don't output error messages for short reads.  In particular,
+		 * DB recovery processing may request pages never written to
+		 * disk or for which only some part have been written to disk,
+		 * in which case we won't find the page.  The caller must know
+		 * how to handle the error.
+		 */
+		if (can_create == 0) {
 			ret = DB_PAGE_NOTFOUND;
 			goto err;
 		}
+
+		/* Clear any bytes that need to be cleared. */
+		len = mfp->clear_len == 0 ? pagesize : mfp->clear_len;
+		memset(bhp->buf, 0, len);
+
+#if defined(DIAGNOSTIC) || defined(UMRW)
 		/*
-		 * Clear any bytes that need to be cleared -- if we did a short
-		 * read, we assume that a page was not completely written and
-		 * clear even the bytes that we read.  This is so our caller
-		 * isn't surprised (for example, if the first sector only of a
-		 * DB page was written, the LSN will indicate that the page was
-		 * updated, but the page contents will be wrong).  Support for
-		 * page checksums might make this unnecessary in the future --
-		 * I would prefer not to discard data potentially written by
-		 * the application, under any circumstances.
-		 *
 		 * If we're running in diagnostic mode, corrupt any bytes on
 		 * the page that are unknown quantities for the caller.
 		 */
-		len = mfp->clear_len == 0 ? pagesize : mfp->clear_len;
-		memset(bhp->buf, 0, len);
-#if defined(DIAGNOSTIC) || defined(UMRW)
 		if (len < pagesize)
 			memset(bhp->buf + len, CLEAR_BYTE, pagesize - len);
 #endif
-	}
+		++mfp->stat.st_page_create;
+	} else
+		++mfp->stat.st_page_in;
 
 	/* Call any pgin function. */
 	ret = mfp->ftype == 0 ? 0 : __memp_pg(dbmfp, bhp, 1);
 
-	/* Unlock the buffer and reacquire the region lock. */
+	/* Unlock the buffer and reacquire the hash bucket lock. */
 err:	MUTEX_UNLOCK(dbenv, &bhp->mutex);
-	R_LOCK(dbenv, dbmp->reginfo);
+	MUTEX_LOCK(dbenv, mutexp);
 
 	/*
 	 * If no errors occurred, the data is now valid, clear the BH_TRASH
 	 * flag; regardless, clear the lock bit and let other threads proceed.
 	 */
 	F_CLR(bhp, BH_LOCKED);
-	if (ret == 0) {
+	if (ret == 0)
 		F_CLR(bhp, BH_TRASH);
 
-		/* Update the statistics. */
-		if (created)
-			++mfp->stat.st_page_create;
-		else
-			++mfp->stat.st_page_in;
-	}
-
 	return (ret);
 }
 
 /*
  * __memp_pgwrite --
  *	Write a page to a file.
- *
- * PUBLIC: int __memp_pgwrite
- * PUBLIC:     __P((DB_MPOOL *, DB_MPOOLFILE *, BH *, int *, int *));
  */
-int
-__memp_pgwrite(dbmp, dbmfp, bhp, restartp, wrotep)
+static int
+__memp_pgwrite(dbmp, dbmfp, hp, bhp)
 	DB_MPOOL *dbmp;
 	DB_MPOOLFILE *dbmfp;
+	DB_MPOOL_HASH *hp;
 	BH *bhp;
-	int *restartp, *wrotep;
 {
 	DB_ENV *dbenv;
 	DB_IO db_io;
 	DB_LSN lsn;
-	MPOOL *c_mp, *mp;
 	MPOOLFILE *mfp;
 	size_t nw;
-	int callpgin, dosync, ret, syncfail;
-	const char *fail;
+	int callpgin, ret;
 
 	dbenv = dbmp->dbenv;
-	mp = dbmp->reginfo[0].primary;
 	mfp = dbmfp == NULL ? NULL : dbmfp->mfp;
-
-	if (restartp != NULL)
-		*restartp = 0;
-	if (wrotep != NULL)
-		*wrotep = 0;
-	callpgin = 0;
-
-	/* We should never be called with a clean or a locked buffer. */
-	DB_ASSERT(F_ISSET(bhp, BH_DIRTY));
-	DB_ASSERT(!F_ISSET(bhp, BH_LOCKED));
+	callpgin = ret = 0;
 
 	/*
-	 * Lock the buffer, set the I/O in progress flag, and discard the
-	 * region lock.
+	 * We should never be called with a clean or trash buffer.
+	 * The sync code does call us with already locked buffers.
 	 */
-	MUTEX_LOCK(dbenv, &bhp->mutex, dbenv->lockfhp);
-	F_SET(bhp, BH_LOCKED);
-	R_UNLOCK(dbenv, dbmp->reginfo);
+	DB_ASSERT(F_ISSET(bhp, BH_DIRTY));
+	DB_ASSERT(!F_ISSET(bhp, BH_TRASH));
 
-	/* Tell the caller that the region lock was discarded. */
-	if (restartp != NULL)
-		*restartp = 1;
+	/*
+	 * If we have not already traded the hash bucket lock for the buffer
+	 * lock, do so now.
+	 */
+	if (!F_ISSET(bhp, BH_LOCKED)) {
+		F_SET(bhp, BH_LOCKED);
+		MUTEX_LOCK(dbenv, &bhp->mutex);
+		MUTEX_UNLOCK(dbenv, &hp->hash_mutex);
+	}
 
 	/*
 	 * It's possible that the underlying file doesn't exist, either
@@ -342,13 +323,9 @@ __memp_pgwrite(dbmp, dbmfp, bhp, restartp, wrotep)
 
 	/*
 	 * If the page is in a file for which we have LSN information, we have
-	 * to ensure the appropriate log records are on disk.  If the page is
-	 * being written as part of a sync operation, the flush has been done
-	 * already, unless it was modified by the application *after* the sync
-	 * was scheduled.
+	 * to ensure the appropriate log records are on disk.
 	 */
-	if (LOGGING_ON(dbenv) && !IS_RECOVERING(dbenv) && mfp->lsn_off != -1 &&
-	    (!F_ISSET(bhp, BH_SYNC) || F_ISSET(bhp, BH_SYNC_LOGFLSH))) {
+	if (LOGGING_ON(dbenv) && mfp->lsn_off != -1) {
 		memcpy(&lsn, bhp->buf + mfp->lsn_off, sizeof(DB_LSN));
 		if ((ret = dbenv->log_flush(dbenv, &lsn)) != 0)
 			goto err;
@@ -361,7 +338,7 @@ __memp_pgwrite(dbmp, dbmfp, bhp, restartp, wrotep)
 	 * !!!
 	 * One special case.  There is a single field on the meta-data page,
 	 * the last-page-number-in-the-file field, for which we do not log
-	 * changes.  So, if the page was original created in a database that
+	 * changes.  If the page was originally created in a database that
 	 * didn't have logging turned on, we can see a page marked dirty but
 	 * for which no corresponding log record has been written.  However,
 	 * the only way that a page can be created for which there isn't a
@@ -381,7 +358,8 @@ __memp_pgwrite(dbmp, dbmfp, bhp, restartp, wrotep)
 
 		dblp = dbenv->lg_handle;
 		lp = dblp->reginfo.primary;
-		if (log_compare(&lp->s_lsn, &LSN(bhp->buf)) <= 0) {
+		if (!IS_NOT_LOGGED_LSN(LSN(bhp->buf)) &&
+		    log_compare(&lp->s_lsn, &LSN(bhp->buf)) <= 0) {
 			R_LOCK(dbenv, &dblp->reginfo);
 			DB_ASSERT(log_compare(&lp->s_lsn, &LSN(bhp->buf)) > 0);
 			R_UNLOCK(dbenv, &dblp->reginfo);
@@ -394,9 +372,7 @@ __memp_pgwrite(dbmp, dbmfp, bhp, restartp, wrotep)
 	 * that the contents of the buffer will need to be passed through pgin
 	 * before they are reused.
 	 */
-	if (mfp->ftype == 0)
-		ret = 0;
-	else {
+	if (mfp->ftype != 0) {
 		callpgin = 1;
 		if ((ret = __memp_pg(dbmfp, bhp, 0)) != 0)
 			goto err;
@@ -405,17 +381,16 @@ __memp_pgwrite(dbmp, dbmfp, bhp, restartp, wrotep)
 	/* Temporary files may not yet have been created. */
 	if (!F_ISSET(dbmfp->fhp, DB_FH_VALID)) {
 		MUTEX_THREAD_LOCK(dbenv, dbmp->mutexp);
-		if (!F_ISSET(dbmfp->fhp, DB_FH_VALID) &&
-		    ((ret = __db_appname(dbenv, DB_APP_TMP, NULL, NULL,
-		    DB_OSO_CREATE | DB_OSO_EXCL | DB_OSO_TEMP,
-		    dbmfp->fhp, NULL)) != 0 ||
-		    !F_ISSET(dbmfp->fhp, DB_FH_VALID))) {
-			MUTEX_THREAD_UNLOCK(dbenv, dbmp->mutexp);
+		ret = F_ISSET(dbmfp->fhp, DB_FH_VALID) ? 0 :
+		    __db_appname(dbenv, DB_APP_TMP, NULL,
+		    F_ISSET(dbenv, DB_ENV_DIRECT_DB) ? DB_OSO_DIRECT : 0,
+		    dbmfp->fhp, NULL);
+		MUTEX_THREAD_UNLOCK(dbenv, dbmp->mutexp);
+		if (ret != 0) {
 			__db_err(dbenv,
 			    "unable to create temporary backing file");
 			goto err;
 		}
-		MUTEX_THREAD_UNLOCK(dbenv, dbmp->mutexp);
 	}
 
 	/* Write the page. */
@@ -425,104 +400,44 @@ __memp_pgwrite(dbmp, dbmfp, bhp, restartp, wrotep)
 	db_io.pgno = bhp->pgno;
 	db_io.buf = bhp->buf;
 	if ((ret = __os_io(dbenv, &db_io, DB_IO_WRITE, &nw)) != 0) {
-		fail = "write";
-		goto syserr;
-	}
-	if (nw != mfp->stat.st_pagesize) {
-		ret = EIO;
-		fail = "write";
-		goto syserr;
+		__db_err(dbenv, "%s: write failed for page %lu",
+		    __memp_fn(dbmfp), (u_long)bhp->pgno);
+		goto err;
 	}
+	++mfp->stat.st_page_out;
 
+err:
 file_dead:
 	/*
 	 * !!!
 	 * Once we pass this point, dbmfp and mfp may be NULL, we may not have
 	 * a valid file reference.
 	 *
-	 * Unlock the buffer and reacquire the region lock.
+	 * Unlock the buffer and reacquire the hash lock.
 	 */
 	MUTEX_UNLOCK(dbenv, &bhp->mutex);
-	R_LOCK(dbenv, dbmp->reginfo);
+	MUTEX_LOCK(dbenv, &hp->hash_mutex);
 
 	/*
-	 * Clean up the flags based on a successful write.
-	 *
 	 * If we rewrote the page, it will need processing by the pgin
 	 * routine before reuse.
 	 */
 	if (callpgin)
 		F_SET(bhp, BH_CALLPGIN);
-	F_CLR(bhp, BH_DIRTY | BH_DIRTY_CREATE | BH_LOCKED);
 
 	/*
-	 * If we write a buffer for which a checkpoint is waiting, update
-	 * the count of pending buffers (both in the mpool as a whole and
-	 * for this file).  If the count for this file goes to zero, set a
-	 * flag so we flush the writes.
+	 * Update the hash bucket statistics, reset the flags.
+	 * If we were successful, the page is no longer dirty.
 	 */
-	dosync = 0;
-	if (F_ISSET(bhp, BH_SYNC)) {
-		F_CLR(bhp, BH_SYNC | BH_SYNC_LOGFLSH);
+	if (ret == 0) {
+		DB_ASSERT(hp->hash_page_dirty != 0);
+		--hp->hash_page_dirty;
 
-		--mp->lsn_cnt;
-		if (mfp != NULL)
-			dosync = --mfp->lsn_cnt == 0 ? 1 : 0;
+		F_CLR(bhp, BH_DIRTY | BH_DIRTY_CREATE);
 	}
 
-	/* Update the page clean/dirty statistics. */
-	c_mp = BH_TO_CACHE(dbmp, bhp);
-	++c_mp->stat.st_page_clean;
-	DB_ASSERT(c_mp->stat.st_page_dirty != 0);
-	--c_mp->stat.st_page_dirty;
-
-	/* Update I/O statistics. */
-	if (mfp != NULL)
-		++mfp->stat.st_page_out;
-
-	/*
-	 * Do the sync after everything else has been updated, so any incoming
-	 * checkpoint doesn't see inconsistent information.
-	 *
-	 * XXX:
-	 * Don't lock the region around the sync, fsync(2) has no atomicity
-	 * issues.
-	 *
-	 * XXX:
-	 * We ignore errors from the sync -- it makes no sense to return an
-	 * error to the calling process, so set a flag causing the checkpoint
-	 * to be retried later.  There is a possibility, of course, that a
-	 * subsequent checkpoint was started and that we're going to force it
-	 * to fail.  That should be unlikely, and fixing it would be difficult.
-	 */
-	if (dosync) {
-		R_UNLOCK(dbenv, dbmp->reginfo);
-		syncfail = __os_fsync(dbenv, dbmfp->fhp) != 0;
-		R_LOCK(dbenv, dbmp->reginfo);
-		if (syncfail)
-			F_SET(mp, MP_LSN_RETRY);
-	}
-
-	if (wrotep != NULL)
-		*wrotep = 1;
-
-	return (0);
-
-syserr:	__db_err(dbenv, "%s: %s failed for page %lu",
-	    __memp_fn(dbmfp), fail, (u_long)bhp->pgno);
-
-err:	/* Unlock the buffer and reacquire the region lock. */
-	MUTEX_UNLOCK(dbenv, &bhp->mutex);
-	R_LOCK(dbenv, dbmp->reginfo);
-
-	/*
-	 * Clean up the flags based on a failure.
-	 *
-	 * The page remains dirty but we remove our lock.  If we rewrote the
-	 * page, it will need processing by the pgin routine before reuse.
-	 */
-	if (callpgin)
-		F_SET(bhp, BH_CALLPGIN);
+	/* Regardless, clear any sync wait-for count and remove our lock. */
+	bhp->ref_sync = 0;
 	F_CLR(bhp, BH_LOCKED);
 
 	return (ret);
@@ -541,15 +456,17 @@ __memp_pg(dbmfp, bhp, is_pgin)
 	int is_pgin;
 {
 	DBT dbt, *dbtp;
+	DB_ENV *dbenv;
 	DB_MPOOL *dbmp;
 	DB_MPREG *mpreg;
 	MPOOLFILE *mfp;
 	int ftype, ret;
 
 	dbmp = dbmfp->dbmp;
+	dbenv = dbmp->dbenv;
 	mfp = dbmfp->mfp;
 
-	MUTEX_THREAD_LOCK(dbmp->dbenv, dbmp->mutexp);
+	MUTEX_THREAD_LOCK(dbenv, dbmp->mutexp);
 
 	ftype = mfp->ftype;
 	for (mpreg = LIST_FIRST(&dbmp->dbregq);
@@ -563,28 +480,28 @@ __memp_pg(dbmfp, bhp, is_pgin)
 			dbt.data = R_ADDR(dbmp->reginfo, mfp->pgcookie_off);
 			dbtp = &dbt;
 		}
-		MUTEX_THREAD_UNLOCK(dbmp->dbenv, dbmp->mutexp);
+		MUTEX_THREAD_UNLOCK(dbenv, dbmp->mutexp);
 
 		if (is_pgin) {
 			if (mpreg->pgin != NULL &&
-			    (ret = mpreg->pgin(dbmp->dbenv,
+			    (ret = mpreg->pgin(dbenv,
 			    bhp->pgno, bhp->buf, dbtp)) != 0)
 				goto err;
 		} else
 			if (mpreg->pgout != NULL &&
-			    (ret = mpreg->pgout(dbmp->dbenv,
+			    (ret = mpreg->pgout(dbenv,
 			    bhp->pgno, bhp->buf, dbtp)) != 0)
 				goto err;
 		break;
 	}
 
 	if (mpreg == NULL)
-		MUTEX_THREAD_UNLOCK(dbmp->dbenv, dbmp->mutexp);
+		MUTEX_THREAD_UNLOCK(dbenv, dbmp->mutexp);
 
 	return (0);
 
-err:	MUTEX_THREAD_UNLOCK(dbmp->dbenv, dbmp->mutexp);
-	__db_err(dbmp->dbenv, "%s: %s failed for page %lu",
+err:	MUTEX_THREAD_UNLOCK(dbenv, dbmp->mutexp);
+	__db_err(dbenv, "%s: %s failed for page %lu",
 	    __memp_fn(dbmfp), is_pgin ? "pgin" : "pgout", (u_long)bhp->pgno);
 	return (ret);
 }
@@ -593,56 +510,78 @@ err:	MUTEX_THREAD_UNLOCK(dbmp->dbenv, dbmp->mutexp);
  * __memp_bhfree --
  *	Free a bucket header and its referenced data.
  *
- * PUBLIC: void __memp_bhfree __P((DB_MPOOL *, BH *, int));
+ * PUBLIC: void __memp_bhfree __P((DB_MPOOL *, DB_MPOOL_HASH *, BH *, int));
  */
 void
-__memp_bhfree(dbmp, bhp, free_mem)
+__memp_bhfree(dbmp, hp, bhp, free_mem)
 	DB_MPOOL *dbmp;
+	DB_MPOOL_HASH *hp;
 	BH *bhp;
 	int free_mem;
 {
-	DB_HASHTAB *dbht;
+	DB_ENV *dbenv;
 	MPOOL *c_mp, *mp;
 	MPOOLFILE *mfp;
-	int n_bucket, n_cache;
+	u_int32_t n_cache;
 
+	/*
+	 * Assumes the hash bucket is locked and the MPOOL is not.
+	 */
+	dbenv = dbmp->dbenv;
 	mp = dbmp->reginfo[0].primary;
-	c_mp = BH_TO_CACHE(dbmp, bhp);
-	n_cache = NCACHE(mp, bhp->pgno);
-	n_bucket = NBUCKET(c_mp, bhp->mf_offset, bhp->pgno);
-	dbht = R_ADDR(&dbmp->reginfo[n_cache], c_mp->htab);
+	n_cache = NCACHE(mp, bhp->mf_offset, bhp->pgno);
 
-	/* Delete the buffer header from the hash bucket queue. */
-	SH_TAILQ_REMOVE(&dbht[n_bucket], bhp, hq, __bh);
+	/*
+	 * Delete the buffer header from the hash bucket queue and reset
+	 * the hash bucket's priority, if necessary.
+	 */
+	SH_TAILQ_REMOVE(&hp->hash_bucket, bhp, hq, __bh);
+	if (bhp->priority == hp->hash_priority)
+		hp->hash_priority =
+		    SH_TAILQ_FIRST(&hp->hash_bucket, __bh) == NULL ?
+		    0 : SH_TAILQ_FIRST(&hp->hash_bucket, __bh)->priority;
 
-	/* Delete the buffer header from the LRU queue. */
-	SH_TAILQ_REMOVE(&c_mp->bhq, bhp, q, __bh);
+	/*
+	 * Discard the hash bucket's mutex, it's no longer needed, and
+	 * we don't want to be holding it when acquiring other locks.
+	 */
+	MUTEX_UNLOCK(dbenv, &hp->hash_mutex);
 
-	/* Clear the mutex this buffer recorded */
-	__db_shlocks_clear(&bhp->mutex, &dbmp->reginfo[n_cache],
-	    (REGMAINT *)R_ADDR(&dbmp->reginfo[n_cache], mp->maint_off));
 	/*
 	 * Find the underlying MPOOLFILE and decrement its reference count.
 	 * If this is its last reference, remove it.
 	 */
 	mfp = R_ADDR(dbmp->reginfo, bhp->mf_offset);
+	MUTEX_LOCK(dbenv, &mfp->mutex);
 	if (--mfp->block_cnt == 0 && mfp->mpf_cnt == 0)
 		__memp_mf_discard(dbmp, mfp);
+	else
+		MUTEX_UNLOCK(dbenv, &mfp->mutex);
 
-	DB_ASSERT(c_mp->stat.st_page_clean != 0);
-	--c_mp->stat.st_page_clean;
+	R_LOCK(dbenv, &dbmp->reginfo[n_cache]);
 
 	/*
-	 * If we're not reusing it immediately, free the buffer header
+	 * Clear the mutex this buffer recorded; requires the region lock
+	 * be held.
+	 */
+	__db_shlocks_clear(&bhp->mutex, &dbmp->reginfo[n_cache],
+	    (REGMAINT *)R_ADDR(&dbmp->reginfo[n_cache], mp->maint_off));
+
+	/*
+	 * If we're not reusing the buffer immediately, free the buffer header
 	 * and data for real.
 	 */
-	if (free_mem)
+	if (free_mem) {
 		__db_shalloc_free(dbmp->reginfo[n_cache].addr, bhp);
+		c_mp = dbmp->reginfo[n_cache].primary;
+		c_mp->stat.st_pages--;
+	}
+	R_UNLOCK(dbenv, &dbmp->reginfo[n_cache]);
 }
 
 /*
  * __memp_upgrade --
- *	Upgrade a file descriptor from readonly to readwrite.
+ *	Upgrade a file descriptor from read-only to read-write.
  */
 static int
 __memp_upgrade(dbmp, dbmfp, mfp)
@@ -650,41 +589,58 @@ __memp_upgrade(dbmp, dbmfp, mfp)
 	DB_MPOOLFILE *dbmfp;
 	MPOOLFILE *mfp;
 {
-	DB_FH fh;
+	DB_ENV *dbenv;
+	DB_FH *fhp, *tfhp;
 	int ret;
 	char *rpath;
 
-	/*
-	 * !!!
-	 * We expect the handle to already be locked.
-	 */
-
-	/* Check to see if we've already upgraded. */
-	if (F_ISSET(dbmfp, MP_UPGRADE))
-		return (0);
-
-	/* Check to see if we've already failed. */
-	if (F_ISSET(dbmfp, MP_UPGRADE_FAIL))
-		return (1);
+	dbenv = dbmp->dbenv;
+	fhp = NULL;
+	rpath = NULL;
 
 	/*
 	 * Calculate the real name for this file and try to open it read/write.
 	 * We know we have a valid pathname for the file because it's the only
 	 * way we could have gotten a file descriptor of any kind.
 	 */
-	if ((ret = __db_appname(dbmp->dbenv, DB_APP_DATA,
-	    NULL, R_ADDR(dbmp->reginfo, mfp->path_off), 0, NULL, &rpath)) != 0)
-		return (ret);
-	if (__os_open(dbmp->dbenv, rpath, 0, 0, &fh) != 0) {
+	if ((ret = __os_calloc(dbenv, 1, sizeof(DB_FH), &fhp)) != 0)
+		goto err;
+
+	if ((ret = __db_appname(dbenv, DB_APP_DATA,
+	    R_ADDR(dbmp->reginfo, mfp->path_off), 0, NULL, &rpath)) != 0)
+		goto err;
+
+	if (__os_open(dbenv, rpath,
+	    F_ISSET(mfp, MP_DIRECT) ? DB_OSO_DIRECT : 0, 0, fhp) != 0) {
 		F_SET(dbmfp, MP_UPGRADE_FAIL);
-		ret = 1;
-	} else {
-		/* Swap the descriptors and set the upgrade flag. */
-		(void)__os_closehandle(dbmfp->fhp);
-		*dbmfp->fhp = fh;
-		F_SET(dbmfp, MP_UPGRADE);
-		ret = 0;
+		goto err;
 	}
-	__os_freestr(dbmp->dbenv, rpath);
+
+	/*
+	 * Swap the descriptors and set the upgrade flag.
+	 *
+	 * XXX
+	 * There is a race here.  If another process schedules a read using the
+	 * existing file descriptor and is swapped out before making the system
+	 * call, this code could theoretically close the file descriptor out
+	 * from under it.  While it's very unlikely, this code should still be
+	 * rewritten.
+	 */
+	tfhp = dbmfp->fhp;
+	dbmfp->fhp = fhp;
+	fhp = tfhp;
+
+	(void)__os_closehandle(dbenv, fhp);
+	F_SET(dbmfp, MP_UPGRADE);
+
+	ret = 0;
+	if (0) {
+err:		ret = 1;
+	}
+	if (fhp != NULL)
+		__os_free(dbenv, fhp);
+	if (rpath != NULL)
+		__os_free(dbenv, rpath);
+
 	return (ret);
 }
