@@ -1,7 +1,7 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2001
+ * Copyright (c) 1996-2002
  *	Sleepycat Software.  All rights reserved.
  */
 /*
@@ -39,11 +39,12 @@
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "Id: txn.c,v 11.125 2001/11/16 16:28:17 bostic Exp ";
+static const char revid[] = "Id: txn.c,v 11.179 2002/08/29 17:41:17 margo Exp ";
 #endif /* not lint */
 
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
+#include <stdlib.h>
 
 #if TIME_WITH_SYS_TIME
 #include <sys/time.h>
@@ -60,12 +61,28 @@ static const char revid[] = "Id: txn.c,v 11.125 2001/11/16 16:28:17 bostic Exp "
 #endif
 
 #include "db_int.h"
-#include "db_page.h"
-#include "db_shash.h"
-#include "hash.h"
-#include "lock.h"
-#include "log.h"
-#include "txn.h"
+#include "dbinc/crypto.h"
+#include "dbinc/hmac.h"
+#include "dbinc/db_page.h"
+#include "dbinc/db_shash.h"
+#include "dbinc/hash.h"
+#include "dbinc/lock.h"
+#include "dbinc/log.h"
+#include "dbinc/txn.h"
+
+#define	SET_LOG_FLAGS(dbenv, txnp, lflags)				\
+	do {								\
+		lflags = DB_COMMIT | DB_PERMANENT;			\
+		if (F_ISSET(txnp, TXN_SYNC))				\
+			lflags |= DB_FLUSH;				\
+		else if (!F_ISSET(txnp, TXN_NOSYNC) &&			\
+		    !F_ISSET(dbenv, DB_ENV_TXN_NOSYNC)) {		\
+			if (F_ISSET(dbenv, DB_ENV_TXN_WRITE_NOSYNC))	\
+				lflags |= DB_WRNOSYNC;			\
+			else						\
+				lflags |= DB_FLUSH;			\
+		}							\
+	} while (0)
 
 /*
  * __txn_isvalid enumerated types.  We cannot simply use the transaction
@@ -80,7 +97,7 @@ typedef enum {
 } txnop_t;
 
 static int  __txn_begin_int __P((DB_TXN *, int));
-static void __txn_end __P((DB_TXN *, int));
+static int  __txn_end __P((DB_TXN *, int));
 static int  __txn_isvalid __P((const DB_TXN *, TXN_DETAIL **, txnop_t));
 static int  __txn_set_timeout __P(( DB_TXN *, db_timeout_t, u_int32_t));
 static int  __txn_undo __P((DB_TXN *));
@@ -177,6 +194,7 @@ __txn_begin(dbenv, parent, txnpp, flags)
 	txn->mgrp = dbenv->tx_handle;
 	txn->parent = parent;
 	TAILQ_INIT(&txn->kids);
+	TAILQ_INIT(&txn->events);
 	txn->flags = TXN_MALLOC;
 	if (LF_ISSET(DB_DIRTY_READ))
 		F_SET(txn, TXN_DIRTY_READ);
@@ -221,7 +239,7 @@ __txn_begin(dbenv, parent, txnpp, flags)
 	return (0);
 
 err:
-	__os_free(dbenv, txn, sizeof(DB_TXN));
+	__os_free(dbenv, txn);
 	return (ret);
 }
 
@@ -241,6 +259,8 @@ __txn_xa_begin(dbenv, txn)
 	memset(txn, 0, sizeof(DB_TXN));
 
 	txn->mgrp = dbenv->tx_handle;
+	TAILQ_INIT(&txn->kids);
+	TAILQ_INIT(&txn->events);
 
 	return (__txn_begin_int(txn, 0));
 }
@@ -267,9 +287,12 @@ __txn_compensate_begin(dbenv, txnpp)
 		return (ret);
 
 	txn->mgrp = dbenv->tx_handle;
+	TAILQ_INIT(&txn->kids);
+	TAILQ_INIT(&txn->events);
+	txn->flags = TXN_MALLOC;
+	F_SET(txn, TXN_COMPENSATE);
 
 	*txnpp = txn;
-	F_SET(txn, TXN_COMPENSATE);
 	return (__txn_begin_int(txn, 1));
 }
 
@@ -283,13 +306,13 @@ __txn_begin_int(txn, internal)
 	int internal;
 {
 	DB_ENV *dbenv;
-	DB_LSN begin_lsn;
+	DB_LSN begin_lsn, null_lsn;
 	DB_TXNMGR *mgr;
 	DB_TXNREGION *region;
 	TXN_DETAIL *td;
 	size_t off;
-	u_int32_t id, min_inuse;
-	int ret;
+	u_int32_t id, *ids;
+	int nids, ret;
 
 	mgr = txn->mgrp;
 	dbenv = mgr->dbenv;
@@ -300,81 +323,64 @@ __txn_begin_int(txn, internal)
 	 * need never write records for read-only transactions).  However,
 	 * we do need to find the current LSN so that we can store it in the
 	 * transaction structure, so we can know where to take checkpoints.
+	 *
+	 * XXX
+	 * We should set this value when we write the first log record, not
+	 * here.
 	 */
-	if (LOGGING_ON(dbenv) &&
-	    (ret = dbenv->log_put(dbenv, &begin_lsn, NULL, DB_CURLSN)) != 0)
-		goto err2;
+	if (DBENV_LOGGING(dbenv))
+		__log_txn_lsn(dbenv, &begin_lsn, NULL, NULL);
 
 	R_LOCK(dbenv, &mgr->reginfo);
 	if (!F_ISSET(txn, TXN_COMPENSATE) && F_ISSET(region, TXN_IN_RECOVERY)) {
-		__db_err(dbenv, "operation not permitted during recovery.");
+		__db_err(dbenv, "operation not permitted during recovery");
 		ret = EINVAL;
-		goto err1;
+		goto err;
 	}
 
 	/* Make sure that we aren't still recovering prepared transactions. */
 	if (!internal && region->stat.st_nrestores != 0) {
 		__db_err(dbenv,
-"txn_begin: recovery of prepared but not yet committed transactions is incomplete.");
+    "recovery of prepared but not yet committed transactions is incomplete");
 		ret = EINVAL;
-		goto err1;
+		goto err;
 	}
 
 	/*
-	 * Allocate a new transaction id.  If we wrap around then we
-	 * find the minimum currently in use and make sure we
-	 * can stay below that.  This code is similar to code
-	 * in __lock_id for recovering locker ids.
+	 * Allocate a new transaction id. Our current valid range can span
+	 * the maximum valid value, so check for it and wrap manually.
 	 */
+	if (region->last_txnid == TXN_MAXIMUM &&
+	    region->cur_maxid != TXN_MAXIMUM)
+		region->last_txnid = TXN_MINIMUM - 1;
 
 	if (region->last_txnid == region->cur_maxid) {
-		min_inuse = TXN_INVALID;
-		/* If cur_maxid is at the top we are wrapping around. */
-		if (region->cur_maxid == TXN_INVALID)
-			region->cur_maxid = TXN_MINIMUM;
-
-		/*
-		 * Traverse teh active transactions and find the
-		 * lowest txnid which is in use.
-		 */
+		if ((ret = __os_malloc(dbenv,
+		    sizeof(u_int32_t) * region->maxtxns, &ids)) != 0)
+			goto err;
+		nids = 0;
 		for (td = SH_TAILQ_FIRST(&region->active_txn, __txn_detail);
 		    td != NULL;
-		    td = SH_TAILQ_NEXT(td, links, __txn_detail)) {
-			if (td->txnid > region->cur_maxid
-			     && td->txnid < min_inuse)
-				min_inuse = td->txnid;
-		}
-		/* Set the maxid to the new lowest. */
-		region->cur_maxid = min_inuse;
-		/*
-		 * Check to see if the current txnid is wrapping around.
-		 * Since this should represent the last id allocated
-		 * and we don't use TXN_MINIMUM  things are
-		 * a little tricky, we don't want to allocate
-		 * the first id if its inuse.
-		 */
-		if (region->last_txnid == TXN_INVALID) {
-			if (min_inuse == TXN_MINIMUM + 1)
-				region->last_txnid = min_inuse;
-			else
-				region->last_txnid = TXN_MINIMUM;
-		}
-
-		/* See if we fixed the situation. */
-		if (region->last_txnid == region->cur_maxid) {
-			__db_err(dbenv,
-	"txn_begin: transaction ID wrapped.  Exit the database environment\nand restart the application as if application failure had occurred");
-			ret = EINVAL;
-			goto err1;
-		}
+		    td = SH_TAILQ_NEXT(td, links, __txn_detail))
+			ids[nids++] = td->txnid;
+		region->last_txnid = TXN_MINIMUM - 1;
+		region->cur_maxid = TXN_MAXIMUM;
+		if (nids != 0)
+			__db_idspace(ids, nids,
+			    &region->last_txnid, &region->cur_maxid);
+		__os_free(dbenv, ids);
+		if (DBENV_LOGGING(dbenv) &&
+		    (ret = __txn_recycle_log(dbenv, NULL,
+		    &null_lsn, 0, region->last_txnid, region->cur_maxid)) != 0)
+			goto err;
 	}
 
 	/* Allocate a new transaction detail structure. */
 	if ((ret =
 	    __db_shalloc(mgr->reginfo.addr, sizeof(TXN_DETAIL), 0, &td)) != 0) {
 		__db_err(dbenv,
-		     "Unable to allocate memory for transaction detail");
-		goto err1;
+		    "Unable to allocate memory for transaction detail");
+		goto err;
 	}
 
 	/* Place transaction on active transaction list. */
@@ -400,7 +406,7 @@ __txn_begin_int(txn, internal)
 
 	ZERO_LSN(txn->last_lsn);
 	txn->txnid = id;
-	txn->off = off;
+	txn->off = (u_int32_t)off;
 
 	txn->abort = __txn_abort;
 	txn->commit = __txn_commit;
@@ -416,7 +422,7 @@ __txn_begin_int(txn, internal)
 	if (txn->parent != NULL && LOCKING_ON(dbenv))
 		if ((ret = __lock_addfamilylocker(dbenv,
 		    txn->parent->txnid, txn->txnid)) != 0)
-			goto err2;
+			return (ret);
 
 	if (F_ISSET(txn, TXN_MALLOC)) {
 		MUTEX_THREAD_LOCK(dbenv, mgr->mutexp);
@@ -426,9 +432,8 @@ __txn_begin_int(txn, internal)
 
 	return (0);
 
-err1:	R_UNLOCK(dbenv, &mgr->reginfo);
-
-err2:	return (ret);
+err:	R_UNLOCK(dbenv, &mgr->reginfo);
+	return (ret);
 }
 
 /*
@@ -445,13 +450,15 @@ __txn_commit(txnp, flags)
 	DB_ENV *dbenv;
 	DB_LOCKREQ request;
 	DB_TXN *kid;
-	int ret;
+	TXN_DETAIL *td;
+	u_int32_t lflags;
+	int ret, t_ret;
 
 	dbenv = txnp->mgrp->dbenv;
 
 	PANIC_CHECK(dbenv);
 
-	if ((ret = __txn_isvalid(txnp, NULL, TXN_OP_COMMIT)) != 0)
+	if ((ret = __txn_isvalid(txnp, &td, TXN_OP_COMMIT)) != 0)
 		return (ret);
 
 	/*
@@ -477,15 +484,33 @@ __txn_commit(txnp, flags)
 	}
 
 	/*
-	 * Commit any unresolved children.  If there's an error, abort any
-	 * unresolved children and the parent.
+	 * Commit any unresolved children.  If anyone fails to commit,
+	 * then try to abort the rest of the kids and then abort the parent.
+	 * Abort should never fail; if it does, we bail out immediately.
 	 */
 	while ((kid = TAILQ_FIRST(&txnp->kids)) != NULL)
-		if ((ret = kid->commit(kid, flags)) != 0) {
+		if ((ret = kid->commit(kid, flags)) != 0)
 			while ((kid = TAILQ_FIRST(&txnp->kids)) != NULL)
-				(void)kid->abort(kid);
-			goto err;
-		}
+				if ((t_ret = kid->abort(kid)) != 0)
+					return (__db_panic(dbenv, t_ret));
+
+	/*
+	 * Process any aborted pages from our children.
+	 * We delay putting pages on the free list that are newly
+	 * allocated and then aborted so that we can undo other
+	 * allocations, if necessary, without worrying about
+	 * these pages which were not on the free list before.
+	 */
+	if (txnp->txn_list != NULL) {
+		t_ret = __db_do_the_limbo(dbenv, NULL, txnp, txnp->txn_list);
+		__db_txnlist_end(dbenv, txnp->txn_list);
+		txnp->txn_list = NULL;
+		if (t_ret != 0 && ret == 0)
+			ret = t_ret;
+	}
+
+	if (ret != 0)
+		goto err;
 
 	/*
 	 * If there are any log records, write a log record and sync the log,
@@ -494,19 +519,27 @@ __txn_commit(txnp, flags)
 	 * abort (if its parent aborts), and otherwise its parent or ultimate
 	 * ancestor will write synchronously.
 	 */
-	if (LOGGING_ON(dbenv) && !IS_ZERO_LSN(txnp->last_lsn)) {
+	if (DBENV_LOGGING(dbenv) && !IS_ZERO_LSN(txnp->last_lsn)) {
 		if (txnp->parent == NULL) {
+			/*
+			 * We are about to free all the read locks
+			 * for this transaction below.  Some of those
+			 * locks might be handle locks which should
+			 * not be freed, because they will be freed
+			 * when the handle is closed.  Check the
+			 * events and preprocess any trades now so
+			 * that we don't release the locks below.
+			 */
+			if ((ret = __txn_doevents(dbenv, txnp, 0, 1)) != 0)
+				goto err;
 			request.op = DB_LOCK_PUT_READ;
 			if (LOCKING_ON(dbenv) && (ret = dbenv->lock_vec(
 			   dbenv, txnp->txnid, 0, &request, 1, NULL)) != 0)
 				goto err;
 
+			SET_LOG_FLAGS(dbenv, txnp, lflags);
 			if ((ret = __txn_regop_log(dbenv,
-			    txnp, &txnp->last_lsn,
-			    DB_COMMIT |
-			    ((F_ISSET(dbenv, DB_ENV_TXN_NOSYNC) &&
-			    !F_ISSET(txnp, TXN_SYNC)) ||
-			    F_ISSET(txnp, TXN_NOSYNC) ?  0 : DB_FLUSH),
+			    txnp, &txnp->last_lsn, lflags,
 			    TXN_COMMIT, (int32_t)time(NULL))) != 0)
 				goto err;
 		} else {
@@ -521,10 +554,23 @@ __txn_commit(txnp, flags)
 		}
 	}
 
-	__txn_end(txnp, 1);
-	return (0);
+	/* This is OK because __txn_end can only fail with a panic. */
+	return (__txn_end(txnp, 1));
 
-err:	(void)txnp->abort(txnp);
+err:	/*
+	 * If we are prepared, then we "must" be able to commit.  We
+	 * panic here because even though the coordinator might be
+	 * able to retry it is not clear it would know to do that.
+	 * Otherwise  we'll try to abort.  If that is successful,
+	 * then we return whatever was in ret (i.e., the reason we failed).
+	 * If the abort was unsuccessful, then abort probably returned
+	 * DB_RUNRECOVERY and we need to propagate that up.
+	 */
+	if (td->status == TXN_PREPARED)
+		return (__db_panic(dbenv, ret));
+		
+	if ((t_ret = txnp->abort(txnp)) != 0)
+		ret = t_ret;
 	return (ret);
 }
 
@@ -542,55 +588,70 @@ __txn_abort(txnp)
 	DB_LOCKREQ request;
 	DB_TXN *kid;
 	TXN_DETAIL *td;
-	int ret, t_ret;
+	u_int32_t lflags;
+	int ret;
 
 	dbenv = txnp->mgrp->dbenv;
 
 	PANIC_CHECK(dbenv);
 
+	/* Ensure that abort always fails fatally. */
 	if ((ret = __txn_isvalid(txnp, &td, TXN_OP_ABORT)) != 0)
-		return (ret);
+		return (__db_panic(dbenv, ret));
 
-	/* Abort any unresolved children. */
+	/*
+	 * Try to abort any unresolved children.
+	 *
+	 * Abort either succeeds or panics the region.  As soon as we
+	 * see any failure, we just get out of here and return the panic
+	 * up.
+	 */
 	while ((kid = TAILQ_FIRST(&txnp->kids)) != NULL)
-		if ((t_ret = kid->abort(kid)) != 0 && ret == 0)
-			ret = t_ret;
+		if ((ret = kid->abort(kid)) != 0)
+			return (ret);
 
 	if (LOCKING_ON(dbenv)) {
+		/*
+		 * We are about to free all the read locks for this transaction
+		 * below.  Some of those locks might be handle locks which
+		 * should not be freed, because they will be freed when the
+		 * handle is closed.  Check the events and preprocess any
+		 * trades now so that we don't release the locks below.
+		 */
+		if ((ret = __txn_doevents(dbenv, txnp, 0, 1)) != 0)
+			return (__db_panic(dbenv, ret));
+
 		/* Turn off timeouts. */
-		if ((t_ret = __lock_set_timeout(dbenv,
-		     txnp->txnid, 0, DB_SET_TXN_TIMEOUT)) != 0 && ret == 0)
-			ret = t_ret;
-		if ((t_ret = __lock_set_timeout(dbenv,
-		     txnp->txnid, 0, DB_SET_LOCK_TIMEOUT)) != 0 && ret == 0)
-			ret = t_ret;
+		if ((ret = __lock_set_timeout(dbenv,
+		    txnp->txnid, 0, DB_SET_TXN_TIMEOUT)) != 0)
+			return (__db_panic(dbenv, ret));
+
+		if ((ret = __lock_set_timeout(dbenv,
+		    txnp->txnid, 0, DB_SET_LOCK_TIMEOUT)) != 0)
+			return (__db_panic(dbenv, ret));
 
 		request.op = DB_LOCK_UPGRADE_WRITE;
 		if ((ret = dbenv->lock_vec(
 		    dbenv, txnp->txnid, 0, &request, 1, NULL)) != 0)
-			(void)__db_panic(dbenv, ret);
+			return (__db_panic(dbenv, ret));
 	}
-	if ((t_ret = __txn_undo(txnp)) != 0 && ret == 0)
-		ret = t_ret;
+	if ((ret = __txn_undo(txnp)) != 0)
+		return (__db_panic(dbenv, ret));
 
 	/*
 	 * Normally, we do not need to log aborts.  However, if we
 	 * are a distributed transaction (i.e., we have a prepare),
 	 * then we log the abort so we know that this transaction
-	 * was actually completed.  Even if the log fails; abort/undo
-	 * the transaction.
+	 * was actually completed.
 	 */
-	if (LOGGING_ON(dbenv) && td->status == TXN_PREPARED &&
-	    (t_ret = __txn_regop_log(dbenv, txnp, &txnp->last_lsn,
-	    (F_ISSET(dbenv, DB_ENV_TXN_NOSYNC) &&
-	    !F_ISSET(txnp, TXN_SYNC)) ||
-	    F_ISSET(txnp, TXN_NOSYNC) ?  0 : DB_FLUSH,
-	    TXN_ABORT, (int32_t)time(NULL))) != 0 && ret == 0)
-		ret = t_ret;
+	SET_LOG_FLAGS(dbenv, txnp, lflags);
+	if (DBENV_LOGGING(dbenv) && td->status == TXN_PREPARED &&
+	    (ret = __txn_regop_log(dbenv, txnp, &txnp->last_lsn,
+	    lflags, TXN_ABORT, (int32_t)time(NULL))) != 0)
+		return (__db_panic(dbenv, ret));
 
-	__txn_end(txnp, 0);
-
-	return (ret);
+	/* __txn_end always panics if it errors, so pass the return along. */
+	return (__txn_end(txnp, 0));
 }
 
 /*
@@ -632,9 +693,9 @@ __txn_discard(txnp, flags)
 	}
 	MUTEX_THREAD_UNLOCK(dbenv, txnp->mgrp->mutexp);
 	if (freep != NULL)
-		__os_free(dbenv, freep, sizeof(*freep));
+		__os_free(dbenv, freep);
 
-	return (ret);
+	return (0);
 }
 
 /*
@@ -652,6 +713,7 @@ __txn_prepare(txnp, gid)
 	DB_ENV *dbenv;
 	DB_TXN *kid;
 	TXN_DETAIL *td;
+	u_int32_t lflags;
 	int ret;
 
 	dbenv = txnp->mgrp->dbenv;
@@ -661,11 +723,6 @@ __txn_prepare(txnp, gid)
 	if ((ret = __txn_isvalid(txnp, &td, TXN_OP_PREPARE)) != 0)
 		return (ret);
 
-	if (txnp->parent != NULL) {
-		__db_err(dbenv, "Prepare disallowed on child transactions.");
-		return (EINVAL);
-	}
-
 	/* Commit any unresolved children. */
 	while ((kid = TAILQ_FIRST(&txnp->kids)) != NULL)
 		if ((ret = kid->commit(kid, DB_TXN_NOSYNC)) != 0)
@@ -679,7 +736,7 @@ __txn_prepare(txnp, gid)
 	 * of those states, then we are calling prepare directly and we need
 	 * to fill in the td->xid.
 	 */
-	if (LOGGING_ON(dbenv)) {
+	if (DBENV_LOGGING(dbenv)) {
 		memset(&xid, 0, sizeof(xid));
 		if (td->xa_status != TXN_XA_ENDED &&
 		    td->xa_status != TXN_XA_SUSPENDED)
@@ -689,11 +746,9 @@ __txn_prepare(txnp, gid)
 		xid.size = sizeof(td->xid);
 		xid.data = td->xid;
 
+		SET_LOG_FLAGS(dbenv, txnp, lflags);
 		if ((ret = __txn_xa_regop_log(dbenv, txnp, &txnp->last_lsn,
-		    DB_COMMIT | ((F_ISSET(dbenv, DB_ENV_TXN_NOSYNC) &&
-		    !F_ISSET(txnp, TXN_SYNC)) ||
-		    F_ISSET(txnp, TXN_NOSYNC) ? 0 : DB_FLUSH), TXN_PREPARE,
-		    &xid, td->format, td->gtrid, td->bqual,
+		    lflags, TXN_PREPARE, &xid, td->format, td->gtrid, td->bqual,
 		    &td->begin_lsn)) != 0) {
 			__db_err(dbenv, "DB_TXN->prepare: log_write failed %s",
 			    db_strerror(ret));
@@ -758,7 +813,7 @@ __txn_isvalid(txnp, tdp, op)
 	if (!F_ISSET(txnp, TXN_COMPENSATE) &&
 	    F_ISSET(region, TXN_IN_RECOVERY)) {
 		__db_err(mgrp->dbenv,
-		    "operation not permitted during recovery.");
+		    "operation not permitted during recovery");
 		goto err;
 	}
 
@@ -773,42 +828,46 @@ __txn_isvalid(txnp, tdp, op)
 	if (tdp != NULL)
 		*tdp = tp;
 
-	/*
-	 * Discard is a special case since we're just tossing the
-	 * per-process space; there are a lot of problems with the
-	 * transaction that we can tolerate.
-	 */
-	if (op == TXN_OP_DISCARD) {
+	/* Handle any operation specific checks. */
+	switch (op) {
+	case TXN_OP_DISCARD:
+		/*
+		 * Since we're just tossing the per-process space; there are
+		 * a lot of problems with the transaction that we can tolerate.
+		 */
+
 		/* Transaction is already been reused. */
 		if (txnp->txnid != tp->txnid)
 			return (0);
 
-		/*
-		 * We have the right transaction, we'd better be
-		 * a restored transaction.  However, if it's not,
-		 * we won't panic the environment since it means
-		 * that someone has a bad user-handle, not that the
-		 * transaction is corrupt.
-		 */
+		/* What we've got had better be a restored transaction. */
 		if (!F_ISSET(tp, TXN_RESTORED)) {
 			__db_err(mgrp->dbenv, "not a restored transaction");
-			return (EINVAL);
+			return (__db_panic(mgrp->dbenv, EINVAL));
 		}
 
-		/*
-		 * No matter what state the transaction is in, we can
-		 * blow away this process' handle.
-		 */
 		return (0);
+	case TXN_OP_PREPARE:
+		if (txnp->parent != NULL) {
+			/*
+			 * This is not fatal, because you could imagine an
+			 * application that simply prepares everybody because
+			 * it doesn't distinguish between children and parents.
+			 * I'm not arguing this is good, but I could imagine
+			 * someone doing it.
+			 */
+			__db_err(mgrp->dbenv,
+			    "Prepare disallowed on child transactions");
+			return (EINVAL);
+		}
+		break;
+	case TXN_OP_ABORT:
+	case TXN_OP_COMMIT:
+	default:
+		break;
 	}
 
 	switch (tp->status) {
-	case TXN_ABORTED:
-	case TXN_COMMITTED:
-	default:
-		__db_err(mgrp->dbenv, "transaction already %s",
-		    tp->status == TXN_COMMITTED ? "committed" : "aborted");
-		goto err;
 	case TXN_PREPARED:
 		if (op == TXN_OP_PREPARE) {
 			__db_err(mgrp->dbenv, "transaction already prepared");
@@ -822,6 +881,12 @@ __txn_isvalid(txnp, tdp, op)
 		break;
 	case TXN_RUNNING:
 		break;
+	case TXN_ABORTED:
+	case TXN_COMMITTED:
+	default:
+		__db_err(mgrp->dbenv, "transaction already %s",
+		    tp->status == TXN_COMMITTED ? "committed" : "aborted");
+		goto err;
 	}
 
 	return (0);
@@ -838,7 +903,7 @@ err:	/*
  * __txn_end --
  *	Internal transaction end routine.
  */
-static void
+static int
 __txn_end(txnp, is_commit)
 	DB_TXN *txnp;
 	int is_commit;
@@ -855,20 +920,24 @@ __txn_end(txnp, is_commit)
 	region = mgr->reginfo.primary;
 	do_closefiles = 0;
 
+	/* Process commit events. */
+	if ((ret = __txn_doevents(dbenv, txnp, is_commit, 0)) != 0)
+		return (__db_panic(dbenv, ret));
+
 	/* Release the locks. */
 	request.op = txnp->parent == NULL ||
 	    is_commit == 0 ? DB_LOCK_PUT_ALL : DB_LOCK_INHERIT;
 
 	/*
-	 * __txn_end cannot return an error, we MUST return success/failure
-	 * from commit or abort, ignoring any internal errors.  So, we panic
-	 * if something goes wrong.  We can't deadlock here because we're not
-	 * acquiring any new locks, so DB_LOCK_DEADLOCK is just as fatal as
-	 * any other error.
+	 * __txn_end cannot return an simple error, we MUST return
+	 * success/failure from commit or abort, ignoring any internal
+	 * errors.  So, we panic if something goes wrong.  We can't
+	 * deadlock here because we're not acquiring any new locks,
+	 * so DB_LOCK_DEADLOCK is just as fatal as any other error.
 	 */
 	if (LOCKING_ON(dbenv) && (ret = dbenv->lock_vec(
 	    dbenv, txnp->txnid, DB_LOCK_FREE_LOCKER, &request, 1, NULL)) != 0)
-		(void)__db_panic(dbenv, ret);
+		return (__db_panic(dbenv, ret));
 
 	/* End the transaction. */
 	R_LOCK(dbenv, &mgr->reginfo);
@@ -879,6 +948,7 @@ __txn_end(txnp, is_commit)
 		region->stat.st_nrestores--;
 		do_closefiles = region->stat.st_nrestores == 0;
 	}
+
 	__db_shalloc_free(mgr->reginfo.addr, tp);
 
 	if (is_commit)
@@ -895,7 +965,7 @@ __txn_end(txnp, is_commit)
 	 */
 	if (LOCKING_ON(dbenv) && (ret =
 	    __lock_freefamilylocker(dbenv->lk_handle, txnp->txnid)) != 0)
-		(void)__db_panic(dbenv, ret);
+		return (__db_panic(dbenv, ret));
 	if (txnp->parent != NULL)
 		TAILQ_REMOVE(&txnp->parent->kids, txnp, klinks);
 
@@ -905,16 +975,17 @@ __txn_end(txnp, is_commit)
 		TAILQ_REMOVE(&mgr->txn_chain, txnp, links);
 		MUTEX_THREAD_UNLOCK(dbenv, mgr->mutexp);
 
-		__os_free(dbenv, txnp, sizeof(*txnp));
+		__os_free(dbenv, txnp);
 	}
 
 	if (do_closefiles) {
 		F_SET((DB_LOG *)dbenv->lg_handle, DBLOG_RECOVER);
-		__log_close_files(dbenv);
+		(void)__dbreg_close_files(dbenv);
 		F_CLR((DB_LOG *)dbenv->lg_handle, DBLOG_RECOVER);
 		mgr->n_discards = 0;
 		(void)dbenv->txn_checkpoint(dbenv, 0, 0, DB_FORCE);
 	}
+	return (0);
 }
 
 /*
@@ -930,8 +1001,9 @@ __txn_undo(txnp)
 	DB_ENV *dbenv;
 	DB_LOGC *logc;
 	DB_LSN key_lsn;
+	DB_TXN *ptxn;
 	DB_TXNMGR *mgr;
-	int ret, t_ret, threaded;
+	int ret, t_ret;
 	void *txnlist;
 
 	mgr = txnp->mgrp;
@@ -939,7 +1011,7 @@ __txn_undo(txnp)
 	logc = NULL;
 	txnlist = NULL;
 
-	if (!LOGGING_ON(dbenv))
+	if (!DBENV_LOGGING(dbenv))
 		return (0);
 
 	/*
@@ -948,15 +1020,26 @@ __txn_undo(txnp)
 	 * allocation here and use DB_DBT_USERMEM.
 	 */
 	memset(&rdbt, 0, sizeof(rdbt));
-	threaded = F_ISSET(dbenv, DB_ENV_THREAD) ? 1 : 0;
-	if (threaded)
-		F_SET(&rdbt, DB_DBT_MALLOC);
 
 	key_lsn = txnp->last_lsn;
 
-	/* Allocate a transaction list for children or aborted page creates. */
-	if ((ret = __db_txnlist_init(dbenv, 0, 0, &txnlist)) != 0)
+	/*
+	 * Allocate a txnlist for children and aborted page allocs.
+	 * We need to associate the list with the maximal parent
+	 * so that aborted pages are recovered when that transaction
+	 * is commited or aborted.
+	 */
+	for (ptxn = txnp->parent; ptxn != NULL && ptxn->parent != NULL;)
+		ptxn = ptxn->parent;
+
+	if (ptxn != NULL && ptxn->txn_list != NULL)
+		txnlist = ptxn->txn_list;
+	else if (txnp->txn_list != NULL)
+		txnlist = txnp->txn_list;
+	else if ((ret = __db_txnlist_init(dbenv, 0, 0, NULL, &txnlist)) != 0)
 		return (ret);
+	else if (ptxn != NULL)
+		ptxn->txn_list = txnlist;
 
 	if (F_ISSET(txnp, TXN_CHILDCOMMIT) &&
 	    (ret = __db_txnlist_lsninit(dbenv, txnlist, &txnp->last_lsn)) != 0)
@@ -970,12 +1053,9 @@ __txn_undo(txnp)
 		 * before the current one in the key_lsn argument.
 		 */
 		if ((ret = logc->get(logc, &key_lsn, &rdbt, DB_SET)) == 0) {
-			ret = __db_dispatch(dbenv, NULL,
-			    &rdbt, &key_lsn, DB_TXN_ABORT, txnlist);
-			if (threaded && rdbt.data != NULL) {
-				__os_free(dbenv, rdbt.data, rdbt.size);
-				rdbt.data = NULL;
-			}
+			ret = __db_dispatch(dbenv, dbenv->recover_dtab,
+			    dbenv->recover_dtab_size, &rdbt, &key_lsn,
+			    DB_TXN_ABORT, txnlist);
 			if (F_ISSET(txnp, TXN_CHILDCOMMIT))
 				(void)__db_txnlist_lsnadd(dbenv,
 				    txnlist, &key_lsn, 0);
@@ -993,13 +1073,12 @@ __txn_undo(txnp)
 		}
 	}
 
-	if (txnlist != NULL)
-		ret = __db_do_the_limbo(dbenv, txnlist);
+	ret = __db_do_the_limbo(dbenv, ptxn, txnp, txnlist);
 
 err:	if (logc != NULL && (t_ret = logc->close(logc, 0)) != 0 && ret == 0)
 		ret = t_ret;
 
-	if (txnlist != NULL)
+	if (ptxn == NULL && txnlist != NULL)
 		__db_txnlist_end(dbenv, txnlist);
 	return (ret);
 }
@@ -1022,141 +1101,106 @@ __txn_checkpoint(dbenv, kbytes, minutes, flags)
 	DB_ENV *dbenv;
 	u_int32_t kbytes, minutes, flags;
 {
-	DB_LOG *dblp;
-	DB_LSN ckp_lsn, sync_lsn, last_ckp;
+	DB_LSN ckp_lsn, last_ckp;
 	DB_TXNMGR *mgr;
 	DB_TXNREGION *region;
-	LOG *lp;
 	TXN_DETAIL *txnp;
 	time_t last_ckp_time, now;
 	u_int32_t bytes, mbytes;
-	int interval, ret;
+	int ret;
 
 	PANIC_CHECK(dbenv);
 	ENV_REQUIRES_CONFIG(dbenv,
 	    dbenv->tx_handle, "txn_checkpoint", DB_INIT_TXN);
 
+	/*
+	 * On a replication client, all transactions are read-only;  therefore,
+	 * a checkpoint is a null-op.
+	 *
+	 * We permit txn_checkpoint, instead of just rendering it illegal,
+	 * so that an application can just let a checkpoint thread continue
+	 * to operate as it gets promoted or demoted between being a
+	 * master and a client.
+	 */
+	if (F_ISSET(dbenv, DB_ENV_REP_CLIENT))
+		return (0);
+
 	mgr = dbenv->tx_handle;
 	region = mgr->reginfo.primary;
-	dblp = dbenv->lg_handle;
-	lp = dblp->reginfo.primary;
 
 	/*
-	 * Check if we need to checkpoint.
+	 * The checkpoint LSN is an LSN such that all transactions begun before
+	 * it are complete.  Our first guess (corrected below based on the list
+	 * of active transactions) is the last-written LSN.
 	 */
-	ZERO_LSN(ckp_lsn);
+	__log_txn_lsn(dbenv, &ckp_lsn, &mbytes, &bytes);
 
-	if (LF_ISSET(DB_FORCE))
-		goto do_ckp;
+	if (!LF_ISSET(DB_FORCE)) {
+		/* Don't checkpoint a quiescent database. */
+		if (bytes == 0 && mbytes == 0)
+			return (0);
 
-	R_LOCK(dbenv, &dblp->reginfo);
-	mbytes = lp->stat.st_wc_mbytes;
-	/*
-	 * We add the current buffer offset so as to count bytes that
-	 * have not yet been written, but are sitting in the log buffer.
-	 */
-	bytes = lp->stat.st_wc_bytes + lp->b_off;
-	ckp_lsn = lp->lsn;
-	R_UNLOCK(dbenv, &dblp->reginfo);
-
-	/* Don't checkpoint a quiescent database. */
-	if (bytes == 0 && mbytes == 0)
-		return (0);
-
-	if (kbytes != 0 && mbytes * 1024 + bytes / 1024 >= (u_int32_t)kbytes)
-		goto do_ckp;
-
-	if (minutes != 0) {
-		(void)time(&now);
-
-		R_LOCK(dbenv, &mgr->reginfo);
-		last_ckp_time = region->time_ckp;
-		R_UNLOCK(dbenv, &mgr->reginfo);
-
-		if (now - last_ckp_time >= (time_t)(minutes * 60))
+		if (kbytes != 0 &&
+		    mbytes * 1024 + bytes / 1024 >= (u_int32_t)kbytes)
 			goto do_ckp;
-	}
 
-	/*
-	 * If we checked time and data and didn't go to checkpoint,
-	 * we're done.
-	 */
-	if (minutes != 0 || kbytes != 0)
-		return (0);
+		if (minutes != 0) {
+			(void)time(&now);
 
-do_ckp:
-	if (IS_ZERO_LSN(ckp_lsn)) {
-		R_LOCK(dbenv, &dblp->reginfo);
-		ckp_lsn = lp->lsn;
-		R_UNLOCK(dbenv, &dblp->reginfo);
-	}
+			R_LOCK(dbenv, &mgr->reginfo);
+			last_ckp_time = region->time_ckp;
+			R_UNLOCK(dbenv, &mgr->reginfo);
 
-	/*
-	 * We have to find an LSN such that all transactions begun
-	 * before that LSN are complete.
-	 */
-	R_LOCK(dbenv, &mgr->reginfo);
-
-	if (IS_ZERO_LSN(region->pending_ckp)) {
-		for (txnp =
-		    SH_TAILQ_FIRST(&region->active_txn, __txn_detail);
-		    txnp != NULL;
-		    txnp = SH_TAILQ_NEXT(txnp, links, __txn_detail)) {
-
-			/*
-			 * Look through the active transactions for the
-			 * lowest begin lsn.
-			 */
-			if (!IS_ZERO_LSN(txnp->begin_lsn) &&
-			    log_compare(&txnp->begin_lsn, &ckp_lsn) < 0)
-				ckp_lsn = txnp->begin_lsn;
+			if (now - last_ckp_time >= (time_t)(minutes * 60))
+				goto do_ckp;
 		}
-		region->pending_ckp = ckp_lsn;
-	} else
-		ckp_lsn = region->pending_ckp;
 
+		/*
+		 * If we checked time and data and didn't go to checkpoint,
+		 * we're done.
+		 */
+		if (minutes != 0 || kbytes != 0)
+			return (0);
+	}
+
+do_ckp:	/* Look through the active transactions for the lowest begin LSN. */
+	R_LOCK(dbenv, &mgr->reginfo);
+	for (txnp = SH_TAILQ_FIRST(&region->active_txn, __txn_detail);
+	    txnp != NULL;
+	    txnp = SH_TAILQ_NEXT(txnp, links, __txn_detail))
+		if (!IS_ZERO_LSN(txnp->begin_lsn) &&
+		    log_compare(&txnp->begin_lsn, &ckp_lsn) < 0)
+			ckp_lsn = txnp->begin_lsn;
 	R_UNLOCK(dbenv, &mgr->reginfo);
 
+	if (MPOOL_ON(dbenv) && (ret = dbenv->memp_sync(dbenv, NULL)) != 0) {
+		__db_err(dbenv,
+		    "txn_checkpoint: failed to flush the buffer cache %s",
+		    db_strerror(ret));
+		return (ret);
+	}
+
 	/*
-	 * Try three times to sync the mpool buffers up to the specified LSN,
-	 * sleeping 1, 2 and 4 seconds between attempts.
+	 * Because we can't be a replication client here, and because
+	 * recovery (somewhat unusually) calls txn_checkpoint and expects
+	 * it to write a log message, LOGGING_ON is the correct macro here.
 	 */
-	if (MPOOL_ON(dbenv))
-		for (interval = 1;;) {
-			/*
-			 * memp_sync may change the lsn you pass it, so don't
-			 * pass it the actual ckp_lsn, pass it a local instead.
-			 */
-			sync_lsn = ckp_lsn;
-			if ((ret = dbenv->memp_sync(dbenv, &sync_lsn)) == 0)
-				break;
-
-			/*
-			 * ret == DB_INCOMPLETE means there are still buffers
-			 * to flush, the checkpoint is not complete.
-			 */
-			if (ret == DB_INCOMPLETE) {
-				if (interval > 4)
-					return (ret);
-
-				(void)__os_sleep(dbenv, interval, 0);
-				interval *= 2;
-			} else {
-				__db_err(dbenv,
-		    "txn_checkpoint: failure when flushing the buffer cache %s",
-				    db_strerror(ret));
-				return (ret);
-			}
-		}
-
 	if (LOGGING_ON(dbenv)) {
 		R_LOCK(dbenv, &mgr->reginfo);
 		last_ckp = region->last_ckp;
-		ZERO_LSN(region->pending_ckp);
 		R_UNLOCK(dbenv, &mgr->reginfo);
 
-		if ((ret = __txn_ckp_log(dbenv,
-		    NULL, &ckp_lsn, DB_CHECKPOINT | DB_FLUSH, &ckp_lsn,
+		/*
+		 * Put out records for the open files before we log
+		 * the checkpoint.  The records are certain to be at
+		 * or after ckp_lsn, but before the checkpoint record
+		 * itself, so they're sure to be included if we start
+		 * recovery from the ckp_lsn contained in this
+		 * checkpoint.
+		 */
+		if ((ret = __dbreg_open_files(dbenv)) != 0 ||
+		    (ret = __txn_ckp_log(dbenv,
+		    NULL, &ckp_lsn, DB_PERMANENT | DB_FLUSH, &ckp_lsn,
 		    &last_ckp, (int32_t)time(NULL))) != 0) {
 			__db_err(dbenv,
 			    "txn_checkpoint: log failed at LSN [%ld %ld] %s",
@@ -1165,23 +1209,40 @@ do_ckp:
 			return (ret);
 		}
 
-		R_LOCK(dbenv, &mgr->reginfo);
-
-		/*
-		 * We want to make sure last_ckp only moves forward;  since
-		 * we drop locks above and in log_put, it's possible
-		 * for two calls to __txn_ckp_log to finish in a different
-		 * order from how they were called.
-		 */
-		if (log_compare(&region->last_ckp, &ckp_lsn) < 0) {
-			region->last_ckp = ckp_lsn;
-			(void)time(&region->time_ckp);
-		}
-		R_UNLOCK(dbenv, &mgr->reginfo);
+		__txn_updateckp(dbenv, &ckp_lsn);
 	}
 	return (0);
 }
 
+/*
+ * __txn_getckp --
+ *	Get the LSN of the last transaction checkpoint.
+ *
+ * PUBLIC: int __txn_getckp __P((DB_ENV *, DB_LSN *));
+ */
+int
+__txn_getckp(dbenv, lsnp)
+	DB_ENV *dbenv;
+	DB_LSN *lsnp;
+{
+	DB_LSN lsn;
+	DB_TXNMGR *mgr;
+	DB_TXNREGION *region;
+
+	mgr = dbenv->tx_handle;
+	region = mgr->reginfo.primary;
+
+	R_LOCK(dbenv, &mgr->reginfo);
+	lsn = region->last_ckp;
+	R_UNLOCK(dbenv, &mgr->reginfo);
+
+	if (IS_ZERO_LSN(lsn))
+		return (DB_NOTFOUND);
+
+	*lsnp = lsn;
+	return (0);
+}
+
 /*
  * __txn_activekids --
  *	Return if this transaction has any active children.
@@ -1198,7 +1259,7 @@ __txn_activekids(dbenv, rectype, txnp)
 	 * On a child commit, we know that there are children (i.e., the
 	 * commiting child at the least.  In that case, skip this check.
 	 */
-	if (rectype == DB_txn_child)
+	if (F_ISSET(txnp, TXN_COMPENSATE) || rectype == DB___txn_child)
 		return (0);
 
 	if (TAILQ_FIRST(&txnp->kids) != NULL) {
@@ -1213,30 +1274,61 @@ __txn_activekids(dbenv, rectype, txnp)
  *	Force an abort record into the log if the commit record
  *	failed to get to disk.
  *
- * PUBLIC: void __txn_force_abort __P((u_int8_t *));
+ * PUBLIC: int __txn_force_abort __P((DB_ENV *, u_int8_t *));
  */
-void
-__txn_force_abort(buffer)
+int
+__txn_force_abort(dbenv, buffer)
+	DB_ENV *dbenv;
 	u_int8_t *buffer;
 {
-	u_int32_t cksum, offset, opcode;
-	u_int8_t *bp;
+	DB_CIPHER *db_cipher;
+	HDR *hdr;
+	u_int32_t offset, opcode, rec_len, rec_type, sum_len;
+	u_int8_t *bp, *key, chksum[DB_MAC_KEY];
+	size_t hdrsize;
+	int ret;
+
+	db_cipher = dbenv->crypto_handle;
 
 	/*
 	 * This routine depends on the layout of HDR and the __txn_regop
 	 * record in txn.src.  We are passed the beginning of the commit
 	 * record in the log buffer and overwrite the commit with an abort
-	 * and recalculate the checksum.
-	 *
-	 *       rectype		   txn_num		lsn
+	 * and recalculate the checksum.  We may be passed a txn_xa_regop
+	 * that is, an XA prepare), there's no need to overwrite that one.
 	 */
+	hdr = (HDR *)buffer;
+	memcpy(&rec_type, hdr, sizeof(rec_type));
+	if (rec_type == DB___txn_xa_regop)
+		return (0);
+
 	offset = sizeof(u_int32_t) + sizeof(u_int32_t) + sizeof(DB_LSN);
-	bp = buffer + sizeof(HDR) + offset;
+	rec_len = offset + sizeof(u_int32_t) + sizeof(int32_t);
+	if (CRYPTO_ON(dbenv)) {
+		key = db_cipher->mac_key;
+		hdrsize = HDR_CRYPTO_SZ;
+		sum_len = DB_MAC_KEY;
+		if ((ret = db_cipher->decrypt(dbenv, db_cipher->data,
+		    &hdr->iv[0], buffer + hdrsize, rec_len)) != 0)
+			return (__db_panic(dbenv, ret));
+	} else {
+		key = NULL;
+		hdrsize = HDR_NORMAL_SZ;
+		sum_len = sizeof(u_int32_t);
+	}
+	bp = buffer + hdrsize + offset;
 	opcode = TXN_ABORT;
 	memcpy(bp, &opcode, sizeof(opcode));
-	cksum = __ham_func4(NULL, buffer + sizeof(HDR),
-	    offset + sizeof(u_int32_t) + sizeof(int32_t));
-	memcpy(buffer + SSZ(HDR, cksum), &cksum, sizeof(cksum));
+
+	if (CRYPTO_ON(dbenv) &&
+	    (ret = db_cipher->encrypt(dbenv,
+	    db_cipher->data, &hdr->iv[0], buffer + hdrsize, rec_len)) != 0)
+		return (__db_panic(dbenv, ret));
+
+	__db_chksum(buffer + hdrsize, rec_len, key, chksum);
+	memcpy(buffer + SSZ(HDR, chksum), &chksum, sum_len);
+
+	return (0);
 }
 
 /*
@@ -1245,15 +1337,15 @@ __txn_force_abort(buffer)
  * were in the midst of taking care of restored transactions.  If
  * so, then we need to close the files that we opened.
  *
- * PUBLIC: void __txn_preclose __P((DB_ENV *));
+ * PUBLIC: int __txn_preclose __P((DB_ENV *));
  */
-void
+int
 __txn_preclose(dbenv)
 	DB_ENV *dbenv;
 {
 	DB_TXNMGR *mgr;
 	DB_TXNREGION *region;
-	int do_closefiles;
+	int do_closefiles, ret;
 
 	mgr = (DB_TXNMGR *)dbenv->tx_handle;
 	region = mgr->reginfo.primary;
@@ -1273,7 +1365,64 @@ __txn_preclose(dbenv)
 		 * that will confuse future recoveries.
 		 */
 		F_SET((DB_LOG *)dbenv->lg_handle, DBLOG_RECOVER);
-		__log_close_files(dbenv);
+		ret = __dbreg_close_files(dbenv);
 		F_CLR((DB_LOG *)dbenv->lg_handle, DBLOG_RECOVER);
+	} else
+		ret = 0;
+
+	return (ret);
+}
+
+/*
+ * __txn_reset --
+ *	Reset the last txnid to its minimum value, and log the reset.
+ *
+ * PUBLIC: int __txn_reset __P((DB_ENV *));
+ */
+int
+__txn_reset(dbenv)
+	DB_ENV *dbenv;
+{
+	DB_LSN scrap;
+	DB_TXNREGION *region;
+
+	region = ((DB_TXNMGR *)dbenv->tx_handle)->reginfo.primary;
+	region->last_txnid = TXN_MINIMUM;
+
+	DB_ASSERT(LOGGING_ON(dbenv));
+	return (__txn_recycle_log(dbenv,
+	    NULL, &scrap, 0, TXN_MINIMUM, TXN_MAXIMUM));
+}
+
+/*
+ * __txn_updateckp --
+ *	Update the last_ckp field in the transaction region.  This happens
+ * at the end of a normal checkpoint and also when a replication client
+ * receives a checkpoint record.
+ *
+ * PUBLIC: void __txn_updateckp __P((DB_ENV *, DB_LSN *));
+ */
+void
+__txn_updateckp(dbenv, lsnp)
+	DB_ENV *dbenv;
+	DB_LSN *lsnp;
+{
+	DB_TXNMGR *mgr;
+	DB_TXNREGION *region;
+
+	mgr = dbenv->tx_handle;
+	region = mgr->reginfo.primary;
+
+	/*
+	 * We want to make sure last_ckp only moves forward;  since
+	 * we drop locks above and in log_put, it's possible
+	 * for two calls to __txn_ckp_log to finish in a different
+	 * order from how they were called.
+	 */
+	R_LOCK(dbenv, &mgr->reginfo);
+	if (log_compare(&region->last_ckp, lsnp) < 0) {
+		region->last_ckp = *lsnp;
+		(void)time(&region->time_ckp);
 	}
+	R_UNLOCK(dbenv, &mgr->reginfo);
 }
