@@ -6,8 +6,13 @@
 #include "rpmio_internal.h"
 #include <rpmlib.h>
 #include <rpmmacro.h>	/* XXX for %_i18ndomains */
+
+#include <rpmfi.h>
+
+#include "legacy.h"
 #include "manifest.h"
 #include "misc.h"
+
 #include "debug.h"
 
 /*@access pgpDig @*/
@@ -32,10 +37,16 @@ static /*@only@*/ char * triggertypeFormat(int_32 type, const void * data,
 
     if (type != RPM_INT32_TYPE)
 	val = xstrdup(_("(not a number)"));
+    else if (*item & RPMSENSE_TRIGGERPREIN)
+	val = xstrdup("prein");
     else if (*item & RPMSENSE_TRIGGERIN)
 	val = xstrdup("in");
-    else
+    else if (*item & RPMSENSE_TRIGGERUN)
 	val = xstrdup("un");
+    else if (*item & RPMSENSE_TRIGGERPOSTUN)
+	val = xstrdup("postun");
+    else
+	val = xstrdup("");
     return val;
 }
 
@@ -131,25 +142,25 @@ static /*@only@*/ char * fflagsFormat(int_32 type, const void * data,
  * @todo Permit selectable display formats (i.e. binary).
  * @param type		tag type
  * @param data		tag value
- * @param formatPrefix
- * @param padding
- * @param element	(unused)
+ * @param formatPrefix	(unused)
+ * @param padding	(unused)
+ * @param element	no. bytes of binary data
  * @return		formatted string
  */
 static /*@only@*/ char * armorFormat(int_32 type, const void * data, 
-		/*@unused@*/ char * formatPrefix, int padding, int element)
+		/*@unused@*/ char * formatPrefix, /*@unused@*/ int padding,
+		int element)
 	/*@*/
 {
     const char * enc;
-    const char * s;
-    char * t;
-    char * val;
+    const unsigned char * s;
+    size_t ns;
     int atype;
-    int lc, ns, nt;
 
     switch (type) {
     case RPM_BIN_TYPE:
 	s = data;
+	/* XXX HACK ALERT: element field abused as no. bytes of binary data. */
 	ns = element;
 	atype = PGPARMOR_SIGNATURE;	/* XXX check pkt for signature */
 	break;
@@ -171,49 +182,8 @@ static /*@only@*/ char * armorFormat(int_32 type, const void * data,
 	/*@notreached@*/ break;
     }
 
-    nt = ((ns + 2) / 3) * 4;
-    /*@-globs@*/
-    /* Add additional bytes necessary for eol string(s). */
-    if (b64encode_chars_per_line > 0 && b64encode_eolstr != NULL) {
-	lc = (nt + b64encode_chars_per_line - 1) / b64encode_chars_per_line;
-       if (((nt + b64encode_chars_per_line - 1) % b64encode_chars_per_line) != 0)
-        ++lc;
-	nt += lc * strlen(b64encode_eolstr);
-    }
-    /*@=globs@*/
-
-    nt += 512;	/* XXX slop for armor and crc */
-
-/*@-boundswrite@*/
-    val = t = xmalloc(nt + padding + 1);
-    *t = '\0';
-    t = stpcpy(t, "-----BEGIN PGP ");
-    t = stpcpy(t, pgpValStr(pgpArmorTbl, atype));
-    /*@-globs@*/
-    t = stpcpy( stpcpy(t, "-----\nVersion: rpm-"), RPMVERSION);
-    /*@=globs@*/
-    t = stpcpy(t, " (beecrypt-2.2.0)\n\n");
-
-    if ((enc = b64encode(s, ns)) != NULL) {
-	t = stpcpy(t, enc);
-	enc = _free(enc);
-	if ((enc = b64crc(s, ns)) != NULL) {
-	    *t++ = '=';
-	    t = stpcpy(t, enc);
-	    enc = _free(enc);
-	}
-    }
-	
-    t = stpcpy(t, "-----END PGP ");
-    t = stpcpy(t, pgpValStr(pgpArmorTbl, atype));
-    t = stpcpy(t, "-----\n");
-/*@=boundswrite@*/
-
-    /*@-branchstate@*/
-    if (s != data) s = _free(s);
-    /*@=branchstate@*/
-
-    return val;
+    /* XXX this doesn't use padding directly, assumes enough slop in retval. */
+    return pgpArmorWrap(atype, s, ns);
 }
 
 /**
@@ -221,9 +191,9 @@ static /*@only@*/ char * armorFormat(int_32 type, const void * data,
  * @todo Permit selectable display formats (i.e. binary).
  * @param type		tag type
  * @param data		tag value
- * @param formatPrefix
+ * @param formatPrefix	(unused)
  * @param padding
- * @param element	(unused)
+ * @param element
  * @return		formatted string
  */
 static /*@only@*/ char * base64Format(int_32 type, const void * data, 
@@ -238,7 +208,9 @@ static /*@only@*/ char * base64Format(int_32 type, const void * data,
 	const char * enc;
 	char * t;
 	int lc;
-	int nt = ((element + 2) / 3) * 4;
+	/* XXX HACK ALERT: element field abused as no. bytes of binary data. */
+	size_t ns = element;
+	size_t nt = ((ns + 2) / 3) * 4;
 
 /*@-boundswrite@*/
 	/*@-globs@*/
@@ -254,7 +226,7 @@ static /*@only@*/ char * base64Format(int_32 type, const void * data,
 	val = t = xmalloc(nt + padding + 1);
 
 	*t = '\0';
-	if ((enc = b64encode(data, element)) != NULL) {
+	if ((enc = b64encode(data, ns)) != NULL) {
 	    t = stpcpy(t, enc);
 	    enc = _free(enc);
 	}
@@ -265,7 +237,50 @@ static /*@only@*/ char * base64Format(int_32 type, const void * data,
 }
 
 /**
- * Display signature fingerprint and time.
+ */
+static size_t xmlstrlen(const char * s)
+	/*@*/
+{
+    size_t len = 0;
+    int c;
+
+/*@-boundsread@*/
+    while ((c = *s++) != '\0')
+/*@=boundsread@*/
+    {
+	switch (c) {
+	case '<': case '>':	len += 4;	/*@switchbreak@*/ break;
+	case '&':		len += 5;	/*@switchbreak@*/ break;
+	default:		len += 1;	/*@switchbreak@*/ break;
+	}
+    }
+    return len;
+}
+
+/**
+ */
+static char * xmlstrcpy(/*@returned@*/ char * t, const char * s)
+	/*@modifies t @*/
+{
+    char * te = t;
+    int c;
+
+/*@-bounds@*/
+    while ((c = *s++) != '\0') {
+	switch (c) {
+	case '<':	te = stpcpy(te, "&lt;");	/*@switchbreak@*/ break;
+	case '>':	te = stpcpy(te, "&gt;");	/*@switchbreak@*/ break;
+	case '&':	te = stpcpy(te, "&amp;");	/*@switchbreak@*/ break;
+	default:	*te++ = c;			/*@switchbreak@*/ break;
+	}
+    }
+    *te = '\0';
+/*@=bounds@*/
+    return t;
+}
+
+/**
+ * Wrap tag data in simple header xml markup.
  * @param type		tag type
  * @param data		tag value
  * @param formatPrefix
@@ -273,11 +288,109 @@ static /*@only@*/ char * base64Format(int_32 type, const void * data,
  * @param element	(unused)
  * @return		formatted string
  */
+/*@-bounds@*/
+static /*@only@*/ char * xmlFormat(int_32 type, const void * data, 
+		char * formatPrefix, int padding,
+		/*@unused@*/ int element)
+	/*@modifies formatPrefix @*/
+{
+    const char * xtag = NULL;
+    size_t nb;
+    char * val;
+    const char * s = NULL;
+    char * t, * te;
+    unsigned long anint = 0;
+    int xx;
+
+/*@-branchstate@*/
+    switch (type) {
+    case RPM_I18NSTRING_TYPE:
+    case RPM_STRING_TYPE:
+	s = data;
+	xtag = "string";
+	break;
+    case RPM_BIN_TYPE:
+    {	int cpl = b64encode_chars_per_line;
+/*@-mods@*/
+	b64encode_chars_per_line = 0;
+/*@=mods@*/
+/*@-formatconst@*/
+	s = base64Format(type, data, formatPrefix, padding, element);
+/*@=formatconst@*/
+/*@-mods@*/
+	b64encode_chars_per_line = cpl;
+/*@=mods@*/
+	xtag = "base64";
+    }	break;
+    case RPM_CHAR_TYPE:
+    case RPM_INT8_TYPE:
+	anint = *((uint_8 *) data);
+	break;
+    case RPM_INT16_TYPE:
+	anint = *((uint_16 *) data);
+	break;
+    case RPM_INT32_TYPE:
+	anint = *((uint_32 *) data);
+	break;
+    case RPM_NULL_TYPE:
+    case RPM_STRING_ARRAY_TYPE:
+    default:
+	return xstrdup(_("(invalid xml type)"));
+	/*@notreached@*/ break;
+    }
+/*@=branchstate@*/
+
+/*@-branchstate@*/
+    if (s == NULL) {
+	int tlen = 32;
+	t = memset(alloca(tlen+1), 0, tlen+1);
+	xx = snprintf(t, tlen, "%lu", anint);
+	s = t;
+	xtag = "integer";
+    }
+/*@=branchstate@*/
+
+    nb = 2 * strlen(xtag) + sizeof("\t<></>") + xmlstrlen(s);
+    te = t = alloca(nb);
+    te = stpcpy( stpcpy( stpcpy(te, "\t<"), xtag), ">");
+    te = xmlstrcpy(te, s);
+    te += strlen(te);
+    te = stpcpy( stpcpy( stpcpy(te, "</"), xtag), ">");
+
+    /* XXX s was malloc'd */
+/*@-branchstate@*/
+    if (!strcmp(xtag, "base64"))
+	s = _free(s);
+/*@=branchstate@*/
+
+    nb += padding;
+    val = xmalloc(nb+1);
+/*@-boundswrite@*/
+    strcat(formatPrefix, "s");
+/*@=boundswrite@*/
+/*@-formatconst@*/
+    xx = snprintf(val, nb, formatPrefix, t);
+/*@=formatconst@*/
+    val[nb] = '\0';
+
+    return val;
+}
+/*@=bounds@*/
+
+/**
+ * Display signature fingerprint and time.
+ * @param type		tag type
+ * @param data		tag value
+ * @param formatPrefix	(unused)
+ * @param padding
+ * @param element	(unused)
+ * @return		formatted string
+ */
 static /*@only@*/ char * pgpsigFormat(int_32 type, const void * data, 
 		/*@unused@*/ char * formatPrefix, /*@unused@*/ int padding,
 		/*@unused@*/ int element)
-	/*@globals fileSystem @*/
-	/*@modifies fileSystem @*/
+	/*@globals fileSystem, internalState @*/
+	/*@modifies fileSystem, internalState @*/
 {
     char * val, * t;
 
@@ -380,11 +493,12 @@ static /*@only@*/ char * depflagsFormat(int_32 type, const void * data,
 {
     char * val;
     char buf[10];
-    int anint = *((int_32 *) data);
+    int anint;
 
     if (type != RPM_INT32_TYPE) {
 	val = xstrdup(_("(not a number)"));
     } else {
+	anint = *((int_32 *) data);
 	buf[0] = '\0';
 
 /*@-boundswrite@*/
@@ -411,10 +525,10 @@ static /*@only@*/ char * depflagsFormat(int_32 type, const void * data,
 /**
  * Retrieve mounted file system paths.
  * @param h		header
- * @retval type		address of tag type
- * @retval data		address of tag value pointer
- * @retval count	address of no. of data items
- * @retval freeData	address of data-was-malloc'ed indicator
+ * @retval *type	tag type
+ * @retval *data	tag value
+ * @retval *count	no. of data items
+ * @retval *freeData	data-was-malloc'ed indicator
  * @return		0 on success
  */
 static int fsnamesTag( /*@unused@*/ Header h, /*@out@*/ int_32 * type,
@@ -443,10 +557,10 @@ static int fsnamesTag( /*@unused@*/ Header h, /*@out@*/ int_32 * type,
 /**
  * Retrieve install prefixes.
  * @param h		header
- * @retval type		address of tag type
- * @retval data		address of tag value pointer
- * @retval count	address of no. of data items
- * @retval freeData	address of data-was-malloc'ed indicator
+ * @retval *type	tag type
+ * @retval *data	tag value
+ * @retval *count	no. of data items
+ * @retval *freeData	data-was-malloc'ed indicator
  * @return		0 on success
  */
 static int instprefixTag(Header h, /*@null@*/ /*@out@*/ rpmTagType * type,
@@ -481,16 +595,16 @@ static int instprefixTag(Header h, /*@null@*/ /*@out@*/ rpmTagType * type,
 /**
  * Retrieve mounted file system space.
  * @param h		header
- * @retval type		address of tag type
- * @retval data		address of tag value pointer
- * @retval count	address of no. of data items
- * @retval freeData	address of data-was-malloc'ed indicator
+ * @retval *type	tag type
+ * @retval *data	tag value
+ * @retval *count	no. of data items
+ * @retval *freeData	data-was-malloc'ed indicator
  * @return		0 on success
  */
 static int fssizesTag(Header h, /*@out@*/ rpmTagType * type,
 		/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
 		/*@out@*/ int * freeData)
-	/*@globals rpmGlobalMacroContext,
+	/*@globals rpmGlobalMacroContext, h_errno,
 		fileSystem, internalState @*/
 	/*@modifies *type, *data, *count, *freeData, rpmGlobalMacroContext,
 		fileSystem, internalState @*/
@@ -508,7 +622,7 @@ static int fssizesTag(Header h, /*@out@*/ rpmTagType * type,
 	numFiles = 0;
 	filenames = NULL;
     } else {
-	rpmBuildFileList(h, &filenames, &numFiles);
+	rpmfiBuildFNames(h, RPMTAG_BASENAMES, &filenames, &numFiles);
     }
 
 /*@-boundswrite@*/
@@ -541,13 +655,12 @@ static int fssizesTag(Header h, /*@out@*/ rpmTagType * type,
 /**
  * Retrieve trigger info.
  * @param h		header
- * @retval type		address of tag type
- * @retval data		address of tag value pointer
- * @retval count	address of no. of data items
- * @retval freeData	address of data-was-malloc'ed indicator
+ * @retval *type	tag type
+ * @retval *data	tag value
+ * @retval *count	no. of data items
+ * @retval *freeData	data-was-malloc'ed indicator
  * @return		0 on success
  */
-/*@-bounds@*/	/* LCL: segfault */
 static int triggercondsTag(Header h, /*@out@*/ rpmTagType * type,
 		/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
 		/*@out@*/ int * freeData)
@@ -582,6 +695,7 @@ static int triggercondsTag(Header h, /*@out@*/ rpmTagType * type,
     *data = conds = xmalloc(sizeof(*conds) * numScripts);
     *count = numScripts;
     *type = RPM_STRING_ARRAY_TYPE;
+/*@-bounds@*/
     for (i = 0; i < numScripts; i++) {
 	chptr = xstrdup("");
 
@@ -607,21 +721,21 @@ static int triggercondsTag(Header h, /*@out@*/ rpmTagType * type,
 
 	conds[i] = chptr;
     }
+/*@=bounds@*/
 
     names = hfd(names, tnt);
     versions = hfd(versions, tvt);
 
     return 0;
 }
-/*@=bounds@*/
 
 /**
  * Retrieve trigger type info.
  * @param h		header
- * @retval type		address of tag type
- * @retval data		address of tag value pointer
- * @retval count	address of no. of data items
- * @retval freeData	address of data-was-malloc'ed indicator
+ * @retval *type	tag type
+ * @retval *data	tag value
+ * @retval *count	no. of data items
+ * @retval *freeData	data-was-malloc'ed indicator
  * @return		0 on success
  */
 static int triggertypeTag(Header h, /*@out@*/ rpmTagType * type,
@@ -653,20 +767,26 @@ static int triggertypeTag(Header h, /*@out@*/ rpmTagType * type,
     *data = conds = xmalloc(sizeof(*conds) * numScripts);
     *count = numScripts;
     *type = RPM_STRING_ARRAY_TYPE;
+/*@-bounds@*/
     for (i = 0; i < numScripts; i++) {
 	for (j = 0; j < numNames; j++) {
 	    if (indices[j] != i)
 		/*@innercontinue@*/ continue;
 
-	    if (flags[j] & RPMSENSE_TRIGGERIN)
+	    if (flags[j] & RPMSENSE_TRIGGERPREIN)
+		conds[i] = xstrdup("prein");
+	    else if (flags[j] & RPMSENSE_TRIGGERIN)
 		conds[i] = xstrdup("in");
 	    else if (flags[j] & RPMSENSE_TRIGGERUN)
 		conds[i] = xstrdup("un");
-	    else
+	    else if (flags[j] & RPMSENSE_TRIGGERPOSTUN)
 		conds[i] = xstrdup("postun");
+	    else
+		conds[i] = xstrdup("");
 	    /*@innerbreak@*/ break;
 	}
     }
+/*@=bounds@*/
 
     return 0;
 }
@@ -674,10 +794,10 @@ static int triggertypeTag(Header h, /*@out@*/ rpmTagType * type,
 /**
  * Retrieve file paths.
  * @param h		header
- * @retval type		address of tag type
- * @retval data		address of tag value pointer
- * @retval count	address of no. of data items
- * @retval freeData	address of data-was-malloc'ed indicator
+ * @retval *type	tag type
+ * @retval *data	tag value
+ * @retval *count	no. of data items
+ * @retval *freeData	data-was-malloc'ed indicator
  * @return		0 on success
  */
 static int filenamesTag(Header h, /*@out@*/ rpmTagType * type,
@@ -689,7 +809,7 @@ static int filenamesTag(Header h, /*@out@*/ rpmTagType * type,
 {
     *type = RPM_STRING_ARRAY_TYPE;
 
-    rpmBuildFileList(h, (const char ***) data, count);
+    rpmfiBuildFNames(h, RPMTAG_BASENAMES, (const char ***) data, count);
     *freeData = 1;
 
     *freeData = 0;	/* XXX WTFO? */
@@ -697,6 +817,75 @@ static int filenamesTag(Header h, /*@out@*/ rpmTagType * type,
     return 0; 
 }
 
+/**
+ * Retrieve file classes.
+ * @param h		header
+ * @retval *type	tag type
+ * @retval *data	tag value
+ * @retval *count	no. of data items
+ * @retval *freeData	data-was-malloc'ed indicator
+ * @return		0 on success
+ */
+static int fileclassTag(Header h, /*@out@*/ rpmTagType * type,
+		/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
+		/*@out@*/ int * freeData)
+	/*@globals fileSystem @*/
+	/*@modifies h, *type, *data, *count, *freeData, fileSystem @*/
+	/*@requires maxSet(type) >= 0 /\ maxSet(data) >= 0
+		/\ maxSet(count) >= 0 /\ maxSet(freeData) >= 0 @*/
+{
+    *type = RPM_STRING_ARRAY_TYPE;
+    rpmfiBuildFClasses(h, (const char ***) data, count);
+    *freeData = 1;
+    return 0; 
+}
+
+/**
+ * Retrieve file provides.
+ * @param h		header
+ * @retval *type	tag type
+ * @retval *data	tag value
+ * @retval *count	no. of data items
+ * @retval *freeData	data-was-malloc'ed indicator
+ * @return		0 on success
+ */
+static int fileprovideTag(Header h, /*@out@*/ rpmTagType * type,
+		/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
+		/*@out@*/ int * freeData)
+	/*@globals fileSystem @*/
+	/*@modifies h, *type, *data, *count, *freeData, fileSystem @*/
+	/*@requires maxSet(type) >= 0 /\ maxSet(data) >= 0
+		/\ maxSet(count) >= 0 /\ maxSet(freeData) >= 0 @*/
+{
+    *type = RPM_STRING_ARRAY_TYPE;
+    rpmfiBuildFDeps(h, RPMTAG_PROVIDENAME, (const char ***) data, count);
+    *freeData = 1;
+    return 0; 
+}
+
+/**
+ * Retrieve file requires.
+ * @param h		header
+ * @retval *type	tag type
+ * @retval *data	tag value
+ * @retval *count	no. of data items
+ * @retval *freeData	data-was-malloc'ed indicator
+ * @return		0 on success
+ */
+static int filerequireTag(Header h, /*@out@*/ rpmTagType * type,
+		/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
+		/*@out@*/ int * freeData)
+	/*@globals fileSystem @*/
+	/*@modifies h, *type, *data, *count, *freeData, fileSystem @*/
+	/*@requires maxSet(type) >= 0 /\ maxSet(data) >= 0
+		/\ maxSet(count) >= 0 /\ maxSet(freeData) >= 0 @*/
+{
+    *type = RPM_STRING_ARRAY_TYPE;
+    rpmfiBuildFDeps(h, RPMTAG_REQUIRENAME, (const char ***) data, count);
+    *freeData = 1;
+    return 0; 
+}
+
 /* I18N look aside diversions */
 
 /*@-exportlocal -exportheadervar@*/
@@ -707,23 +896,22 @@ int _nl_msg_cat_cntr;	/* XXX GNU gettext voodoo */
 static const char * language = "LANGUAGE";
 
 /*@observer@*/ /*@unchecked@*/
-static const char * _macro_i18ndomains =
-		"%{?_i18ndomains:%{_i18ndomains}}";
+static const char * _macro_i18ndomains = "%{?_i18ndomains}";
 
 /**
  * Retrieve i18n text.
  * @param h		header
  * @param tag		tag
- * @retval type		address of tag type
- * @retval data		address of tag value pointer
- * @retval count	address of no. of data items
- * @retval freeData	address of data-was-malloc'ed indicator
+ * @retval *type	tag type
+ * @retval *data	tag value
+ * @retval *count	no. of data items
+ * @retval *freeData	data-was-malloc'ed indicator
  * @return		0 on success
  */
 static int i18nTag(Header h, int_32 tag, /*@out@*/ rpmTagType * type,
 		/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
 		/*@out@*/ int * freeData)
-	/*@globals rpmGlobalMacroContext @*/
+	/*@globals rpmGlobalMacroContext, h_errno @*/
 	/*@modifies *type, *data, *count, *freeData, rpmGlobalMacroContext @*/
 	/*@requires maxSet(type) >= 0 /\ maxSet(data) >= 0
 		/\ maxSet(count) >= 0 /\ maxSet(freeData) >= 0 @*/
@@ -804,16 +992,16 @@ static int i18nTag(Header h, int_32 tag, /*@out@*/ rpmTagType * type,
 /**
  * Retrieve summary text.
  * @param h		header
- * @retval type		address of tag type
- * @retval data		address of tag value pointer
- * @retval count	address of no. of data items
- * @retval freeData	address of data-was-malloc'ed indicator
+ * @retval *type	tag type
+ * @retval *data	tag value
+ * @retval *count	no. of data items
+ * @retval *freeData	data-was-malloc'ed indicator
  * @return		0 on success
  */
 static int summaryTag(Header h, /*@out@*/ rpmTagType * type,
 		/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
 		/*@out@*/ int * freeData)
-	/*@globals rpmGlobalMacroContext @*/
+	/*@globals rpmGlobalMacroContext, h_errno @*/
 	/*@modifies *type, *data, *count, *freeData, rpmGlobalMacroContext @*/
 	/*@requires maxSet(type) >= 0 /\ maxSet(data) >= 0
 		/\ maxSet(count) >= 0 /\ maxSet(freeData) >= 0 @*/
@@ -824,16 +1012,16 @@ static int summaryTag(Header h, /*@out@*/ rpmTagType * type,
 /**
  * Retrieve description text.
  * @param h		header
- * @retval type		address of tag type
- * @retval data		address of tag value pointer
- * @retval count	address of no. of data items
- * @retval freeData	address of data-was-malloc'ed indicator
+ * @retval *type	tag type
+ * @retval *data	tag value
+ * @retval *count	no. of data items
+ * @retval *freeData	data-was-malloc'ed indicator
  * @return		0 on success
  */
 static int descriptionTag(Header h, /*@out@*/ rpmTagType * type,
 		/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
 		/*@out@*/ int * freeData)
-	/*@globals rpmGlobalMacroContext @*/
+	/*@globals rpmGlobalMacroContext, h_errno @*/
 	/*@modifies *type, *data, *count, *freeData, rpmGlobalMacroContext @*/
 	/*@requires maxSet(type) >= 0 /\ maxSet(data) >= 0
 		/\ maxSet(count) >= 0 /\ maxSet(freeData) >= 0 @*/
@@ -844,16 +1032,16 @@ static int descriptionTag(Header h, /*@out@*/ rpmTagType * type,
 /**
  * Retrieve group text.
  * @param h		header
- * @retval type		address of tag type
- * @retval data		address of tag value pointer
- * @retval count	address of no. of data items
- * @retval freeData	address of data-was-malloc'ed indicator
+ * @retval *type	tag type
+ * @retval *data	tag value
+ * @retval *count	no. of data items
+ * @retval *freeData	data-was-malloc'ed indicator
  * @return		0 on success
  */
 static int groupTag(Header h, /*@out@*/ rpmTagType * type,
 		/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
 		/*@out@*/ int * freeData)
-	/*@globals rpmGlobalMacroContext @*/
+	/*@globals rpmGlobalMacroContext, h_errno @*/
 	/*@modifies *type, *data, *count, *freeData, rpmGlobalMacroContext @*/
 	/*@requires maxSet(type) >= 0 /\ maxSet(data) >= 0
 		/\ maxSet(count) >= 0 /\ maxSet(freeData) >= 0 @*/
@@ -863,23 +1051,27 @@ static int groupTag(Header h, /*@out@*/ rpmTagType * type,
 
 /*@-type@*/ /* FIX: cast? */
 const struct headerSprintfExtension_s rpmHeaderFormats[] = {
-    { HEADER_EXT_TAG, "RPMTAG_GROUP", { groupTag } },
-    { HEADER_EXT_TAG, "RPMTAG_DESCRIPTION", { descriptionTag } },
-    { HEADER_EXT_TAG, "RPMTAG_SUMMARY", { summaryTag } },
-    { HEADER_EXT_TAG, "RPMTAG_FILENAMES", { filenamesTag } },
-    { HEADER_EXT_TAG, "RPMTAG_FSSIZES", { fssizesTag } },
-    { HEADER_EXT_TAG, "RPMTAG_FSNAMES", { fsnamesTag } },
-    { HEADER_EXT_TAG, "RPMTAG_INSTALLPREFIX", { instprefixTag } },
-    { HEADER_EXT_TAG, "RPMTAG_TRIGGERCONDS", { triggercondsTag } },
-    { HEADER_EXT_TAG, "RPMTAG_TRIGGERTYPE", { triggertypeTag } },
-    { HEADER_EXT_FORMAT, "armor", { armorFormat } },
-    { HEADER_EXT_FORMAT, "base64", { base64Format } },
-    { HEADER_EXT_FORMAT, "pgpsig", { pgpsigFormat } },
-    { HEADER_EXT_FORMAT, "depflags", { depflagsFormat } },
-    { HEADER_EXT_FORMAT, "fflags", { fflagsFormat } },
-    { HEADER_EXT_FORMAT, "perms", { permsFormat } },
-    { HEADER_EXT_FORMAT, "permissions", { permsFormat } },
-    { HEADER_EXT_FORMAT, "triggertype", { triggertypeFormat } },
-    { HEADER_EXT_MORE, NULL, { (void *) headerDefaultFormats } }
+    { HEADER_EXT_TAG, "RPMTAG_GROUP",		{ groupTag } },
+    { HEADER_EXT_TAG, "RPMTAG_DESCRIPTION",	{ descriptionTag } },
+    { HEADER_EXT_TAG, "RPMTAG_SUMMARY",		{ summaryTag } },
+    { HEADER_EXT_TAG, "RPMTAG_FILECLASS",	{ fileclassTag } },
+    { HEADER_EXT_TAG, "RPMTAG_FILENAMES",	{ filenamesTag } },
+    { HEADER_EXT_TAG, "RPMTAG_FILEPROVIDE",	{ fileprovideTag } },
+    { HEADER_EXT_TAG, "RPMTAG_FILEREQUIRE",	{ filerequireTag } },
+    { HEADER_EXT_TAG, "RPMTAG_FSSIZES",		{ fssizesTag } },
+    { HEADER_EXT_TAG, "RPMTAG_FSNAMES",		{ fsnamesTag } },
+    { HEADER_EXT_TAG, "RPMTAG_INSTALLPREFIX",	{ instprefixTag } },
+    { HEADER_EXT_TAG, "RPMTAG_TRIGGERCONDS",	{ triggercondsTag } },
+    { HEADER_EXT_TAG, "RPMTAG_TRIGGERTYPE",	{ triggertypeTag } },
+    { HEADER_EXT_FORMAT, "armor",		{ armorFormat } },
+    { HEADER_EXT_FORMAT, "base64",		{ base64Format } },
+    { HEADER_EXT_FORMAT, "pgpsig",		{ pgpsigFormat } },
+    { HEADER_EXT_FORMAT, "depflags",		{ depflagsFormat } },
+    { HEADER_EXT_FORMAT, "fflags",		{ fflagsFormat } },
+    { HEADER_EXT_FORMAT, "perms",		{ permsFormat } },
+    { HEADER_EXT_FORMAT, "permissions",		{ permsFormat } },
+    { HEADER_EXT_FORMAT, "triggertype",		{ triggertypeFormat } },
+    { HEADER_EXT_FORMAT, "xml",			{ xmlFormat } },
+    { HEADER_EXT_MORE, NULL,		{ (void *) headerDefaultFormats } }
 } ;
 /*@=type@*/
