@@ -12,7 +12,7 @@
 #include "rpmts.h"
 
 #include "misc.h"	/* XXX stripTrailingChar() */
-#include "legacy.h"	/* XXX providePackageNVR() and compressFileList() */
+#include "legacy.h"	/* XXX legacyRetrofit() */
 #include "rpmlead.h"
 
 #include "header_internal.h"	/* XXX headerCheck */
@@ -23,14 +23,13 @@
 
 /*@access pgpDig @*/
 /*@access pgpDigParams @*/
-/*@access rpmts @*/
 /*@access Header @*/		/* XXX compared with NULL */
 /*@access entryInfo @*/		/* XXX headerCheck */
 /*@access indexEntry @*/	/* XXX headerCheck */
 /*@access FD_t @*/		/* XXX stealing digests */
 
 /*@unchecked@*/
-static int _print_pkts = 0;
+int _print_pkts = 0;
 
 /*@unchecked@*/
 static unsigned int nkeyids_max = 256;
@@ -360,7 +359,7 @@ rpmRC headerCheck(rpmts ts, const void * uh, size_t uc, const char ** msg)
     if (uc > 0 && pvlen != uc) {
 	(void) snprintf(buf, sizeof(buf),
 		_("blob size(%d): BAD, 8 + 16 * il(%d) + dl(%d)\n"),
-		uc, il, dl);
+		(int)uc, (int)il, (int)dl);
 	goto exit;
     }
 
@@ -610,6 +609,7 @@ verifyinfo_exit:
 
 	break;
     default:
+	sig = _free(sig);
 	break;
     }
 
@@ -624,6 +624,8 @@ verifyinfo_exit:
 /*@=boundswrite@*/
 
     rpmtsCleanDig(ts);
+    if (info->tag == RPMTAG_SHA1HEADER)
+	sig = _free(sig);
     return rc;
 }
 
@@ -652,7 +654,7 @@ rpmRC rpmReadHeader(rpmts ts, FD_t fd, Header *hdrp, const char ** msg)
     memset(block, 0, sizeof(block));
     if ((xx = timedRead(fd, (char *)block, sizeof(block))) != sizeof(block)) {
 	(void) snprintf(buf, sizeof(buf),
-		_("hdr size(%d): BAD, read returned %d\n"), sizeof(block), xx);
+		_("hdr size(%d): BAD, read returned %d\n"), (int)sizeof(block), xx);
 	goto exit;
     }
     if (memcmp(block, header_magic, sizeof(header_magic))) {
