@@ -33,6 +33,9 @@ extern int _fsm_debug;
 /*@unchecked@*/
 extern int _hdr_debug;
 
+/*@unchecked@*/
+extern int _print_pkts;
+
 /*@unchecked@*/
 extern int _psm_debug;
 
@@ -45,6 +48,10 @@ extern int _rpmdb_debug;
 /*@unchecked@*/
 extern int _rpmds_debug;
 
+/* XXX avoid -lrpmbuild linkage. */
+/*@unchecked@*/
+       int _rpmfc_debug;
+
 /*@unchecked@*/
 extern int _rpmfi_debug;
 
@@ -57,6 +64,9 @@ extern int _rpmte_debug;
 /*@unchecked@*/
 extern int _rpmts_debug;
 
+/*@unchecked@*/
+extern int _rpmts_stats;
+
 /*@unchecked@*/
 extern int noLibio;
 /*@=exportheadervar@*/
@@ -270,6 +280,8 @@ struct poptOption rpmcliAllPoptTable[] = {
  { "poptdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_popt_debug, -1,
 	N_("debug option/argument processing"), NULL},
 #endif
+ { "prtpkts", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_print_pkts, -1,
+	NULL, NULL},
  { "psmdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_psm_debug, -1,
 	N_("debug package state machine"), NULL},
  { "rpmaldebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmal_debug, -1,
@@ -278,6 +290,8 @@ struct poptOption rpmcliAllPoptTable[] = {
 	NULL, NULL},
  { "rpmdsdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmds_debug, -1,
 	NULL, NULL},
+ { "rpmfcdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmfc_debug, -1,
+	NULL, NULL},
  { "rpmfidebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmfi_debug, -1,
 	NULL, NULL},
  { "rpmiodebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmio_debug, -1,
@@ -288,6 +302,8 @@ struct poptOption rpmcliAllPoptTable[] = {
 	NULL, NULL},
  { "rpmtsdebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmts_debug, -1,
 	NULL, NULL},
+ { "stats", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_rpmts_stats, -1,
+	NULL, NULL},
  { "urldebug", '\0', POPT_ARG_VAL|POPT_ARGFLAG_DOC_HIDDEN, &_url_debug, -1,
 	N_("debug URL cache handling"), NULL},
 
@@ -332,9 +348,8 @@ rpmcliInit(int argc, char *const argv[], struct poptOption * optionsTable)
     }
 /*@=globs =mods@*/
 
-#if !defined(__LCLINT__)
+#if defined(ENABLE_NLS) && !defined(__LCLINT__)
     (void) setlocale(LC_ALL, "" );
-
     (void) bindtextdomain(PACKAGE, LOCALEDIR);
     (void) textdomain(PACKAGE);
 #endif
