@@ -11,6 +11,8 @@
 #include "cpio.h"	/* XXX CPIO_FOO */
 #include "fsm.h"	/* XXX newFSM() */
 
+#include "rpmds.h"
+
 #define	_RPMFI_INTERNAL
 #include "rpmfi.h"
 
@@ -19,6 +21,8 @@
 #include "rpmts.h"
 
 #include "misc.h"	/* XXX stripTrailingChar */
+#include "legacy.h"	/* XXX domd5 */
+#include "rpmmacro.h"	/* XXX rpmCleanPath */
 
 #include "debug.h"
 
@@ -263,6 +267,68 @@ int_32 rpmfiFInode(rpmfi fi)
     return finode;
 }
 
+uint_32 rpmfiColor(rpmfi fi)
+{
+    uint_32 color = 0;
+
+    if (fi != NULL)
+	/* XXX ignore all but lsnibble for now. */
+	color = fi->color & 0xf;
+    return color;
+}
+
+uint_32 rpmfiFColor(rpmfi fi)
+{
+    uint_32 fcolor = 0;
+
+    if (fi != NULL && fi->i >= 0 && fi->i < fi->fc) {
+/*@-boundsread@*/
+	if (fi->fcolors != NULL)
+	    /* XXX ignore all but lsnibble for now. */
+	    fcolor = (fi->fcolors[fi->i] & 0x0f);
+/*@=boundsread@*/
+    }
+    return fcolor;
+}
+
+const char * rpmfiFClass(rpmfi fi)
+{
+    const char * fclass = NULL;
+    int cdictx;
+
+    if (fi != NULL && fi->fcdictx != NULL && fi->i >= 0 && fi->i < fi->fc) {
+/*@-boundsread@*/
+	cdictx = fi->fcdictx[fi->i];
+	if (fi->cdict != NULL && cdictx >= 0 && cdictx < fi->ncdict)
+	    fclass = fi->cdict[cdictx];
+/*@=boundsread@*/
+    }
+    return fclass;
+}
+
+int_32 rpmfiFDepends(rpmfi fi, const int_32 ** fddictp)
+{
+    int fddictx = -1;
+    int fddictn = 0;
+    const int_32 * fddict = NULL;
+
+    if (fi != NULL && fi->i >= 0 && fi->i < fi->fc) {
+/*@-boundsread@*/
+	if (fi->fddictn != NULL)
+	    fddictn = fi->fddictn[fi->i];
+	if (fddictn > 0 && fi->fddictx != NULL)
+	    fddictx = fi->fddictx[fi->i];
+	if (fi->ddict != NULL && fddictx >= 0 && (fddictx+fddictn) <= fi->nddict)
+	    fddict = fi->ddict + fddictx;
+/*@=boundsread@*/
+    }
+/*@-boundswrite -dependenttrans -onlytrans @*/
+    if (fddictp)
+	*fddictp = fddict;
+/*@=boundswrite =dependenttrans =onlytrans @*/
+    return fddictn;
+}
+
 int_32 rpmfiFNlink(rpmfi fi)
 {
     int_32 nlink = 0;
@@ -433,6 +499,120 @@ fileTypes whatis(uint_16 mode)
     return REG;
 }
 
+/*@-boundsread@*/
+int rpmfiCompare(const rpmfi afi, const rpmfi bfi)
+	/*@*/
+{
+    fileTypes awhat = whatis(rpmfiFMode(afi));
+    fileTypes bwhat = whatis(rpmfiFMode(bfi));
+
+    if (awhat != bwhat) return 1;
+
+    if (awhat == LINK) {
+	const char * alink = rpmfiFLink(afi);
+	const char * blink = rpmfiFLink(bfi);
+	if (alink == blink) return 0;
+	if (alink == NULL) return 1;
+	if (blink == NULL) return -1;
+	return strcmp(alink, blink);
+    } else if (awhat == REG) {
+	const unsigned char * amd5 = rpmfiMD5(afi);
+	const unsigned char * bmd5 = rpmfiMD5(bfi);
+	if (amd5 == bmd5) return 0;
+	if (amd5 == NULL) return 1;
+	if (bmd5 == NULL) return -1;
+	return memcmp(amd5, bmd5, 16);
+    }
+
+    return 0;
+}
+/*@=boundsread@*/
+
+/*@-boundsread@*/
+fileAction rpmfiDecideFate(const rpmfi ofi, rpmfi nfi, int skipMissing)
+{
+    const char * fn = rpmfiFN(nfi);
+    int newFlags = rpmfiFFlags(nfi);
+    char buffer[1024];
+    fileTypes dbWhat, newWhat, diskWhat;
+    struct stat sb;
+    int save = (newFlags & RPMFILE_NOREPLACE) ? FA_ALTNAME : FA_SAVE;
+
+    if (lstat(fn, &sb)) {
+	/*
+	 * The file doesn't exist on the disk. Create it unless the new
+	 * package has marked it as missingok, or allfiles is requested.
+	 */
+	if (skipMissing && (newFlags & RPMFILE_MISSINGOK)) {
+	    rpmMessage(RPMMESS_DEBUG, _("%s skipped due to missingok flag\n"),
+			fn);
+	    return FA_SKIP;
+	} else {
+	    return FA_CREATE;
+	}
+    }
+
+    diskWhat = whatis((int_16)sb.st_mode);
+    dbWhat = whatis(rpmfiFMode(ofi));
+    newWhat = whatis(rpmfiFMode(nfi));
+
+    /*
+     * RPM >= 2.3.10 shouldn't create config directories -- we'll ignore
+     * them in older packages as well.
+     */
+    if (newWhat == XDIR)
+	return FA_CREATE;
+
+    if (diskWhat != newWhat)
+	return save;
+    else if (newWhat != dbWhat && diskWhat != dbWhat)
+	return save;
+    else if (dbWhat != newWhat)
+	return FA_CREATE;
+    else if (dbWhat != LINK && dbWhat != REG)
+	return FA_CREATE;
+
+    /*
+     * This order matters - we'd prefer to CREATE the file if at all
+     * possible in case something else (like the timestamp) has changed.
+     */
+    if (dbWhat == REG) {
+	const unsigned char * omd5, * nmd5;
+	if (domd5(fn, buffer, 0, NULL))
+	    return FA_CREATE;	/* assume file has been removed */
+	omd5 = rpmfiMD5(ofi);
+	if (omd5 && !memcmp(omd5, buffer, 16))
+	    return FA_CREATE;	/* unmodified config file, replace. */
+	nmd5 = rpmfiMD5(nfi);
+/*@-nullpass@*/
+	if (omd5 && nmd5 && !memcmp(omd5, nmd5, 16))
+	    return FA_SKIP;	/* identical file, don't bother. */
+/*@=nullpass@*/
+    } else /* dbWhat == LINK */ {
+	const char * oFLink, * nFLink;
+	memset(buffer, 0, sizeof(buffer));
+	if (readlink(fn, buffer, sizeof(buffer) - 1) == -1)
+	    return FA_CREATE;	/* assume file has been removed */
+	oFLink = rpmfiFLink(ofi);
+	if (oFLink && !strcmp(oFLink, buffer))
+	    return FA_CREATE;	/* unmodified config file, replace. */
+	nFLink = rpmfiFLink(nfi);
+/*@-nullpass@*/
+	if (oFLink && nFLink && !strcmp(oFLink, nFLink))
+	    return FA_SKIP;	/* identical file, don't bother. */
+/*@=nullpass@*/
+    }
+
+    /*
+     * The config file on the disk has been modified, but
+     * the ones in the two packages are different. It would
+     * be nice if RPM was smart enough to at least try and
+     * merge the difference ala CVS, but...
+     */
+    return save;
+}
+/*@=boundsread@*/
+
 /*@observer@*/
 const char *const rpmfiTypeString(rpmfi fi)
 {
@@ -461,7 +641,7 @@ Header relocateFileList(const rpmts ts, rpmfi fi,
 		Header origH, fileAction * actions)
 	/*@modifies ts, fi, origH, actions @*/
 {
-    rpmte p = fi->te;
+    rpmte p = rpmtsRelocateElement(ts);
     HGE_t hge = fi->hge;
     HAE_t hae = fi->hae;
     HME_t hme = fi->hme;
@@ -480,8 +660,9 @@ Header relocateFileList(const rpmts ts, rpmfi fi,
     int_32 fileCount;
     int_32 dirCount;
     uint_32 * fFlags = NULL;
+    uint_32 * fColors = NULL;
+    uint_32 * dColors = NULL;
     uint_16 * fModes = NULL;
-    char * skipDirList;
     Header h;
     int nrelocated = 0;
     int fileAlloced = 0;
@@ -495,6 +676,7 @@ Header relocateFileList(const rpmts ts, rpmfi fi,
 			(void **) &validRelocations, &numValid))
 	numValid = 0;
 
+assert(p != NULL);
     numRelocations = 0;
     if (p->relocs)
 	while (p->relocs[numRelocations].newPath ||
@@ -649,10 +831,11 @@ Header relocateFileList(const rpmts ts, rpmfi fi,
     xx = hge(h, RPMTAG_DIRINDEXES, NULL, (void **) &dirIndexes, NULL);
     xx = hge(h, RPMTAG_DIRNAMES, NULL, (void **) &dirNames, &dirCount);
     xx = hge(h, RPMTAG_FILEFLAGS, NULL, (void **) &fFlags, NULL);
+    xx = hge(h, RPMTAG_FILECOLORS, NULL, (void **) &fColors, NULL);
     xx = hge(h, RPMTAG_FILEMODES, NULL, (void **) &fModes, NULL);
 
-    skipDirList = alloca(dirCount * sizeof(*skipDirList));
-    memset(skipDirList, 0, dirCount * sizeof(*skipDirList));
+    dColors = alloca(dirCount * sizeof(*dColors));
+    memset(dColors, 0, dirCount * sizeof(*dColors));
 
     newDirIndexes = alloca(sizeof(*newDirIndexes) * fileCount);
     memcpy(newDirIndexes, dirIndexes, sizeof(*newDirIndexes) * fileCount);
@@ -670,19 +853,6 @@ Header relocateFileList(const rpmts ts, rpmfi fi,
 	fileTypes ft;
 	int fnlen;
 
-	/*
-	 * If only adding libraries of different arch into an already
-	 * installed package, skip all other files.
-	 */
-	if (rpmteMultiLib(p) && !isFileMULTILIB((fFlags[i]))) {
-	    if (actions) {
-		actions[i] = FA_SKIPMULTILIB;
-		rpmMessage(RPMMESS_DEBUG, _("excluding multilib path %s%s\n"), 
-			dirNames[dirIndexes[i]], baseNames[i]);
-	    }
-	    continue;
-	}
-
 	len = reldel +
 		strlen(dirNames[dirIndexes[i]]) + strlen(baseNames[i]) + 1;
 	/*@-branchstate@*/
@@ -691,9 +861,12 @@ Header relocateFileList(const rpmts ts, rpmfi fi,
 	    fn = xrealloc(fn, fileAlloced);
 	}
 	/*@=branchstate@*/
+
+assert(fn != NULL);		/* XXX can't happen */
 	*fn = '\0';
 	fnlen = stpcpy( stpcpy(fn, dirNames[dirIndexes[i]]), baseNames[i]) - fn;
 
+dColors[dirIndexes[i]] |= fColors[i];
 	/*
 	 * See if this file path needs relocating.
 	 */
@@ -723,7 +896,9 @@ Header relocateFileList(const rpmts ts, rpmfi fi,
 	}
 	if (j < 0) continue;
 
+/*@-nullderef@*/ /* FIX: fModes may be NULL */
 	ft = whatis(fModes[i]);
+/*@=nullderef@*/
 
 	/* On install, a relocate to NULL means skip the path. */
 	if (relocations[j].newPath == NULL) {
@@ -738,8 +913,6 @@ Header relocateFileList(const rpmts ts, rpmfi fi,
 			/*@innercontinue@*/ continue;
 		    /*@innerbreak@*/ break;
 		}
-		if (j < dirCount)
-		    skipDirList[j] = 1;
 	    }
 	    if (actions) {
 		actions[i] = FA_SKIPNSTATE;
@@ -809,6 +982,9 @@ Header relocateFileList(const rpmts ts, rpmfi fi,
     for (i = dirCount - 1; i >= 0; i--) {
 	for (j = numRelocations - 1; j >= 0; j--) {
 
+	    /* XXX Don't autorelocate uncolored directories. */
+if (j == p->autorelocatex && (dColors[i] == 0 || !(dColors[i] & 0x1))) continue;
+
 	    if (relocations[j].oldPath == NULL) /* XXX can't happen */
 		/*@innercontinue@*/ continue;
 	    len = strcmp(relocations[j].oldPath, "/")
@@ -828,8 +1004,16 @@ Header relocateFileList(const rpmts ts, rpmfi fi,
 	    if (relocations[j].newPath) { /* Relocate the path */
 		const char * s = relocations[j].newPath;
 		char * t = alloca(strlen(s) + strlen(dirNames[i]) - len + 1);
+		size_t slen;
 
 		(void) stpcpy( stpcpy(t, s) , dirNames[i] + len);
+
+		/* Unfortunatly rpmCleanPath strips the trailing slash.. */
+		(void) rpmCleanPath(t);
+		slen = strlen(t);
+		t[slen] = '/';
+		t[slen+1] = '\0';
+
 		if (actions)
 		    rpmMessage(RPMMESS_DEBUG,
 			_("relocating directory %s to %s\n"), dirNames[i], t);
@@ -877,7 +1061,9 @@ Header relocateFileList(const rpmts ts, rpmfi fi,
 
     baseNames = hfd(baseNames, RPM_STRING_ARRAY_TYPE);
     dirNames = hfd(dirNames, RPM_STRING_ARRAY_TYPE);
+/*@-dependenttrans@*/
     fn = _free(fn);
+/*@=dependenttrans@*/
 
     return h;
 }
@@ -907,6 +1093,8 @@ fprintf(stderr, "*** fi %p\t%s[%d]\n", fi, fi->Type, fi->fc);
 	fi->fmd5s = hfd(fi->fmd5s, -1);
 	fi->md5s = _free(fi->md5s);
 
+	fi->cdict = hfd(fi->cdict, -1);
+
 	fi->fuser = hfd(fi->fuser, -1);
 	fi->fuids = _free(fi->fuids);
 	fi->fgroup = hfd(fi->fgroup, -1);
@@ -924,6 +1112,13 @@ fprintf(stderr, "*** fi %p\t%s[%d]\n", fi, fi->Type, fi->fc);
 	    fi->frdevs = _free(fi->frdevs);
 	    fi->finodes = _free(fi->finodes);
 	    fi->dil = _free(fi->dil);
+
+	    fi->fcolors = _free(fi->fcolors);
+	    fi->fcdictx = _free(fi->fcdictx);
+	    fi->ddict = _free(fi->ddict);
+	    fi->fddictx = _free(fi->fddictx);
+	    fi->fddictn = _free(fi->fddictn);
+
 	}
 	/*@=evalorder@*/
     }
@@ -980,6 +1175,7 @@ rpmfi rpmfiNew(rpmts ts, Header h, rpmTag tagN, int scareMem)
     HGE_t hge =
 	(scareMem ? (HGE_t) headerGetEntryMinMemory : (HGE_t) headerGetEntry);
     HFD_t hfd = headerFreeData;
+    rpmte p;
     rpmfi fi = NULL;
     const char * Type;
     uint_32 * uip;
@@ -1032,6 +1228,19 @@ rpmfi rpmfiNew(rpmts ts, Header h, rpmTag tagN, int scareMem)
     xx = hge(h, RPMTAG_FILEFLAGS, NULL, (void **) &fi->fflags, NULL);
     xx = hge(h, RPMTAG_FILEVERIFYFLAGS, NULL, (void **) &fi->vflags, NULL);
     xx = hge(h, RPMTAG_FILESIZES, NULL, (void **) &fi->fsizes, NULL);
+
+    xx = hge(h, RPMTAG_FILECOLORS, NULL, (void **) &fi->fcolors, NULL);
+    fi->color = 0;
+    if (fi->fcolors != NULL)
+    for (i = 0; i < fi->fc; i++)
+	fi->color |= fi->fcolors[i];
+    xx = hge(h, RPMTAG_CLASSDICT, NULL, (void **) &fi->cdict, &fi->ncdict);
+    xx = hge(h, RPMTAG_FILECLASS, NULL, (void **) &fi->fcdictx, NULL);
+
+    xx = hge(h, RPMTAG_DEPENDSDICT, NULL, (void **) &fi->ddict, &fi->nddict);
+    xx = hge(h, RPMTAG_FILEDEPENDSX, NULL, (void **) &fi->fddictx, NULL);
+    xx = hge(h, RPMTAG_FILEDEPENDSN, NULL, (void **) &fi->fddictn, NULL);
+
     xx = hge(h, RPMTAG_FILESTATES, NULL, (void **) &fi->fstates, NULL);
     if (xx == 0 || fi->fstates == NULL)
 	fi->fstates = xcalloc(fi->fc, sizeof(*fi->fstates));
@@ -1053,24 +1262,28 @@ if (fi->actions == NULL)
     xx = hge(h, RPMTAG_FILELINKTOS, NULL, (void **) &fi->flinks, NULL);
     xx = hge(h, RPMTAG_FILELANGS, NULL, (void **) &fi->flangs, NULL);
 
+    fi->fmd5s = NULL;
     xx = hge(h, RPMTAG_FILEMD5S, NULL, (void **) &fi->fmd5s, NULL);
 
-    t = xmalloc(fi->fc * 16);
-    fi->md5s = t;
-    for (i = 0; i < fi->fc; i++) {
-	const char * fmd5;
-	int j;
+    fi->md5s = NULL;
+    if (fi->fmd5s) {
+	t = xmalloc(fi->fc * 16);
+	fi->md5s = t;
+	for (i = 0; i < fi->fc; i++) {
+	    const char * fmd5;
+	    int j;
 
-	fmd5 = fi->fmd5s[i];
-	if (!(fmd5 && *fmd5 != '\0')) {
-	    memset(t, 0, 16);
-	    t += 16;
-	    continue;
+	    fmd5 = fi->fmd5s[i];
+	    if (!(fmd5 && *fmd5 != '\0')) {
+		memset(t, 0, 16);
+		t += 16;
+		continue;
+	    }
+	    for (j = 0; j < 16; j++, t++, fmd5 += 2)
+		*t = (nibble(fmd5[0]) << 4) | nibble(fmd5[1]);
 	}
-	for (j = 0; j < 16; j++, t++, fmd5 += 2)
-	    *t = (nibble(fmd5[0]) << 4) | nibble(fmd5[1]);
+	fi->fmd5s = hfd(fi->fmd5s, -1);
     }
-    fi->fmd5s = hfd(fi->fmd5s, -1);
 
     /* XXX TR_REMOVED doesn;t need fmtimes, frdevs or finodes */
     xx = hge(h, RPMTAG_FILEMTIMES, NULL, (void **) &fi->fmtimes, NULL);
@@ -1086,8 +1299,59 @@ if (fi->actions == NULL)
 
     if (ts != NULL)
     if (fi != NULL)
-    if (fi->te != NULL && rpmteType(fi->te) == TR_ADDED) {
+    if ((p = rpmtsRelocateElement(ts)) != NULL && rpmteType(p) == TR_ADDED
+     && !headerIsEntry(h, RPMTAG_SOURCEPACKAGE)
+     && !headerIsEntry(h, RPMTAG_ORIGBASENAMES))
+    {
+	const char * fmt = rpmGetPath("%{?_autorelocate_path}", NULL);
+	const char * errstr;
+	char * newPath;
 	Header foo;
+
+	/* XXX error handling. */
+	newPath = headerSprintf(h, fmt, rpmTagTable, rpmHeaderFormats, &errstr);
+	fmt = _free(fmt);
+
+#if __ia64__
+	/* XXX On ia64, change leading /emul/ix86 -> /emul/ia32, ick. */
+	if (newPath != NULL && *newPath != '\0'
+	 && strlen(newPath) >= (sizeof("/emul/i386")-1)
+	 && newPath[0] == '/' && newPath[1] == 'e' && newPath[2] == 'm'
+	 && newPath[3] == 'u' && newPath[4] == 'l' && newPath[5] == '/'
+	 && newPath[6] == 'i' && newPath[8] == '8' && newPath[9] == '6')
+	{
+	    newPath[7] = 'a';
+	    newPath[8] = '3';
+	    newPath[9] = '2';
+	}
+#endif
+
+	/* XXX Make sure autoreloc is not already specified. */
+	i = p->nrelocs;
+	if (newPath != NULL && *newPath != '\0' && p->relocs != NULL)
+	for (i = 0; i < p->nrelocs; i++) {
+	   if (strcmp(p->relocs[i].oldPath, "/"))
+		continue;
+	   if (strcmp(p->relocs[i].newPath, newPath))
+		continue;
+	   break;
+	}
+
+	/* XXX test for incompatible arch triggering autorelocation is dumb. */
+	if (newPath != NULL && *newPath != '\0' && i == p->nrelocs
+	 && p->archScore == 0)
+	{
+	    p->relocs =
+		xrealloc(p->relocs, (p->nrelocs + 2) * sizeof(*p->relocs));
+	    p->relocs[p->nrelocs].oldPath = xstrdup("/");
+	    p->relocs[p->nrelocs].newPath = xstrdup(newPath);
+	    p->autorelocatex = p->nrelocs;
+	    p->nrelocs++;
+	    p->relocs[p->nrelocs].oldPath = NULL;
+	    p->relocs[p->nrelocs].newPath = NULL;
+	}
+	newPath = _free(newPath);
+
 /* XXX DYING */
 if (fi->actions == NULL)
 	fi->actions = xcalloc(fi->fc, sizeof(*fi->actions));
@@ -1108,6 +1372,17 @@ if (fi->actions == NULL)
 	_fdupe(fi, vflags);
 	_fdupe(fi, fmodes);
 	_fdupe(fi, dil);
+
+	_fdupe(fi, fcolors);
+	_fdupe(fi, fcdictx);
+
+	if (fi->ddict != NULL)
+	    fi->ddict = memcpy(xmalloc(fi->nddict * sizeof(*fi->ddict)),
+			fi->ddict, fi->nddict * sizeof(*fi->ddict));
+
+	_fdupe(fi, fddictx);
+	_fdupe(fi, fddictn);
+
 	fi->h = headerFree(fi->h);
     }
 
@@ -1137,3 +1412,157 @@ fprintf(stderr, "*** fi %p\t%s[%d]\n", fi, Type, (fi ? fi->fc : 0));
     return rpmfiLink(fi, (fi ? fi->Type : NULL));
     /*@=compdef =nullstate@*/
 }
+
+void rpmfiBuildFClasses(Header h,
+	/*@out@*/ const char *** fclassp, /*@out@*/ int * fcp)
+{
+    int scareMem = 1;
+    rpmfi fi = rpmfiNew(NULL, h, RPMTAG_BASENAMES, scareMem);
+    const char * FClass;
+    const char ** av;
+    int ac;
+    size_t nb;
+    char * t;
+
+    if ((ac = rpmfiFC(fi)) <= 0) {
+	av = NULL;
+	ac = 0;
+	goto exit;
+    }
+
+    /* Compute size of file class argv array blob. */
+    nb = (ac + 1) * sizeof(*av);
+    fi = rpmfiInit(fi, 0);
+    if (fi != NULL)
+    while (rpmfiNext(fi) >= 0) {
+	FClass = rpmfiFClass(fi);
+	if (FClass && *FClass != '\0')
+	    nb += strlen(FClass);
+	nb += 1;
+    }
+
+    /* Create and load file class argv array. */
+    av = xmalloc(nb);
+    t = ((char *) av) + ((ac + 1) * sizeof(*av));
+    ac = 0;
+    fi = rpmfiInit(fi, 0);
+    if (fi != NULL)
+    while (rpmfiNext(fi) >= 0) {
+	FClass = rpmfiFClass(fi);
+	av[ac++] = t;
+	if (FClass && *FClass != '\0')
+	    t = stpcpy(t, FClass);
+	*t++ = '\0';
+    }
+    av[ac] = NULL;
+    /*@=branchstate@*/
+
+exit:
+    fi = rpmfiFree(fi);
+    /*@-branchstate@*/
+    if (fclassp)
+	*fclassp = av;
+    else
+	av = _free(av);
+    /*@=branchstate@*/
+    if (fcp) *fcp = ac;
+}
+
+void rpmfiBuildFDeps(Header h, rpmTag tagN,
+	/*@out@*/ const char *** fdepsp, /*@out@*/ int * fcp)
+{
+    int scareMem = 1;
+    rpmfi fi = rpmfiNew(NULL, h, RPMTAG_BASENAMES, scareMem);
+    rpmds ds = NULL;
+    const char ** av;
+    int ac;
+    size_t nb;
+    char * t;
+    char deptype = 'R';
+    char mydt;
+    const char * DNEVR;
+    const int_32 * ddict;
+    unsigned ix;
+    int ndx;
+
+    if ((ac = rpmfiFC(fi)) <= 0) {
+	av = NULL;
+	ac = 0;
+	goto exit;
+    }
+
+    if (tagN == RPMTAG_PROVIDENAME)
+	deptype = 'P';
+    else if (tagN == RPMTAG_REQUIRENAME)
+	deptype = 'R';
+
+    ds = rpmdsNew(h, tagN, scareMem);
+
+    /* Compute size of file depends argv array blob. */
+    nb = (ac + 1) * sizeof(*av);
+    fi = rpmfiInit(fi, 0);
+    if (fi != NULL)
+    while (rpmfiNext(fi) >= 0) {
+	ddict = NULL;
+	ndx = rpmfiFDepends(fi, &ddict);
+	if (ddict != NULL)
+	while (ndx-- > 0) {
+	    ix = *ddict++;
+	    mydt = ((ix >> 24) & 0xff);
+	    if (mydt != deptype)
+		/*@innercontinue@*/ continue;
+	    ix &= 0x00ffffff;
+	    (void) rpmdsSetIx(ds, ix-1);
+	    if (rpmdsNext(ds) < 0)
+		/*@innercontinue@*/ continue;
+	    DNEVR = rpmdsDNEVR(ds);
+	    if (DNEVR != NULL)
+		nb += strlen(DNEVR+2) + 1;
+	}
+	nb += 1;
+    }
+
+    /* Create and load file depends argv array. */
+    av = xmalloc(nb);
+    t = ((char *) av) + ((ac + 1) * sizeof(*av));
+    ac = 0;
+    /*@-branchstate@*/
+    fi = rpmfiInit(fi, 0);
+    if (fi != NULL)
+    while (rpmfiNext(fi) >= 0) {
+	av[ac++] = t;
+	ddict = NULL;
+	ndx = rpmfiFDepends(fi, &ddict);
+	if (ddict != NULL)
+	while (ndx-- > 0) {
+	    ix = *ddict++;
+	    mydt = ((ix >> 24) & 0xff);
+	    if (mydt != deptype)
+		/*@innercontinue@*/ continue;
+	    ix &= 0x00ffffff;
+	    (void) rpmdsSetIx(ds, ix-1);
+	    if (rpmdsNext(ds) < 0)
+		/*@innercontinue@*/ continue;
+	    DNEVR = rpmdsDNEVR(ds);
+	    if (DNEVR != NULL) {
+		t = stpcpy(t, DNEVR+2);
+		*t++ = ' ';
+		*t = '\0';
+	    }
+	}
+	*t++ = '\0';
+    }
+    /*@=branchstate@*/
+    av[ac] = NULL;
+
+exit:
+    fi = rpmfiFree(fi);
+    ds = rpmdsFree(ds);
+    /*@-branchstate@*/
+    if (fdepsp)
+	*fdepsp = av;
+    else
+	av = _free(av);
+    /*@=branchstate@*/
+    if (fcp) *fcp = ac;
+}
