@@ -3,6 +3,9 @@
 
 /** \ingroup rpmcli rpmrc rpmdep rpmtrans rpmdb lead signature header payload dbi
  * \file lib/rpmlib.h
+ *
+ * In Memoriam: Steve Taylor <staylor@redhat.com> was here, now he's not.
+ *
  */
 
 #include "rpmio.h"
@@ -67,6 +70,10 @@ void * _free(/*@only@*/ /*@null@*/ /*@out@*/ const void * p)
  */
 typedef /*@abstract@*/ /*@refcounted@*/ struct rpmts_s * rpmts;
 
+/** \ingroup rpmbuild
+ */
+typedef struct Spec_s * Spec;
+
 /** \ingroup rpmtrans
  * An added/available package retrieval key.
  */
@@ -108,9 +115,9 @@ typedef /*@abstract@*/ struct _rpmdbMatchIterator * rpmdbMatchIterator;
 /** \ingroup header
  * Return name, version, release strings from header.
  * @param h		header
- * @retval np		address of name pointer (or NULL)
- * @retval vp		address of version pointer (or NULL)
- * @retval rp		address of release pointer (or NULL)
+ * @retval *np		name pointer (or NULL)
+ * @retval *vp		version pointer (or NULL)
+ * @retval *rp		release pointer (or NULL)
  * @return		0 always
  */
 int headerNVR(Header h,
@@ -119,6 +126,24 @@ int headerNVR(Header h,
 		/*@null@*/ /*@out@*/ const char ** rp)
 	/*@modifies *np, *vp, *rp @*/;
 
+/** \ingroup header
+ * Return name, epoch, version, release, arch strings from header.
+ * @param h		header
+ * @retval *np		name pointer (or NULL)
+ * @retval *ep		epoch pointer (or NULL)
+ * @retval *vp		version pointer (or NULL)
+ * @retval *rp		release pointer (or NULL)
+ * @retval *ap		arch pointer (or NULL)
+ * @return		0 always
+ */
+int headerNEVRA(Header h,
+		/*@null@*/ /*@out@*/ const char ** np,
+		/*@null@*/ /*@out@*/ /*@unused@*/ const char ** ep,
+		/*@null@*/ /*@out@*/ const char ** vp,
+		/*@null@*/ /*@out@*/ const char ** rp,
+		/*@null@*/ /*@out@*/ const char ** ap)
+	/*@modifies *np, *vp, *rp, *ap @*/;
+
 /** \ingroup header
  * Translate and merge legacy signature tags into header.
  * @todo Remove headerSort() through headerInitIterator() modifies sig.
@@ -138,23 +163,6 @@ void headerMergeLegacySigs(Header h, const Header sigh)
 Header headerRegenSigHeader(const Header h, int noArchiveSize)
 	/*@modifies h @*/;
 
-/** \ingroup header
- * Retrieve file names from header.
- * The representation of file names in package headers changed in rpm-4.0.
- * Originally, file names were stored as an array of paths. In rpm-4.0,
- * file names are stored as separate arrays of dirname's and basename's,
- * with a dirname index to associate the correct dirname with each basname.
- * This function is used to retrieve file names independent of how the
- * file names are represented in the package header.
- * 
- * @param h		header
- * @retval fileListPtr	address of array of file names
- * @retval fileCountPtr	address of number of files
- */
-void rpmBuildFileList(Header h, /*@out@*/ const char *** fileListPtr, 
-		/*@out@*/ int * fileCountPtr)
-	/*@modifies *fileListPtr, *fileCountPtr @*/;
-
 /** \ingroup header
  * Retrieve tag info from header.
  * This is a "dressed" entry to headerGetEntry to do:
@@ -232,6 +240,7 @@ typedef enum rpmTag_e {
     RPMTAG_SIGPGP		= RPMTAG_SIG_BASE+3,
     RPMTAG_SIGLEMD5_2		= RPMTAG_SIG_BASE+4,	/*!< internal - obsolete */
     RPMTAG_SIGMD5	        = RPMTAG_SIG_BASE+5,
+#define	RPMTAG_PKGID	RPMTAG_SIGMD5
     RPMTAG_SIGGPG	        = RPMTAG_SIG_BASE+6,
     RPMTAG_SIGPGP5	        = RPMTAG_SIG_BASE+7,	/*!< internal - obsolete */
 
@@ -241,11 +250,16 @@ typedef enum rpmTag_e {
     RPMTAG_DSAHEADER		= RPMTAG_SIG_BASE+11,
     RPMTAG_RSAHEADER		= RPMTAG_SIG_BASE+12,
     RPMTAG_SHA1HEADER		= RPMTAG_SIG_BASE+13,
+#define	RPMTAG_HDRID	RPMTAG_SHA1HEADER
 
     RPMTAG_NAME  		= 1000,
+#define	RPMTAG_N	RPMTAG_NAME
     RPMTAG_VERSION		= 1001,
+#define	RPMTAG_V	RPMTAG_VERSION
     RPMTAG_RELEASE		= 1002,
+#define	RPMTAG_R	RPMTAG_RELEASE
     RPMTAG_EPOCH   		= 1003,
+#define	RPMTAG_E	RPMTAG_EPOCH
 #define	RPMTAG_SERIAL	RPMTAG_EPOCH	/* backward comaptibility */
     RPMTAG_SUMMARY		= 1004,
     RPMTAG_DESCRIPTION		= 1005,
@@ -378,7 +392,7 @@ typedef enum rpmTag_e {
     RPMTAG_PAYLOADFORMAT	= 1124,
     RPMTAG_PAYLOADCOMPRESSOR	= 1125,
     RPMTAG_PAYLOADFLAGS		= 1126,
-    RPMTAG_MULTILIBS		= 1127,
+    RPMTAG_INSTALLCOLOR		= 1127, /*!< transaction color when installed */
     RPMTAG_INSTALLTID		= 1128,
     RPMTAG_REMOVETID		= 1129,
 /*@-enummemuse@*/
@@ -393,6 +407,13 @@ typedef enum rpmTag_e {
     RPMTAG_CACHEPKGPATH		= 1137,
     RPMTAG_CACHEPKGSIZE		= 1138,
     RPMTAG_CACHEPKGMTIME	= 1139,
+    RPMTAG_FILECOLORS		= 1140,
+    RPMTAG_FILECLASS		= 1141,
+    RPMTAG_CLASSDICT		= 1142,
+    RPMTAG_FILEDEPENDSX		= 1143,
+    RPMTAG_FILEDEPENDSN		= 1144,
+    RPMTAG_DEPENDSDICT		= 1145,
+    RPMTAG_SOURCEPKGID		= 1146,
 /*@-enummemuse@*/
     RPMTAG_FIRSTFREE_TAG	/*!< internal */
 /*@=enummemuse@*/
@@ -407,7 +428,8 @@ typedef enum rpmfileState_e {
     RPMFILE_STATE_NORMAL 	= 0,
     RPMFILE_STATE_REPLACED 	= 1,
     RPMFILE_STATE_NOTINSTALLED	= 2,
-    RPMFILE_STATE_NETSHARED	= 3
+    RPMFILE_STATE_NETSHARED	= 3,
+    RPMFILE_STATE_WRONGCOLOR	= 4
 } rpmfileState;
 #define	RPMFILE_STATE_MISSING	-1	/* XXX used for unavailable data */
 
@@ -418,29 +440,22 @@ typedef	enum rpmfileAttrs_e {
 /*@-enummemuse@*/
     RPMFILE_NONE	= 0,
 /*@=enummemuse@*/
-    RPMFILE_CONFIG	= (1 << 0),	/*!< from %%config */
-    RPMFILE_DOC		= (1 << 1),	/*!< from %%doc */
-/*@-enummemuse@*/
-    RPMFILE_DONOTUSE	= (1 << 2),	/*!< @todo (unimplemented) from %donotuse. */
-/*@=enummemuse@*/
-    RPMFILE_MISSINGOK	= (1 << 3),	/*!< from %%config(missingok) */
-    RPMFILE_NOREPLACE	= (1 << 4),	/*!< from %%config(noreplace) */
-    RPMFILE_SPECFILE	= (1 << 5),	/*!< @todo (unnecessary) marks 1st file in srpm. */
-    RPMFILE_GHOST	= (1 << 6),	/*!< from %%ghost */
-    RPMFILE_LICENSE	= (1 << 7),	/*!< from %%license */
-    RPMFILE_README	= (1 << 8),	/*!< from %%readme */
-    RPMFILE_EXCLUDE	= (1 << 9),	/*!< from %%exclude */
-    RPMFILE_UNPATCHED	= (1 << 10)	/*!< placeholder (SuSE) */
+    RPMFILE_CONFIG	= (1 <<  0),	/*!< from %%config */
+    RPMFILE_DOC		= (1 <<  1),	/*!< from %%doc */
+    RPMFILE_ICON	= (1 <<  2),	/*!< from %%donotuse. */
+    RPMFILE_MISSINGOK	= (1 <<  3),	/*!< from %%config(missingok) */
+    RPMFILE_NOREPLACE	= (1 <<  4),	/*!< from %%config(noreplace) */
+    RPMFILE_SPECFILE	= (1 <<  5),	/*!< @todo (unnecessary) marks 1st file in srpm. */
+    RPMFILE_GHOST	= (1 <<  6),	/*!< from %%ghost */
+    RPMFILE_LICENSE	= (1 <<  7),	/*!< from %%license */
+    RPMFILE_README	= (1 <<  8),	/*!< from %%readme */
+    RPMFILE_EXCLUDE	= (1 <<  9),	/*!< from %%exclude */
+    RPMFILE_UNPATCHED	= (1 << 10),	/*!< placeholder (SuSE) */
+    RPMFILE_PUBKEY	= (1 << 11)	/*!< from %%pubkey */
 } rpmfileAttrs;
-#define	RPMFILE_MULTILIB_SHIFT		9
-#define	RPMFILE_MULTILIB(N)		((N) << RPMFILE_MULTILIB_SHIFT)
-#define	RPMFILE_MULTILIB_MASK		RPMFILE_MULTILIB(7)
 
 #define	RPMFILE_ALL	~(RPMFILE_NONE)
 
-/* XXX Check file flags for multilib marker. */
-#define	isFileMULTILIB(_fflags)		((_fflags) & RPMFILE_MULTILIB_MASK)
-
 /**
  * Dependency Attributes.
  */
@@ -468,7 +483,7 @@ typedef	enum rpmsenseFlags_e {
     RPMSENSE_TRIGGERIN	= (1 << 16),	/*!< %triggerin dependency. */
     RPMSENSE_TRIGGERUN	= (1 << 17),	/*!< %triggerun dependency. */
     RPMSENSE_TRIGGERPOSTUN = (1 << 18),	/*!< %triggerpostun dependency. */
-    RPMSENSE_MULTILIB	= (1 << 19),
+	/* (1 << 19) unused. */
     RPMSENSE_SCRIPT_PREP = (1 << 20),	/*!< %prep build dependency. */
     RPMSENSE_SCRIPT_BUILD = (1 << 21),	/*!< %build build dependency. */
     RPMSENSE_SCRIPT_INSTALL = (1 << 22),/*!< %install build dependency. */
@@ -478,7 +493,8 @@ typedef	enum rpmsenseFlags_e {
     RPMSENSE_TRIGGERPREIN = (1 << 25),	/*!< @todo Implement %triggerprein. */
 /*@=enummemuse@*/
     RPMSENSE_KEYRING	= (1 << 26),
-    RPMSENSE_PATCHES	= (1 << 27)
+    RPMSENSE_PATCHES	= (1 << 27),
+    RPMSENSE_CONFIG	= (1 << 28)
 } rpmsenseFlags;
 
 #define	RPMSENSE_SENSEMASK	15	 /* Mask to get senses, ie serial, */
@@ -487,8 +503,6 @@ typedef	enum rpmsenseFlags_e {
 #define	RPMSENSE_TRIGGER	\
 	(RPMSENSE_TRIGGERIN | RPMSENSE_TRIGGERUN | RPMSENSE_TRIGGERPOSTUN)
 
-#define	isDependsMULTILIB(_dflags)	((_dflags) & RPMSENSE_MULTILIB)
-
 #define	_ALL_REQUIRES_MASK	(\
     RPMSENSE_INTERP | \
     RPMSENSE_SCRIPT_PRE | \
@@ -793,11 +807,11 @@ typedef enum fileAction_e {
     FA_ERASE,		/*!< ... to be removed. */
     FA_SKIPNSTATE,	/*!< ... untouched, state "not installed". */
     FA_SKIPNETSHARED,	/*!< ... untouched, state "netshared". */
-    FA_SKIPMULTILIB	/*!< ... untouched. @todo state "multilib" ???. */
+    FA_SKIPCOLOR	/*!< ... untouched, state "wrong color". */
 } fileAction;
 
 #define XFA_SKIPPING(_a)	\
-    ((_a) == FA_SKIP || (_a) == FA_SKIPNSTATE || (_a) == FA_SKIPNETSHARED || (_a) == FA_SKIPMULTILIB)
+    ((_a) == FA_SKIP || (_a) == FA_SKIPNSTATE || (_a) == FA_SKIPNETSHARED || (_a) == FA_SKIPCOLOR)
 
 /**
  * File types.
@@ -920,7 +934,7 @@ typedef enum rpmtransFlags_e {
 /*@-enummemuse@*/
     RPMTRANS_FLAG_KEEPOBSOLETE	= (1 <<  7),	/*!< @todo Document. */
 /*@=enummemuse@*/
-    RPMTRANS_FLAG_MULTILIB	= (1 <<  8),	/*!< @todo Document. */
+	/* (1 << 8) unused. */
     RPMTRANS_FLAG_DIRSTASH	= (1 <<  9),	/*!< from --dirstash */
     RPMTRANS_FLAG_REPACKAGE	= (1 << 10),	/*!< from --repackage */
 
@@ -947,12 +961,11 @@ typedef enum rpmtransFlags_e {
 /*@=enummemuse@*/
     RPMTRANS_FLAG_APPLYONLY	= (1 << 25),
 
-/*@-enummemuse@*/
-    RPMTRANS_FLAG_CHAINSAW	= (1 << 26),
-/*@=enummemuse@*/
+    RPMTRANS_FLAG_ANACONDA	= (1 << 26),
     RPMTRANS_FLAG_NOMD5		= (1 << 27),	/*!< from --nomd5 */
     RPMTRANS_FLAG_NOSUGGEST	= (1 << 28),	/*!< from --nosuggest */
-    RPMTRANS_FLAG_ADDINDEPS	= (1 << 29)
+    RPMTRANS_FLAG_ADDINDEPS	= (1 << 29),	/*!< from --aid */
+    RPMTRANS_FLAG_NOCONFIGS	= (1 << 30)	/*!< from --noconfigs */
 } rpmtransFlags;
 
 #define	_noTransScripts		\
