@@ -7,11 +7,14 @@
  */
 
 #include "rpmps.h"
+#include "rpmsw.h"
 
 /*@-exportlocal@*/
 /*@unchecked@*/
 extern int _rpmts_debug;
 /*@unchecked@*/
+extern int _rpmts_stats;
+/*@unchecked@*/
 extern int _fps_debug;
 /*@=exportlocal@*/
 
@@ -59,6 +62,30 @@ typedef enum rpmVSFlags_e {
     RPMVSF_NODSA |		\
     RPMVSF_NORSA )
 
+/**
+ * Indices for timestamps.
+ */
+typedef	enum rpmtsOpX_e {
+    RPMTS_OP_TOTAL		=  0,
+    RPMTS_OP_CHECK		=  1,
+    RPMTS_OP_ORDER		=  2,
+    RPMTS_OP_FINGERPRINT	=  3,
+    RPMTS_OP_REPACKAGE		=  4,
+    RPMTS_OP_INSTALL		=  5,
+    RPMTS_OP_ERASE		=  6,
+    RPMTS_OP_SCRIPTLETS		=  7,
+    RPMTS_OP_COMPRESS		=  8,
+    RPMTS_OP_UNCOMPRESS		=  9,
+    RPMTS_OP_DIGEST		= 10,
+    RPMTS_OP_SIGNATURE		= 11,
+    RPMTS_OP_DBADD		= 12,
+    RPMTS_OP_DBREMOVE		= 13,
+    RPMTS_OP_DBGET		= 14,
+    RPMTS_OP_DBPUT		= 15,
+    RPMTS_OP_DBDEL		= 16,
+    RPMTS_OP_MAX		= 17
+} rpmtsOpX;
+
 #if defined(_RPMTS_INTERNAL)
 
 #include "rpmhash.h"	/* XXX hashTable */
@@ -80,8 +107,8 @@ struct diskspaceInfo_s {
     signed long bneeded;	/*!< No. of blocks needed. */
     signed long ineeded;	/*!< No. of inodes needed. */
     int bsize;			/*!< File system block size. */
-    signed long bavail;		/*!< No. of blocks available. */
-    signed long iavail;		/*!< No. of inodes available. */
+    signed long long bavail;	/*!< No. of blocks available. */
+    signed long long iavail;	/*!< No. of inodes available. */
 };
 
 /** \ingroup rpmts
@@ -114,6 +141,7 @@ struct rpmts_s {
 /*@null@*/
     int (*solve) (rpmts ts, rpmds key, const void * data)
 	/*@modifies ts @*/;	/*!< Search for NEVRA key. */
+/*@relnull@*/
     const void * solveData;	/*!< Solve callback data */
     int nsuggests;		/*!< No. of depCheck suggestions. */
 /*@only@*/ /*@null@*/
@@ -156,6 +184,9 @@ struct rpmts_s {
     int numAvailablePackages;	/*!< No. available package instances. */
 #endif
 
+/*@null@*/
+    rpmte relocateElement;	/*!< Element to use when relocating packages. */
+
 /*@owned@*/ /*@relnull@*/
     rpmte * order;		/*!< Packages sorted by dependencies. */
     int orderCount;		/*!< No. of transaction elements. */
@@ -172,6 +203,8 @@ struct rpmts_s {
     int delta;			/*!< Delta for reallocation. */
     int_32 tid;			/*!< Transaction id. */
 
+    uint_32 color;		/*!< Transaction color bits. */
+
     rpmVSFlags vsflags;		/*!< Signature/digest verification flags. */
 
 /*@observer@*/ /*@dependent@*/ /*@null@*/
@@ -187,9 +220,14 @@ struct rpmts_s {
     size_t pkpktlen;		/*!< Current pubkey packet length. */
     unsigned char pksignid[8];	/*!< Current pubkey fingerprint. */
 
+    struct rpmop_s ops[RPMTS_OP_MAX];
+
 /*@null@*/
     pgpDig dig;			/*!< Current signature/pubkey parameters. */
 
+/*@null@*/
+    Spec spec;			/*!< Spec file control structure. */
+
 /*@refs@*/
     int nrefs;			/*!< Reference count. */
 
@@ -206,7 +244,7 @@ extern "C" {
  * @return		0 on success
  */
 int rpmtsCheck(rpmts ts)
-	/*@globals rpmGlobalMacroContext, fileSystem, internalState @*/
+	/*@globals rpmGlobalMacroContext, h_errno, fileSystem, internalState @*/
 	/*@modifies ts, rpmGlobalMacroContext, fileSystem, internalState @*/;
 
 /** \ingroup rpmts
@@ -226,8 +264,8 @@ int rpmtsCheck(rpmts ts)
  * @return		no. of (added) packages that could not be ordered
  */
 int rpmtsOrder(rpmts ts)
-	/*@globals fileSystem, internalState@*/
-	/*@modifies ts, fileSystem, internalState @*/;
+	/*@globals rpmGlobalMacroContext, h_errno, fileSystem, internalState @*/
+	/*@modifies ts, rpmGlobalMacroContext, fileSystem, internalState @*/;
 
 /** \ingroup rpmts
  * Process all package elements in a transaction set.
@@ -238,10 +276,8 @@ int rpmtsOrder(rpmts ts)
  * @return		0 on success, -1 on error, >0 with newProbs set
  */
 int rpmtsRun(rpmts ts, rpmps okProbs, rpmprobFilterFlags ignoreSet)
-	/*@globals rpmGlobalMacroContext,
-		fileSystem, internalState@*/
-	/*@modifies ts, rpmGlobalMacroContext,
-		fileSystem, internalState @*/;
+	/*@globals rpmGlobalMacroContext, h_errno, fileSystem, internalState @*/
+	/*@modifies ts, rpmGlobalMacroContext, fileSystem, internalState @*/;
 
 /** \ingroup rpmts
  * Unreference a transaction instance.
@@ -295,7 +331,7 @@ int rpmtsCloseDB(rpmts ts)
  * @return		0 on success
  */
 int rpmtsOpenDB(rpmts ts, int dbmode)
-	/*@globals rpmGlobalMacroContext, fileSystem, internalState @*/
+	/*@globals rpmGlobalMacroContext, h_errno, fileSystem, internalState @*/
 	/*@modifies ts, rpmGlobalMacroContext, fileSystem, internalState @*/;
 
 /** \ingroup rpmts
@@ -306,7 +342,7 @@ int rpmtsOpenDB(rpmts ts, int dbmode)
  * @return		0 on success
  */
 int rpmtsInitDB(rpmts ts, int dbmode)
-	/*@globals rpmGlobalMacroContext, fileSystem, internalState @*/
+	/*@globals rpmGlobalMacroContext, h_errno, fileSystem, internalState @*/
 	/*@modifies ts, rpmGlobalMacroContext, fileSystem, internalState @*/;
 
 /** \ingroup rpmts
@@ -315,7 +351,7 @@ int rpmtsInitDB(rpmts ts, int dbmode)
  * @return		0 on success
  */
 int rpmtsRebuildDB(rpmts ts)
-	/*@globals rpmGlobalMacroContext, fileSystem, internalState @*/
+	/*@globals rpmGlobalMacroContext, h_errno, fileSystem, internalState @*/
 	/*@modifies ts, rpmGlobalMacroContext, fileSystem, internalState @*/;
 
 /** \ingroup rpmts
@@ -324,7 +360,7 @@ int rpmtsRebuildDB(rpmts ts)
  * @return		0 on success
  */
 int rpmtsVerifyDB(rpmts ts)
-	/*@globals rpmGlobalMacroContext, fileSystem, internalState @*/
+	/*@globals rpmGlobalMacroContext, h_errno, fileSystem, internalState @*/
 	/*@modifies ts, rpmGlobalMacroContext, fileSystem, internalState @*/;
 
 /** \ingroup rpmts
@@ -338,7 +374,7 @@ int rpmtsVerifyDB(rpmts ts)
 /*@only@*/ /*@null@*/
 rpmdbMatchIterator rpmtsInitIterator(const rpmts ts, rpmTag rpmtag,
 			/*@null@*/ const void * keyp, size_t keylen)
-	/*@globals rpmGlobalMacroContext, fileSystem, internalState @*/
+	/*@globals rpmGlobalMacroContext, h_errno, fileSystem, internalState @*/
 	/*@modifies ts, rpmGlobalMacroContext, fileSystem, internalState @*/;
 
 /**
@@ -347,7 +383,7 @@ rpmdbMatchIterator rpmtsInitIterator(const rpmts ts, rpmTag rpmtag,
  * @return		RPMRC_OK on success, RPMRC_NOKEY if not found
  */
 rpmRC rpmtsFindPubkey(rpmts ts)
-	/*@globals rpmGlobalMacroContext, fileSystem, internalState @*/
+	/*@globals rpmGlobalMacroContext, h_errno, fileSystem, internalState @*/
 	/*@modifies ts, rpmGlobalMacroContext, fileSystem, internalState */;
 
 /** \ingroup rpmts
@@ -369,7 +405,7 @@ int rpmtsCloseSDB(rpmts ts)
  */
 /*@-exportlocal@*/
 int rpmtsOpenSDB(rpmts ts, int dbmode)
-	/*@globals rpmGlobalMacroContext, fileSystem, internalState @*/
+	/*@globals rpmGlobalMacroContext, h_errno, fileSystem, internalState @*/
 	/*@modifies ts, rpmGlobalMacroContext, fileSystem, internalState @*/;
 /*@=exportlocal@*/
 
@@ -382,7 +418,7 @@ int rpmtsOpenSDB(rpmts ts, int dbmode)
  */
 /*@-exportlocal@*/
 int rpmtsSolve(rpmts ts, rpmds ds, const void * data)
-	/*@globals rpmGlobalMacroContext, fileSystem, internalState @*/
+	/*@globals rpmGlobalMacroContext, h_errno, fileSystem, internalState @*/
 	/*@modifies ts, rpmGlobalMacroContext, fileSystem, internalState @*/;
 /*@=exportlocal@*/
 
@@ -431,16 +467,16 @@ void rpmtsCleanDig(rpmts ts)
  * @param ts		transaction set
  */
 void rpmtsClean(rpmts ts)
-	/*@globals fileSystem @*/
-	/*@modifies ts, fileSystem @*/;
+	/*@globals fileSystem, internalState @*/
+	/*@modifies ts, fileSystem , internalState@*/;
 
 /** \ingroup rpmts
  * Re-create an empty transaction set.
  * @param ts		transaction set
  */
 void rpmtsEmpty(rpmts ts)
-	/*@globals fileSystem @*/
-	/*@modifies ts, fileSystem @*/;
+	/*@globals fileSystem, internalState @*/
+	/*@modifies ts, fileSystem, internalState @*/;
 
 /** \ingroup rpmts
  * Destroy transaction set, closing the database as well.
@@ -449,8 +485,8 @@ void rpmtsEmpty(rpmts ts)
  */
 /*@null@*/
 rpmts rpmtsFree(/*@killref@*/ /*@only@*//*@null@*/ rpmts ts)
-	/*@globals fileSystem @*/
-	/*@modifies ts, fileSystem @*/;
+	/*@globals fileSystem, internalState @*/
+	/*@modifies ts, fileSystem, internalState @*/;
 
 /** \ingroup rpmts
  * Get verify signatures flag(s).
@@ -703,8 +739,9 @@ int rpmtsNElements(rpmts ts)
  * Return (ordered) transaction set element.
  * @param ts		transaction set
  * @param ix		transaction element index
- * @return		transaction element
+ * @return		transaction element (or NULL)
  */
+/*@null@*/ /*@dependent@*/
 rpmte rpmtsElement(rpmts ts, int ix)
 	/*@*/;
 
@@ -733,6 +770,71 @@ rpmtransFlags rpmtsFlags(rpmts ts)
 rpmtransFlags rpmtsSetFlags(rpmts ts, rpmtransFlags transFlags)
 	/*@modifies ts @*/;
 
+/** \ingroup rpmts
+ * Get spec control structure from transaction set.
+ * @param ts		transaction set
+ * @return		spec control structure
+ */
+/*@null@*/
+Spec rpmtsSpec(rpmts ts)
+	/*@*/;
+
+/** \ingroup rpmts
+ * Set a spec control structure in transaction set.
+ * @param ts		transaction set
+ * @param spec		new spec control structure
+ * @return		previous spec control structure
+ */
+/*@null@*/
+Spec rpmtsSetSpec(rpmts ts, /*@null@*/ Spec spec)
+	/*@modifies ts @*/;
+
+/** \ingroup rpmts
+ * Get current relocate transaction element.
+ * @param ts		transaction set
+ * @return		current relocate transaction element
+ */
+/*@null@*/
+rpmte rpmtsRelocateElement(rpmts ts)
+	/*@*/;
+
+/** \ingroup rpmts
+ * Set current relocate transaction element.
+ * @param ts		transaction set
+ * @param relocateElement new relocate transaction element
+ * @return		previous relocate transaction element
+ */
+/*@null@*/
+rpmte rpmtsSetRelocateElement(rpmts ts, /*@null@*/ rpmte relocateElement)
+	/*@modifies ts @*/;
+
+/**
+ * Retrieve color bits of transaction set.
+ * @param ts		transaction set
+ * @return		color bits
+ */
+uint_32 rpmtsColor(rpmts ts)
+	/*@*/;
+
+/**
+ * Set color bits of transaction set.
+ * @param ts		transaction set
+ * @param color		new color bits
+ * @return		previous color bits
+ */
+uint_32 rpmtsSetColor(rpmts ts, uint_32 color)
+	/*@modifies ts @*/;
+
+/**
+ * Retrieve operation timestamp from a transaction set.
+ * @param ts		transaction set
+ * @param opx		operation timestamp index
+ * @return		pointer to operation timestamp.
+ */
+/*@null@*/
+rpmop rpmtsOp(rpmts ts, rpmtsOpX opx)
+	/*@*/;
+
 /** \ingroup rpmts
  * Set transaction notify callback function and argument.
  *
@@ -755,7 +857,8 @@ int rpmtsSetNotifyCallback(rpmts ts,
  */
 /*@newref@*/
 rpmts rpmtsCreate(void)
-	/*@*/;
+	/*@globals rpmGlobalMacroContext, h_errno, internalState @*/
+	/*@modifies rpmGlobalMacroContext, internalState @*/;
 
 /** \ingroup rpmts
  * Add package to be installed to transaction set.
@@ -773,7 +876,7 @@ rpmts rpmtsCreate(void)
 int rpmtsAddInstallElement(rpmts ts, Header h,
 		/*@exposed@*/ /*@null@*/ const fnpyKey key, int upgrade,
 		/*@null@*/ rpmRelocation * relocs)
-	/*@globals rpmcliPackagesTotal, rpmGlobalMacroContext,
+	/*@globals rpmcliPackagesTotal, rpmGlobalMacroContext, h_errno,
 		fileSystem, internalState @*/
 	/*@modifies ts, h, rpmcliPackagesTotal, rpmGlobalMacroContext,
 		fileSystem, internalState @*/;
@@ -801,8 +904,8 @@ int rpmtsAddEraseElement(rpmts ts, Header h, int dboffset)
 int rpmtsGetKeys(rpmts ts,
 		/*@null@*/ /*@out@*/ fnpyKey ** ep,
 		/*@null@*/ /*@out@*/ int * nep)
-	/*@globals fileSystem @*/
-	/*@modifies ts, ep, nep, fileSystem @*/;
+	/*@globals fileSystem, internalState @*/
+	/*@modifies ts, ep, nep, fileSystem, internalState @*/;
 
 /**
  * Return (malloc'd) header name-version-release string.
@@ -813,6 +916,14 @@ int rpmtsGetKeys(rpmts ts,
 /*@only@*/ char * hGetNEVR(Header h, /*@null@*/ /*@out@*/ const char ** np )
 	/*@modifies *np @*/;
 
+/**
+ * Return header color.
+ * @param h		header
+ * @return		header color
+ */
+uint_32 hGetColor(Header h)
+	/*@modifies h @*/;
+
 #ifdef __cplusplus
 }
 #endif
