@@ -18,10 +18,10 @@
 #include "header_internal.h"
 #include "debug.h"
 
+/*@access FD_t@*/		/* XXX ufdio->read arg1 is void ptr */
 /*@access Header@*/		/* XXX compared with NULL */
 /*@access entryInfo @*/		/* XXX rpmReadSignature */
 /*@access indexEntry @*/	/* XXX rpmReadSignature */
-/*@access FD_t@*/		/* XXX compared with NULL */
 /*@access DIGEST_CTX@*/		/* XXX compared with NULL */
 /*@access pgpDig@*/
 /*@access pgpDigParams@*/
@@ -178,7 +178,7 @@ rpmRC rpmReadSignature(FD_t fd, Header * sighp, sigType sig_type,
     memset(block, 0, sizeof(block));
     if ((xx = timedRead(fd, (char *)block, sizeof(block))) != sizeof(block)) {
 	(void) snprintf(buf, sizeof(buf),
-		_("sigh size(%d): BAD, read returned %d\n"), sizeof(block), xx);
+		_("sigh size(%d): BAD, read returned %d\n"), (int)sizeof(block), xx);
 	goto exit;
     }
     if (memcmp(block, header_magic, sizeof(header_magic))) {
@@ -215,7 +215,7 @@ rpmRC rpmReadSignature(FD_t fd, Header * sighp, sigType sig_type,
     dataStart = (unsigned char *) (pe + il);
     if ((xx = timedRead(fd, (char *)pe, nb)) != nb) {
 	(void) snprintf(buf, sizeof(buf),
-		_("sigh blob(%d): BAD, read returned %d\n"), nb, xx);
+		_("sigh blob(%d): BAD, read returned %d\n"), (int)nb, xx);
 	goto exit;
     }
     
@@ -724,7 +724,7 @@ exit:
     }
     SHA1 = _free(SHA1);
     h = headerFree(h);
-    if (fd) (void) Fclose(fd);
+    if (fd != NULL) (void) Fclose(fd);
     return ret;
 }
 
@@ -985,11 +985,11 @@ verifySizeSignature(const rpmts ts, /*@out@*/ char * t)
     if (size != dig->nbytes) {
 	res = RPMRC_FAIL;
 	t = stpcpy(t, rpmSigString(res));
-	sprintf(t, " Expected(%d) != (%d)\n", size, dig->nbytes);
+	sprintf(t, " Expected(%d) != (%d)\n", (int)size, (int)dig->nbytes);
     } else {
 	res = RPMRC_OK;
 	t = stpcpy(t, rpmSigString(res));
-	sprintf(t, " (%d)", dig->nbytes);
+	sprintf(t, " (%d)", (int)dig->nbytes);
     }
 
 exit:
@@ -1206,7 +1206,7 @@ verifyPGPSignature(rpmts ts, /*@out@*/ char * t,
 	tt = stpcpy(tt, prefix);
 	tt = stpcpy(tt, dig->md5);
 
-	mp32nzero(&dig->rsahm);	mp32nsethex(&dig->rsahm, hexstr);
+	mpnzero(&dig->rsahm);	mpnsethex(&dig->rsahm, hexstr);
 
 	hexstr = _free(hexstr);
 
@@ -1296,7 +1296,7 @@ verifyGPGSignature(rpmts ts, /*@out@*/ char * t,
 #endif
 	xx = rpmDigestFinal(ctx, (void **)&dig->sha1, &dig->sha1len, 1);
 
-	mp32nzero(&dig->hm);	mp32nsethex(&dig->hm, dig->sha1);
+	mpnzero(&dig->hm);	mpnsethex(&dig->hm, dig->sha1);
 
 	/* Compare leading 16 bits of digest for quick check. */
 	signhash16[0] = (*dig->hm.data >> 24) & 0xff;
