@@ -30,11 +30,9 @@
 
 #include "debug.h"
 
-/*@access FD_t @*/		/* XXX compared with NULL */
-/*@access Header @*/		/* XXX compared with NULL */
+/*@access Header @*/		/* XXX ts->notify arg1 is void ptr */
 /*@access rpmps @*/	/* XXX need rpmProblemSetOK() */
 /*@access dbiIndexSet @*/
-/*@access rpmdb @*/
 
 /*@access rpmpsm @*/
 
@@ -82,129 +80,12 @@ static int sharedCmp(const void * one, const void * two)
 }
 
 /**
- */
-/*@-boundsread@*/
-static fileAction decideFileFate(const rpmts ts,
-		const rpmfi ofi, rpmfi nfi)
-	/*@globals fileSystem, internalState @*/
-	/*@modifies nfi, fileSystem, internalState @*/
-{
-    const char * fn = rpmfiFN(nfi);
-    int newFlags = rpmfiFFlags(nfi);
-    char buffer[1024];
-    fileTypes dbWhat, newWhat, diskWhat;
-    struct stat sb;
-    int save = (newFlags & RPMFILE_NOREPLACE) ? FA_ALTNAME : FA_SAVE;
-
-    if (lstat(fn, &sb)) {
-	/*
-	 * The file doesn't exist on the disk. Create it unless the new
-	 * package has marked it as missingok, or allfiles is requested.
-	 */
-	if (!(rpmtsFlags(ts) & RPMTRANS_FLAG_ALLFILES)
-	 && (newFlags & RPMFILE_MISSINGOK))
-	{
-	    rpmMessage(RPMMESS_DEBUG, _("%s skipped due to missingok flag\n"),
-			fn);
-	    return FA_SKIP;
-	} else {
-	    return FA_CREATE;
-	}
-    }
-
-    diskWhat = whatis((int_16)sb.st_mode);
-    dbWhat = whatis(rpmfiFMode(ofi));
-    newWhat = whatis(rpmfiFMode(nfi));
-
-    /*
-     * RPM >= 2.3.10 shouldn't create config directories -- we'll ignore
-     * them in older packages as well.
-     */
-    if (newWhat == XDIR)
-	return FA_CREATE;
-
-    if (diskWhat != newWhat)
-	return save;
-    else if (newWhat != dbWhat && diskWhat != dbWhat)
-	return save;
-    else if (dbWhat != newWhat)
-	return FA_CREATE;
-    else if (dbWhat != LINK && dbWhat != REG)
-	return FA_CREATE;
-
-    /*
-     * This order matters - we'd prefer to CREATE the file if at all
-     * possible in case something else (like the timestamp) has changed.
-     */
-    if (dbWhat == REG) {
-	const unsigned char * omd5, * nmd5;
-	if (domd5(fn, buffer, 0, NULL))
-	    return FA_CREATE;	/* assume file has been removed */
-	omd5 = rpmfiMD5(ofi);
-	if (omd5 && !memcmp(omd5, buffer, 16))
-	    return FA_CREATE;	/* unmodified config file, replace. */
-	nmd5 = rpmfiMD5(nfi);
-/*@-nullpass@*/
-	if (omd5 && nmd5 && !memcmp(omd5, nmd5, 16))
-	    return FA_SKIP;	/* identical file, don't bother. */
-/*@=nullpass@*/
-    } else /* dbWhat == LINK */ {
-	const char * oFLink, * nFLink;
-	memset(buffer, 0, sizeof(buffer));
-	if (readlink(fn, buffer, sizeof(buffer) - 1) == -1)
-	    return FA_CREATE;	/* assume file has been removed */
-	oFLink = rpmfiFLink(ofi);
-	if (oFLink && !strcmp(oFLink, buffer))
-	    return FA_CREATE;	/* unmodified config file, replace. */
-	nFLink = rpmfiFLink(nfi);
-/*@-nullpass@*/
-	if (oFLink && nFLink && !strcmp(oFLink, nFLink))
-	    return FA_SKIP;	/* identical file, don't bother. */
-/*@=nullpass@*/
-     }
-
-    /*
-     * The config file on the disk has been modified, but
-     * the ones in the two packages are different. It would
-     * be nice if RPM was smart enough to at least try and
-     * merge the difference ala CVS, but...
-     */
-    return save;
-}
-/*@=boundsread@*/
-
-/**
- */
-/*@-boundsread@*/
-static int filecmp(rpmfi afi, rpmfi bfi)
-	/*@*/
-{
-    fileTypes awhat = whatis(rpmfiFMode(afi));
-    fileTypes bwhat = whatis(rpmfiFMode(bfi));
-
-    if (awhat != bwhat) return 1;
-
-    if (awhat == LINK) {
-	const char * alink = rpmfiFLink(afi);
-	const char * blink = rpmfiFLink(bfi);
-	if (alink == blink) return 0;
-	if (alink == NULL) return 1;
-	if (blink == NULL) return -1;
-	return strcmp(alink, blink);
-    } else if (awhat == REG) {
-	const unsigned char * amd5 = rpmfiMD5(afi);
-	const unsigned char * bmd5 = rpmfiMD5(bfi);
-	if (amd5 == bmd5) return 0;
-	if (amd5 == NULL) return 1;
-	if (bmd5 == NULL) return -1;
-	return memcmp(amd5, bmd5, 16);
-    }
-
-    return 0;
-}
-/*@=boundsread@*/
-
-/**
+ * @param ts		transaction set
+ * @param p
+ * @param fi		file info set
+ * @param shared
+ * @param sharedCount
+ * @param reportConflicts
  */
 /* XXX only ts->{probs,rpmdb} modified */
 /*@-bounds@*/
@@ -215,6 +96,9 @@ static int handleInstInstalledFiles(const rpmts ts,
 	/*@globals rpmGlobalMacroContext, fileSystem, internalState @*/
 	/*@modifies ts, fi, rpmGlobalMacroContext, fileSystem, internalState @*/
 {
+    uint_32 tscolor = rpmtsColor(ts);
+    uint_32 otecolor, tecolor;
+    uint_32 oFColor, FColor;
     const char * altNEVR = NULL;
     rpmfi otherFi = NULL;
     int numReplaced = 0;
@@ -235,6 +119,18 @@ static int handleInstInstalledFiles(const rpmts ts,
 	mi = rpmdbFreeIterator(mi);
     }
 
+    /* Compute package color. */
+    tecolor = rpmteColor(p);
+    tecolor &= tscolor;
+
+    /* Compute other pkg color. */
+    otecolor = 0;
+    otherFi = rpmfiInit(otherFi, 0);
+    if (otherFi != NULL)
+    while (rpmfiNext(otherFi) >= 0)
+	otecolor |= rpmfiFColor(otherFi);
+    otecolor &= tscolor;
+
     if (otherFi == NULL)
 	return 1;
 
@@ -247,9 +143,13 @@ static int handleInstInstalledFiles(const rpmts ts,
 
 	otherFileNum = shared->otherFileNum;
 	(void) rpmfiSetFX(otherFi, otherFileNum);
+	oFColor = rpmfiFColor(otherFi);
+	oFColor &= tscolor;
 
 	fileNum = shared->pkgFileNum;
 	(void) rpmfiSetFX(fi, fileNum);
+	FColor = rpmfiFColor(fi);
+	FColor &= tscolor;
 
 	isCfgFile = ((rpmfiFFlags(otherFi) | rpmfiFFlags(fi)) & RPMFILE_CONFIG);
 
@@ -262,7 +162,22 @@ static int handleInstInstalledFiles(const rpmts ts,
 	if (XFA_SKIPPING(fi->actions[fileNum]))
 	    continue;
 
-	if (filecmp(otherFi, fi)) {
+	if (rpmfiCompare(otherFi, fi)) {
+
+	    /* Resolve file conflicts to prefer Elf64 (if not forced). */
+	    if (tscolor != 0 && FColor != 0 && FColor != oFColor
+	     && reportConflicts)
+	    {
+		if (oFColor & 0x2) {
+		    fi->actions[fileNum] = FA_SKIP;
+		    reportConflicts = 0;
+		} else
+		if (FColor & 0x2) {
+		    fi->actions[fileNum] = FA_CREATE;
+		    reportConflicts = 0;
+		}
+	    }
+
 	    if (reportConflicts) {
 		rpmpsAppend(ps, RPMPROB_FILE_CONFLICT,
 			rpmteNEVR(p), rpmteKey(p),
@@ -270,7 +185,8 @@ static int handleInstInstalledFiles(const rpmts ts,
 			altNEVR,
 			0);
 	    }
-	    if (!isCfgFile) {
+	    /* Save file identifier to mark as state REPLACED. */
+	    if ( !(isCfgFile || XFA_SKIPPING(fi->actions[fileNum])) ) {
 		/*@-assignexpose@*/ /* FIX: p->replaced, not fi */
 		if (!shared->isRemoved)
 		    fi->replaced[numReplaced++] = *shared;
@@ -278,9 +194,11 @@ static int handleInstInstalledFiles(const rpmts ts,
 	    }
 	}
 
+	/* Determine config file dispostion, skipping missing files (if any). */
 	if (isCfgFile) {
-	    fileAction action;
-	    action = decideFileFate(ts, otherFi, fi);
+	    int skipMissing =
+		((rpmtsFlags(ts) & RPMTRANS_FLAG_ALLFILES) ? 0 : 1);
+	    fileAction action = rpmfiDecideFate(otherFi, fi, skipMissing);
 	    fi->actions[fileNum] = action;
 	}
 	fi->replacedSizes[fileNum] = rpmfiFSize(otherFi);
@@ -508,6 +426,8 @@ static void handleOverlappedFiles(const rpmts ts,
     fi = rpmfiInit(fi, 0);
     if (fi != NULL)
     while ((i = rpmfiNext(fi)) >= 0) {
+	uint_32 tscolor = rpmtsColor(ts);
+	uint_32 oFColor, FColor;
 	struct fingerPrint_s * fiFps;
 	int otherPkgNum, otherFileNum;
 	rpmfi otherFi;
@@ -523,6 +443,8 @@ static void handleOverlappedFiles(const rpmts ts,
 	fiFps = fi->fps + i;
 	FFlags = rpmfiFFlags(fi);
 	FMode = rpmfiFMode(fi);
+	FColor = rpmfiFColor(fi);
+	FColor &= tscolor;
 
 	fixupSize = 0;
 
@@ -584,10 +506,17 @@ static void handleOverlappedFiles(const rpmts ts,
 		/*@innerbreak@*/ break;
 	}
 
+	oFColor = rpmfiFColor(otherFi);
+	oFColor &= tscolor;
+
 /*@-boundswrite@*/
 	switch (rpmteType(p)) {
 	case TR_ADDED:
 	  { struct stat sb;
+	    int reportConflicts =
+		!(rpmtsFilterFlags(ts) & RPMPROB_FILTER_REPLACENEWFILES);
+	    int done = 0;
+
 	    if (otherPkgNum < 0) {
 		/* XXX is this test still necessary? */
 		if (fi->actions[i] != FA_UNKNOWN)
@@ -604,14 +533,40 @@ static void handleOverlappedFiles(const rpmts ts,
 
 assert(otherFi != NULL);
 	    /* Mark added overlapped non-identical files as a conflict. */
-	    if ((rpmtsFilterFlags(ts) & RPMPROB_FILTER_REPLACENEWFILES)
-	     && filecmp(otherFi, fi))
-	    {
-		rpmpsAppend(ps, RPMPROB_NEW_FILE_CONFLICT,
+	    if (rpmfiCompare(otherFi, fi)) {
+
+		/* Resolve file conflicts to prefer Elf64 (if not forced) ... */
+		if (tscolor != 0 && reportConflicts) {
+		    if (FColor & 0x2) {
+			/* ... last Elf64 file is installed ... */
+			if (!XFA_SKIPPING(fi->actions[i]))
+			    otherFi->actions[otherFileNum] = FA_SKIP;
+			fi->actions[i] = FA_CREATE;
+			reportConflicts = 0;
+		    } else
+		    if (oFColor & 0x2) {
+			/* ... first Elf64 file is installed ... */
+			if (XFA_SKIPPING(fi->actions[i]))
+			    otherFi->actions[otherFileNum] = FA_CREATE;
+			fi->actions[i] = FA_SKIP;
+			reportConflicts = 0;
+		    } else
+		    if (FColor == 0 && oFColor == 0) {
+			/* ... otherwise, do both, last in wins. */
+			otherFi->actions[otherFileNum] = FA_CREATE;
+			fi->actions[i] = FA_CREATE;
+			reportConflicts = 0;
+		    }
+		    done = 1;
+		}
+
+		if (reportConflicts) {
+		    rpmpsAppend(ps, RPMPROB_NEW_FILE_CONFLICT,
 			rpmteNEVR(p), rpmteKey(p),
 			fn, NULL,
 			rpmteNEVR(otherFi->te),
 			0);
+		}
 	    }
 
 	    /* Try to get the disk accounting correct even if a conflict. */
@@ -622,7 +577,8 @@ assert(otherFi != NULL);
 		fi->actions[i] = (FFlags & RPMFILE_NOREPLACE)
 			? FA_ALTNAME : FA_SKIP;
 	    } else {
-		fi->actions[i] = FA_CREATE;
+		if (!done)
+		    fi->actions[i] = FA_CREATE;
 	    }
 	  } /*@switchbreak@*/ break;
 
@@ -661,8 +617,8 @@ assert(otherFi != NULL);
 /*@=boundswrite@*/
 
 	/* Update disk space info for a file. */
-	rpmtsUpdateDSI(ts, fiFps->entry->dev,
-                rpmfiFSize(fi), fi->replacedSizes[i], fixupSize, fi->actions[i]);
+	rpmtsUpdateDSI(ts, fiFps->entry->dev, rpmfiFSize(fi),
+		fi->replacedSizes[i], fixupSize, fi->actions[i]);
 
     }
     ps = rpmpsFree(ps);
@@ -722,6 +678,9 @@ static int ensureOlder(rpmts ts,
 }
 
 /**
+ * Skip any files that do not match install policies.
+ * @param ts		transaction set
+ * @param fi		file info set
  */
 /*@-mustmod@*/ /* FIX: fi->actions is modified. */
 /*@-bounds@*/
@@ -729,6 +688,9 @@ static void skipFiles(const rpmts ts, rpmfi fi)
 	/*@globals rpmGlobalMacroContext @*/
 	/*@modifies fi, rpmGlobalMacroContext @*/
 {
+    uint_32 tscolor = rpmtsColor(ts);
+    uint_32 FColor;
+    int noConfigs = (rpmtsFlags(ts) & RPMTRANS_FLAG_NOCONFIGS);
     int noDocs = (rpmtsFlags(ts) & RPMTRANS_FLAG_NODOCS);
     char ** netsharedPaths = NULL;
     const char ** languages;
@@ -773,7 +735,7 @@ static void skipFiles(const rpmts ts, rpmfi fi)
     if (fi != NULL)	/* XXX lclint */
     while ((i = rpmfiNext(fi)) >= 0)
     {
-	char **nsp;
+	char ** nsp;
 
 	bn = rpmfiBN(fi);
 	bnlen = strlen(bn);
@@ -787,7 +749,15 @@ static void skipFiles(const rpmts ts, rpmfi fi)
 
 	/* Don't bother with skipped files */
 	if (XFA_SKIPPING(fi->actions[i])) {
-	    drc[ix]--;
+	    drc[ix]--; dff[ix] = 1;
+	    continue;
+	}
+
+	/* Ignore colored files not in our rainbow. */
+	FColor = rpmfiFColor(fi);
+	if (tscolor && FColor && !(tscolor & FColor)) {
+	    drc[ix]--;	dff[ix] = 1;
+	    fi->actions[i] = FA_SKIPCOLOR;
 	    continue;
 	}
 
@@ -853,6 +823,15 @@ static void skipFiles(const rpmts ts, rpmfi fi)
 	    }
 	}
 
+	/*
+	 * Skip config files if requested.
+	 */
+	if (noConfigs && (rpmfiFFlags(fi) & RPMFILE_CONFIG)) {
+	    drc[ix]--;	dff[ix] = 1;
+	    fi->actions[i] = FA_SKIPNSTATE;
+	    continue;
+	}
+
 	/*
 	 * Skip documentation if requested.
 	 */
@@ -954,6 +933,7 @@ rpmfi rpmtsiFi(const rpmtsi tsi)
 
 int rpmtsRun(rpmts ts, rpmps okProbs, rpmprobFilterFlags ignoreSet)
 {
+    uint_32 tscolor = rpmtsColor(ts);
     int i, j;
     int ourrc = 0;
     int totalFileCount = 0;
@@ -971,15 +951,16 @@ int rpmtsRun(rpmts ts, rpmps okProbs, rpmprobFilterFlags ignoreSet)
     int numRemoved;
     int xx;
 
-    /* FIXME: what if the same package is included in ts twice? */
+    /* XXX programmer error segfault avoidance. */
+    if (rpmtsNElements(ts) <= 0)
+	return -1;
 
     if (rpmtsFlags(ts) & RPMTRANS_FLAG_NOSCRIPTS)
 	(void) rpmtsSetFlags(ts, (rpmtsFlags(ts) | _noTransScripts | _noTransTriggers));
     if (rpmtsFlags(ts) & RPMTRANS_FLAG_NOTRIGGERS)
 	(void) rpmtsSetFlags(ts, (rpmtsFlags(ts) | _noTransTriggers));
 
-    /* XXX MULTILIB is broken, as packages can and do execute /sbin/ldconfig. */
-    if (rpmtsFlags(ts) & (RPMTRANS_FLAG_JUSTDB | RPMTRANS_FLAG_MULTILIB))
+    if (rpmtsFlags(ts) & RPMTRANS_FLAG_JUSTDB)
 	(void) rpmtsSetFlags(ts, (rpmtsFlags(ts) | _noTransScripts | _noTransTriggers));
 
     ts->probs = rpmpsFree(ts->probs);
@@ -1018,7 +999,7 @@ int rpmtsRun(rpmts ts, rpmps okProbs, rpmprobFilterFlags ignoreSet)
      * - count files.
      */
 
-rpmMessage(RPMMESS_DEBUG, _("sanity checking %d elments\n"), rpmtsNElements(ts));
+rpmMessage(RPMMESS_DEBUG, _("sanity checking %d elements\n"), rpmtsNElements(ts));
     ps = rpmtsProblems(ts);
     /* The ordering doesn't matter here */
     pi = rpmtsiInit(ts);
@@ -1030,7 +1011,7 @@ rpmMessage(RPMMESS_DEBUG, _("sanity checking %d elments\n"), rpmtsNElements(ts))
 	    continue;	/* XXX can't happen */
 	fc = rpmfiFC(fi);
 
-	if (!(rpmtsFilterFlags(ts) & RPMPROB_FILTER_IGNOREARCH))
+	if (!(rpmtsFilterFlags(ts) & RPMPROB_FILTER_IGNOREARCH) && !tscolor)
 	    if (!archOkay(rpmteA(p)))
 		rpmpsAppend(ps, RPMPROB_BADARCH,
 			rpmteNEVR(p), rpmteKey(p),
@@ -1052,8 +1033,7 @@ rpmMessage(RPMMESS_DEBUG, _("sanity checking %d elments\n"), rpmtsNElements(ts))
 	    mi = rpmdbFreeIterator(mi);
 	}
 
-	/* XXX multilib should not display "already installed" problems */
-	if (!(rpmtsFilterFlags(ts) & RPMPROB_FILTER_REPLACEPKG) && !rpmteMultiLib(p)) {
+	if (!(rpmtsFilterFlags(ts) & RPMPROB_FILTER_REPLACEPKG)) {
 	    mi = rpmtsInitIterator(ts, RPMTAG_NAME, rpmteN(p), 0);
 	    xx = rpmdbSetIteratorRE(mi, RPMTAG_EPOCH, RPMMIRE_DEFAULT,
 				rpmteE(p));
@@ -1061,6 +1041,12 @@ rpmMessage(RPMMESS_DEBUG, _("sanity checking %d elments\n"), rpmtsNElements(ts))
 				rpmteV(p));
 	    xx = rpmdbSetIteratorRE(mi, RPMTAG_RELEASE, RPMMIRE_DEFAULT,
 				rpmteR(p));
+	    if (tscolor) {
+		xx = rpmdbSetIteratorRE(mi, RPMTAG_ARCH, RPMMIRE_DEFAULT,
+				rpmteA(p));
+		xx = rpmdbSetIteratorRE(mi, RPMTAG_OS, RPMMIRE_DEFAULT,
+				rpmteO(p));
+	    }
 
 	    while (rpmdbNextIterator(mi) != NULL) {
 		rpmpsAppend(ps, RPMPROB_PKG_INSTALLED,
@@ -1157,6 +1143,7 @@ rpmMessage(RPMMESS_DEBUG, _("computing %d file fingerprints\n"), totalFileCount)
 	    continue;	/* XXX can't happen */
 	fc = rpmfiFC(fi);
 
+	(void) rpmswEnter(rpmtsOp(ts, RPMTS_OP_FINGERPRINT), 0);
 	fpLookupList(fpc, fi->dnl, fi->bnl, fi->dil, fc, fi->fps);
 	/*@-branchstate@*/
  	fi = rpmfiInit(fi, 0);
@@ -1169,6 +1156,7 @@ rpmMessage(RPMMESS_DEBUG, _("computing %d file fingerprints\n"), totalFileCount)
 	    /*@=dependenttrans@*/
 	}
 	/*@=branchstate@*/
+	(void) rpmswExit(rpmtsOp(ts, RPMTS_OP_FINGERPRINT), fc);
     }
     pi = rpmtsiFree(pi);
 
@@ -1197,6 +1185,7 @@ rpmMessage(RPMMESS_DEBUG, _("computing file dispositions\n"));
 
 	if (fc == 0) continue;
 
+	(void) rpmswEnter(rpmtsOp(ts, RPMTS_OP_FINGERPRINT), 0);
 	/* Extract file info for all files in this package from the database. */
 	matches = xcalloc(fc, sizeof(*matches));
 	if (rpmdbFindFpList(rpmtsGetRdb(ts), fi->fps, matches, fc)) {
@@ -1296,6 +1285,7 @@ rpmMessage(RPMMESS_DEBUG, _("computing file dispositions\n"));
 	case TR_REMOVED:
 	    /*@switchbreak@*/ break;
 	}
+	(void) rpmswExit(rpmtsOp(ts, RPMTS_OP_FINGERPRINT), fc);
     }
     pi = rpmtsiFree(pi);
     ps = rpmpsFree(ps);
@@ -1364,9 +1354,11 @@ rpmMessage(RPMMESS_DEBUG, _("computing file dispositions\n"));
 				7, numRemoved, NULL, ts->notifyData));
 
 		NOTIFY(ts, (NULL, RPMCALLBACK_REPACKAGE_PROGRESS, progress,
-                        numRemoved, NULL, ts->notifyData));
+			numRemoved, NULL, ts->notifyData));
 		progress++;
 
+		(void) rpmswEnter(rpmtsOp(ts, RPMTS_OP_REPACKAGE), 0);
+
 	/* XXX TR_REMOVED needs CPIO_MAP_{ABSOLUTE,ADDDOT} CPIO_ALL_HARDLINKS */
 		fi->mapflags |= CPIO_MAP_ABSOLUTE;
 		fi->mapflags |= CPIO_MAP_ADDDOT;
@@ -1378,6 +1370,8 @@ rpmMessage(RPMMESS_DEBUG, _("computing file dispositions\n"));
 		fi->mapflags &= ~CPIO_MAP_ADDDOT;
 		fi->mapflags &= ~CPIO_ALL_HARDLINKS;
 
+		(void) rpmswExit(rpmtsOp(ts, RPMTS_OP_REPACKAGE), 0);
+
 		/*@switchbreak@*/ break;
 	    }
 	}
@@ -1405,15 +1399,19 @@ rpmMessage(RPMMESS_DEBUG, _("computing file dispositions\n"));
 	    continue;	/* XXX can't happen */
 	
 	psm = rpmpsmNew(ts, p, fi);
+assert(psm != NULL);
 	psm->unorderedSuccessor =
 		(rpmtsiOc(pi) >= rpmtsUnorderedSuccessors(ts, -1) ? 1 : 0);
 
 	switch (rpmteType(p)) {
 	case TR_ADDED:
+	    (void) rpmswEnter(rpmtsOp(ts, RPMTS_OP_INSTALL), 0);
 
 	    pkgKey = rpmteAddedKey(p);
 
-	    rpmMessage(RPMMESS_DEBUG, "========== +++ %s\n", rpmteNEVR(p));
+	    rpmMessage(RPMMESS_DEBUG, "========== +++ %s %s-%s 0x%x\n",
+		rpmteNEVR(p), rpmteA(p), rpmteO(p), rpmteColor(p));
+
 	    p->h = NULL;
 	    /*@-type@*/ /* FIX: rpmte not opaque */
 	    {
@@ -1460,11 +1458,16 @@ rpmMessage(RPMMESS_DEBUG, _("computing file dispositions\n"));
 		{
 		    char * fstates = fi->fstates;
 		    fileAction * actions = fi->actions;
+		    rpmte savep;
 
 		    fi->fstates = NULL;
 		    fi->actions = NULL;
 		    fi = rpmfiFree(fi);
+
+		    savep = rpmtsSetRelocateElement(ts, p);
 		    fi = rpmfiNew(ts, p->h, RPMTAG_BASENAMES, 1);
+		    (void) rpmtsSetRelocateElement(ts, savep);
+
 		    if (fi != NULL) {	/* XXX can't happen */
 			fi->te = p;
 			fi->fstates = _free(fi->fstates);
@@ -1476,11 +1479,6 @@ rpmMessage(RPMMESS_DEBUG, _("computing file dispositions\n"));
 		}
 		psm->fi = rpmfiLink(p->fi, NULL);
 
-		if (rpmteMultiLib(p))
-		    (void) rpmtsSetFlags(ts, (rpmtsFlags(ts) | RPMTRANS_FLAG_MULTILIB));
-		else
-		    (void) rpmtsSetFlags(ts, (rpmtsFlags(ts) & ~RPMTRANS_FLAG_MULTILIB));
-
 /*@-nullstate@*/ /* FIX: psm->fi may be NULL */
 		if (rpmpsmStage(psm, PSM_PKGINSTALL)) {
 		    ourrc++;
@@ -1504,9 +1502,14 @@ rpmMessage(RPMMESS_DEBUG, _("computing file dispositions\n"));
 
 	    p->h = headerFree(p->h);
 
+	    (void) rpmswExit(rpmtsOp(ts, RPMTS_OP_INSTALL), 0);
+
 	    /*@switchbreak@*/ break;
 	case TR_REMOVED:
-	    rpmMessage(RPMMESS_DEBUG, "========== --- %s\n", rpmteNEVR(p));
+	    (void) rpmswEnter(rpmtsOp(ts, RPMTS_OP_ERASE), 0);
+
+	    rpmMessage(RPMMESS_DEBUG, "========== --- %s %s-%s 0x%x\n",
+		rpmteNEVR(p), rpmteA(p), rpmteO(p), rpmteColor(p));
 	    /*
 	     * XXX This has always been a hack, now mostly broken.
 	     * If install failed, then we shouldn't erase.
@@ -1515,6 +1518,9 @@ rpmMessage(RPMMESS_DEBUG, _("computing file dispositions\n"));
 		if (rpmpsmStage(psm, PSM_PKGERASE))
 		    ourrc++;
 	    }
+
+	    (void) rpmswExit(rpmtsOp(ts, RPMTS_OP_ERASE), 0);
+
 	    /*@switchbreak@*/ break;
 	}
 	xx = rpmdbSync(rpmtsGetRdb(ts));
