@@ -28,11 +28,11 @@
 /** \ingroup python
  * \class Rpmhdr
  * \brief A python header object represents an RPM package header.
- * 
+ *
  * All RPM packages have headers that provide metadata for the package.
  * Header objects can be returned by database queries or loaded from a
  * binary package on disk.
- * 
+ *
  * The ts.hdrFromFdno() function returns the package header from a
  * package on disk, verifying package signatures and digests of the
  * package while reading.
@@ -53,7 +53,7 @@
  *	else:
  *	   print "header is from a binary package"
  * \endcode
- * 
+ *
  * The Python interface to the header data is quite elegant.  It
  * presents the data in a dictionary form.  We'll take the header we
  * just loaded and access the data within it:
@@ -156,7 +156,7 @@ static PyObject * hdrUnload(hdrObject * s, PyObject * args, PyObject *keywords)
     static char *kwlist[] = { "legacyHeader", NULL};
 
     if (!PyArg_ParseTupleAndKeywords(args, keywords, "|i", kwlist, &legacy))
-	return NULL; 
+	return NULL;
 
     h = headerLink(s->h);
     /* XXX this legacy switch is a hack, needs to be removed. */
@@ -214,8 +214,8 @@ static void mungeFilelist(Header h)
 	|| !headerIsEntry (h, RPMTAG_DIRNAMES)
 	|| !headerIsEntry (h, RPMTAG_DIRINDEXES))
 	compressFilelist(h);
-    
-    rpmBuildFileList(h, &fileNames, &count);
+
+    rpmfiBuildFNames(h, RPMTAG_BASENAMES, &fileNames, &count);
 
     if (fileNames == NULL || count <= 0)
 	return;
@@ -227,7 +227,7 @@ static void mungeFilelist(Header h)
     fileNames = _free(fileNames);
 }
 
-/** 
+/**
  */
 static PyObject * rhnUnload(hdrObject * s, PyObject * args)
 	/*@*/
@@ -336,6 +336,11 @@ static int hdr_compare(hdrObject * a, hdrObject * b)
     return rpmVersionCompare(a->h, b->h);
 }
 
+static long hdr_hash(PyObject * h)
+{
+    return (long) h;
+}
+
 /** \ingroup python
  */
 /*@unchecked@*/ /*@observer@*/
@@ -382,7 +387,7 @@ static void hdr_dealloc(hdrObject * s)
     s->md5list = _free(s->md5list);
     s->fileList = _free(s->fileList);
     s->linkList = _free(s->linkList);
-    PyMem_DEL(s);
+    PyObject_Del(s);
 }
 
 /** \ingroup python
@@ -442,7 +447,7 @@ static PyObject * hdr_subscript(hdrObject * s, PyObject * item)
             PyErr_SetString(PyExc_KeyError, "unknown header tag");
             return NULL;
         }
-        
+
         if (!rpmHeaderGetEntry(s->h, tag, &type, &data, &count)) {
             Py_INCREF(Py_None);
             return Py_None;
@@ -580,7 +585,7 @@ static char hdr_doc[] =
  */
 /*@unchecked@*/ /*@observer@*/
 PyTypeObject hdr_Type = {
-	PyObject_HEAD_INIT(NULL)
+	PyObject_HEAD_INIT(&PyType_Type)
 	0,				/* ob_size */
 	"rpm.hdr",			/* tp_name */
 	sizeof(hdrObject),		/* tp_size */
@@ -594,7 +599,7 @@ PyTypeObject hdr_Type = {
 	0,				/* tp_as_number */
 	0,	 			/* tp_as_sequence */
 	&hdr_as_mapping,		/* tp_as_mapping */
-	0,				/* tp_hash */
+	hdr_hash,			/* tp_hash */
 	0,				/* tp_call */
 	0,				/* tp_str */
 	0,				/* tp_getattro */
@@ -627,7 +632,7 @@ PyTypeObject hdr_Type = {
 
 hdrObject * hdr_Wrap(Header h)
 {
-    hdrObject * hdr = PyObject_NEW(hdrObject, &hdr_Type);
+    hdrObject * hdr = PyObject_New(hdrObject, &hdr_Type);
     hdr->h = headerLink(h);
     hdr->fileList = hdr->linkList = hdr->md5list = NULL;
     hdr->uids = hdr->gids = hdr->mtimes = hdr->fileSizes = NULL;
@@ -651,7 +656,7 @@ PyObject * hdrLoad(PyObject * self, PyObject * args)
     int len;
 
     if (!PyArg_ParseTuple(args, "s#", &obj, &len)) return NULL;
-    
+
     /* malloc is needed to avoid surprises from data swab in headerLoad(). */
     copy = malloc(len);
     if (copy == NULL) {
@@ -684,7 +689,7 @@ PyObject * rhnLoad(PyObject * self, PyObject * args)
     int len;
 
     if (!PyArg_ParseTuple(args, "s#", &obj, &len)) return NULL;
-    
+
     /* malloc is needed to avoid surprises from data swab in headerLoad(). */
     copy = malloc(len);
     if (copy == NULL) {
@@ -856,6 +861,7 @@ int rpmMergeHeaders(PyObject * list, FD_t fd, int matchTag)
 	}
 
 	headerFreeIterator(hi);
+	h = headerFree(h);
 
 	Py_BEGIN_ALLOW_THREADS
 	h = headerRead(fd, HEADER_MAGIC_YES);
