@@ -10,18 +10,19 @@
 #define PyObject_HEAD   int _PyObjectHead;
 #endif
 
+#include <rpmio_internal.h>
 #include <rpmcli.h>	/* XXX for rpmCheckSig */
 #include <rpmdb.h>
 
 #include "legacy.h"
 #include "misc.h"
 #include "header_internal.h"
-#include "upgrade.h"
 
 #include "header-py.h"
 #include "rpmal-py.h"
 #include "rpmds-py.h"
 #include "rpmfd-py.h"
+#include "rpmfts-py.h"
 #include "rpmfi-py.h"
 #include "rpmmi-py.h"
 #include "rpmrc-py.h"
@@ -57,197 +58,26 @@ static PyObject * archScore(PyObject * self, PyObject * args)
 
 /**
  */
-static int psGetArchScore(Header h)
+static PyObject * setLogFile (PyObject * self, PyObject * args)
 {
-    void * pkgArch;
-    int type, count;
+    PyObject * fop = NULL;
+    FILE * fp = NULL;
 
-    if (!headerGetEntry(h, RPMTAG_ARCH, &type, (void **) &pkgArch, &count) ||
-        type == RPM_INT8_TYPE)
-       return 150;
-    else
-        return rpmMachineScore(RPM_MACHTABLE_INSTARCH, pkgArch);
-}
-
-/**
- */
-static int pkgCompareVer(void * first, void * second)
-{
-    struct packageInfo ** a = first;
-    struct packageInfo ** b = second;
-    int ret, score1, score2;
-
-    /* put packages w/o names at the end */
-    if (!(*a)->name) return 1;
-    if (!(*b)->name) return -1;
-
-    ret = xstrcasecmp((*a)->name, (*b)->name);
-    if (ret) return ret;
-    score1 = psGetArchScore((*a)->h);
-    if (!score1) return 1;
-    score2 = psGetArchScore((*b)->h);
-    if (!score2) return -1;
-    if (score1 < score2) return -1;
-    if (score1 > score2) return 1;
-    return rpmVersionCompare((*b)->h, (*a)->h);
-}
-
-/**
- */
-static void pkgSort(struct pkgSet * psp)
-{
-    int i;
-    char *name;
-
-    if (psp->numPackages <= 0)
-	return;
-
-    qsort(psp->packages, psp->numPackages, sizeof(*psp->packages),
-	 (void *) pkgCompareVer);
-
-    name = psp->packages[0]->name;
-    if (!name) {
-       psp->numPackages = 0;
-       return;
-    }
-    for (i = 1; i < psp->numPackages; i++) {
-       if (!psp->packages[i]->name) break;
-       if (!strcmp(psp->packages[i]->name, name))
-	   psp->packages[i]->name = NULL;
-       else
-	   name = psp->packages[i]->name;
-    }
-
-    qsort(psp->packages, psp->numPackages, sizeof(*psp->packages),
-	 (void *) pkgCompareVer);
-
-    for (i = 0; i < psp->numPackages; i++)
-       if (!psp->packages[i]->name) break;
-    psp->numPackages = i;
-}
-
-/**
- */
-static PyObject * findUpgradeSet(PyObject * self, PyObject * args)
-{
-    PyObject * hdrList, * result;
-    char * root = "/";
-    int i;
-    struct pkgSet list;
-    hdrObject * hdr;
-
-    if (!PyArg_ParseTuple(args, "O|s", &hdrList, &root)) return NULL;
-
-    if (!PyList_Check(hdrList)) {
-	PyErr_SetString(PyExc_TypeError, "list of headers expected");
+    if (!PyArg_ParseTuple(args, "|O:logSetFile", &fop))
 	return NULL;
-    }
 
-    list.numPackages = PyList_Size(hdrList);
-    list.packages = alloca(sizeof(list.packages) * list.numPackages);
-    for (i = 0; i < list.numPackages; i++) {
-	hdr = (hdrObject *) PyList_GetItem(hdrList, i);
-	if (((PyObject *) hdr)->ob_type != &hdr_Type) {
-	    PyErr_SetString(PyExc_TypeError, "list of headers expected");
+    if (fop) {
+	if (!PyFile_Check(fop)) {
+	    PyErr_SetString(pyrpmError, "requires file object");
 	    return NULL;
 	}
-	list.packages[i] = alloca(sizeof(struct packageInfo));
-	list.packages[i]->h = hdrGetHeader(hdr);
-	list.packages[i]->selected = 0;
-	list.packages[i]->data = hdr;
-
-	headerGetEntry(list.packages[i]->h, RPMTAG_NAME, NULL,
-		      (void **) &list.packages[i]->name, NULL);
+	fp = PyFile_AsFile(fop);
     }
 
-    pkgSort (&list);
+    (void) rpmlogSetFile(fp);
 
-    if (ugFindUpgradePackages(&list, root)) {
-	PyErr_SetString(pyrpmError, "error during upgrade check");
-	return NULL;
-    }
-
-    result = PyList_New(0);
-    for (i = 0; i < list.numPackages; i++) {
-	if (list.packages[i]->selected) {
-	    PyList_Append(result, list.packages[i]->data);
-/*  	    Py_DECREF(list.packages[i]->data); */
-	}
-    }
-
-    return result;
-}
-
-/**
- */
-static PyObject * errorCB = NULL;
-static PyObject * errorData = NULL;
-
-/**
- */
-static void errorcb (void)
-{
-    PyObject * result, * args = NULL;
-
-    if (errorData)
-	args = Py_BuildValue("(O)", errorData);
-
-    result = PyEval_CallObject(errorCB, args);
-    Py_XDECREF(args);
-
-    if (result == NULL) {
-	PyErr_Print();
-	PyErr_Clear();
-    }
-    Py_DECREF (result);
-}
-
-/**
- */
-static PyObject * errorSetCallback (PyObject * self, PyObject * args)
-{
-    PyObject *newCB = NULL, *newData = NULL;
-
-    if (!PyArg_ParseTuple(args, "O|O", &newCB, &newData)) return NULL;
-
-    /* if we're getting a void*, set the error callback to this. */
-    /* also, we can possibly decref any python callbacks we had  */
-    /* and set them to NULL.                                     */
-    if (PyCObject_Check (newCB)) {
-	rpmErrorSetCallback (PyCObject_AsVoidPtr(newCB));
-
-	Py_XDECREF (errorCB);
-	Py_XDECREF (errorData);
-
-	errorCB   = NULL;
-	errorData = NULL;
-	
-	Py_INCREF(Py_None);
-	return Py_None;
-    }
-    
-    if (!PyCallable_Check (newCB)) {
-	PyErr_SetString(PyExc_TypeError, "parameter must be callable");
-	return NULL;
-    }
-
-    Py_XDECREF(errorCB);
-    Py_XDECREF(errorData);
-
-    errorCB = newCB;
-    errorData = newData;
-    
-    Py_INCREF (errorCB);
-    Py_XINCREF (errorData);
-
-    return PyCObject_FromVoidPtr(rpmErrorSetCallback (errorcb), NULL);
-}
-
-/**
- */
-static PyObject * errorString (PyObject * self, PyObject * args)
-{
-    return PyString_FromString(rpmErrorString ());
+    Py_INCREF(Py_None);
+    return (PyObject *) Py_None;
 }
 
 /**
@@ -275,6 +105,16 @@ static PyObject * setEpochPromote (PyObject * self, PyObject * args)
     return (PyObject *) Py_None;
 }
 
+/**
+ */
+static PyObject * setStats (PyObject * self, PyObject * args)
+{
+    if (!PyArg_ParseTuple(args, "i", &_rpmts_stats))
+	return NULL;
+    Py_INCREF(Py_None);
+    return (PyObject *) Py_None;
+}
+
 /**
  */
 static PyMethodDef rpmModuleMethods[] = {
@@ -293,8 +133,7 @@ static PyMethodDef rpmModuleMethods[] = {
 
     { "archscore", (PyCFunction) archScore, METH_VARARGS,
 	NULL },
-    { "findUpgradeSet", (PyCFunction) findUpgradeSet, METH_VARARGS,
-	NULL },
+
     { "headerLoad", (PyCFunction) hdrLoad, METH_VARARGS,
 	NULL },
     { "rhnLoad", (PyCFunction) rhnLoad, METH_VARARGS,
@@ -305,10 +144,10 @@ static PyMethodDef rpmModuleMethods[] = {
 	NULL },
     { "readHeaderListFromFile", (PyCFunction) rpmHeaderFromFile, METH_VARARGS,
 	NULL },
-    { "errorSetCallback", (PyCFunction) errorSetCallback, METH_VARARGS,
-	NULL },
-    { "errorString", (PyCFunction) errorString, METH_VARARGS,
+
+    { "setLogFile", (PyCFunction) setLogFile, METH_VARARGS,
 	NULL },
+
     { "versionCompare", (PyCFunction) versionCompare, METH_VARARGS,
 	NULL },
     { "labelCompare", (PyCFunction) labelCompare, METH_VARARGS,
@@ -317,6 +156,9 @@ static PyMethodDef rpmModuleMethods[] = {
 	NULL },
     { "setEpochPromote", (PyCFunction) setEpochPromote, METH_VARARGS,
 	NULL },
+    { "setStats", (PyCFunction) setStats, METH_VARARGS,
+	NULL },
+
     { "dsSingle", (PyCFunction) rpmds_Single, METH_VARARGS,
 "rpm.dsSingle(TagN, N, [EVR, [Flags]] -> ds\n\
 - Create a single element dependency set.\n" },
@@ -344,6 +186,7 @@ void initrpm(void)
     if (PyType_Ready(&rpmal_Type) < 0) return;
     if (PyType_Ready(&rpmds_Type) < 0) return;
     if (PyType_Ready(&rpmfd_Type) < 0) return;
+    if (PyType_Ready(&rpmfts_Type) < 0) return;
     if (PyType_Ready(&rpmfi_Type) < 0) return;
     if (PyType_Ready(&rpmmi_Type) < 0) return;
 
@@ -385,6 +228,9 @@ void initrpm(void)
     Py_INCREF(&rpmfd_Type);
     PyModule_AddObject(m, "fd", (PyObject *) &rpmfd_Type);
 
+    Py_INCREF(&rpmfts_Type);
+    PyModule_AddObject(m, "fts", (PyObject *) &rpmfts_Type);
+
     Py_INCREF(&rpmfi_Type);
     PyModule_AddObject(m, "fi", (PyObject *) &rpmfi_Type);
 
@@ -404,6 +250,7 @@ void initrpm(void)
     rpmal_Type.ob_type = &PyType_Type;
     rpmds_Type.ob_type = &PyType_Type;
     rpmfd_Type.ob_type = &PyType_Type;
+    rpmfts_Type.ob_type = &PyType_Type;
     rpmfi_Type.ob_type = &PyType_Type;
     rpmmi_Type.ob_type = &PyType_Type;
     rpmte_Type.ob_type = &PyType_Type;
@@ -443,14 +290,19 @@ void initrpm(void)
     REGISTER_ENUM(RPMFILE_STATE_REPLACED);
     REGISTER_ENUM(RPMFILE_STATE_NOTINSTALLED);
     REGISTER_ENUM(RPMFILE_STATE_NETSHARED);
+    REGISTER_ENUM(RPMFILE_STATE_WRONGCOLOR);
 
     REGISTER_ENUM(RPMFILE_CONFIG);
     REGISTER_ENUM(RPMFILE_DOC);
+    REGISTER_ENUM(RPMFILE_ICON);
     REGISTER_ENUM(RPMFILE_MISSINGOK);
     REGISTER_ENUM(RPMFILE_NOREPLACE);
     REGISTER_ENUM(RPMFILE_GHOST);
     REGISTER_ENUM(RPMFILE_LICENSE);
     REGISTER_ENUM(RPMFILE_README);
+    REGISTER_ENUM(RPMFILE_EXCLUDE);
+    REGISTER_ENUM(RPMFILE_UNPATCHED);
+    REGISTER_ENUM(RPMFILE_PUBKEY);
 
     REGISTER_ENUM(RPMDEP_SENSE_REQUIRES);
     REGISTER_ENUM(RPMDEP_SENSE_CONFLICTS);
@@ -471,7 +323,6 @@ void initrpm(void)
     REGISTER_ENUM(RPMSENSE_TRIGGERIN);
     REGISTER_ENUM(RPMSENSE_TRIGGERUN);
     REGISTER_ENUM(RPMSENSE_TRIGGERPOSTUN);
-    REGISTER_ENUM(RPMSENSE_MULTILIB);
     REGISTER_ENUM(RPMSENSE_SCRIPT_PREP);
     REGISTER_ENUM(RPMSENSE_SCRIPT_BUILD);
     REGISTER_ENUM(RPMSENSE_SCRIPT_INSTALL);
@@ -480,6 +331,7 @@ void initrpm(void)
     REGISTER_ENUM(RPMSENSE_TRIGGERPREIN);
     REGISTER_ENUM(RPMSENSE_KEYRING);
     REGISTER_ENUM(RPMSENSE_PATCHES);
+    REGISTER_ENUM(RPMSENSE_CONFIG);
 
     REGISTER_ENUM(RPMTRANS_FLAG_TEST);
     REGISTER_ENUM(RPMTRANS_FLAG_BUILD_PROBS);
@@ -489,19 +341,21 @@ void initrpm(void)
     REGISTER_ENUM(RPMTRANS_FLAG_NODOCS);
     REGISTER_ENUM(RPMTRANS_FLAG_ALLFILES);
     REGISTER_ENUM(RPMTRANS_FLAG_KEEPOBSOLETE);
-    REGISTER_ENUM(RPMTRANS_FLAG_MULTILIB);
     REGISTER_ENUM(RPMTRANS_FLAG_REPACKAGE);
+    REGISTER_ENUM(RPMTRANS_FLAG_REVERSE);
     REGISTER_ENUM(RPMTRANS_FLAG_NOPRE);
     REGISTER_ENUM(RPMTRANS_FLAG_NOPOST);
+    REGISTER_ENUM(RPMTRANS_FLAG_NOTRIGGERPREIN);
     REGISTER_ENUM(RPMTRANS_FLAG_NOTRIGGERIN);
     REGISTER_ENUM(RPMTRANS_FLAG_NOTRIGGERUN);
     REGISTER_ENUM(RPMTRANS_FLAG_NOPREUN);
     REGISTER_ENUM(RPMTRANS_FLAG_NOPOSTUN);
     REGISTER_ENUM(RPMTRANS_FLAG_NOTRIGGERPOSTUN);
-    REGISTER_ENUM(RPMTRANS_FLAG_CHAINSAW);
+    REGISTER_ENUM(RPMTRANS_FLAG_ANACONDA);
     REGISTER_ENUM(RPMTRANS_FLAG_NOMD5);
     REGISTER_ENUM(RPMTRANS_FLAG_NOSUGGEST);
     REGISTER_ENUM(RPMTRANS_FLAG_ADDINDEPS);
+    REGISTER_ENUM(RPMTRANS_FLAG_NOCONFIGS);
 
     REGISTER_ENUM(RPMPROB_FILTER_IGNOREOS);
     REGISTER_ENUM(RPMPROB_FILTER_IGNOREARCH);
@@ -576,6 +430,9 @@ void initrpm(void)
     REGISTER_ENUM(_RPMVSF_NOHEADER);
     REGISTER_ENUM(_RPMVSF_NOPAYLOAD);
 
+    REGISTER_ENUM(TR_ADDED);
+    REGISTER_ENUM(TR_REMOVED);
+
     PyDict_SetItemString(d, "RPMAL_NOMATCH", o=PyInt_FromLong( (long)RPMAL_NOMATCH ));
     Py_DECREF(o);
 }
