@@ -14,10 +14,10 @@ typedef struct rpmrcObject_s rpmrcObject;
 struct rpmrcObject_s {
 #if Py_TPFLAGS_HAVE_ITER	/* XXX backport to python-1.5.2 */
     PyDictObject dict;
-    int state;
 #else
     PyObject_HEAD
 #endif
+    PyObject *md_dict;		/*!< to look like PyModuleObject */
 } ;
 
 /*@unchecked@*/
