@@ -9,6 +9,7 @@
 
 typedef struct rpmtsObject_s {
     PyObject_HEAD
+    PyObject *md_dict;		/*!< to look like PyModuleObject */
     rpmts	ts;
     PyObject * keyList;		/* keeps reference counts correct */
     FD_t scriptFd;
@@ -27,6 +28,7 @@ enum {
 };
 
 rpmtsObject * rpmts_Create(PyObject * s, PyObject * args)
-	/*@*/;
+	/*@globals rpmGlobalMacroContext @*/
+	/*@modifies rpmGlobalMacroContext @*/;
 
 #endif
