@@ -191,13 +191,13 @@ static int db_fini(dbiIndex dbi, const char * dbhome,
 	/*@-moduncon@*/ /* FIX: annotate db3 methods */
 	xx = db_env_create(&dbenv, 0);
 	/*@=moduncon@*/
-	xx = cvtdberr(dbi, "db_env_create", rc, _debug);
+	xx = cvtdberr(dbi, "db_env_create", xx, _debug);
 #if (DB_VERSION_MAJOR == 3 && DB_VERSION_MINOR != 0) || (DB_VERSION_MAJOR == 4)
 	xx = dbenv->remove(dbenv, dbhome, 0);
 #else
 	xx = dbenv->remove(dbenv, dbhome, NULL, 0);
 #endif
-	xx = cvtdberr(dbi, "dbenv->remove", rc, _debug);
+	xx = cvtdberr(dbi, "dbenv->remove", xx, _debug);
 
 	if (dbfile)
 	    rpmMessage(RPMMESS_DEBUG, _("removed  db environment %s/%s\n"),
@@ -213,6 +213,55 @@ static int db3_fsync_disable(/*@unused@*/ int fd)
     return 0;
 }
 
+#if HAVE_LIBPTHREAD
+#if HAVE_PTHREAD_H
+#include <pthread.h>
+#endif
+
+/**
+ * Check that posix mutexes are shared.
+ * @return		0 == shared.
+ */
+static int db3_pthread_nptl(void)
+	/*@*/
+{
+    pthread_mutex_t mutex;
+    pthread_mutexattr_t mutexattr, *mutexattrp = NULL;
+    pthread_cond_t cond;
+    pthread_condattr_t condattr, *condattrp = NULL;
+    int ret = 0;
+
+    ret = pthread_mutexattr_init(&mutexattr);
+    if (ret == 0) {
+	ret = pthread_mutexattr_setpshared(&mutexattr, PTHREAD_PROCESS_SHARED);
+	mutexattrp = &mutexattr;
+    }
+
+    if (ret == 0)
+	ret = pthread_mutex_init(&mutex, mutexattrp);
+    if (mutexattrp != NULL)
+	pthread_mutexattr_destroy(mutexattrp);
+    if (ret)
+	return ret;
+    (void) pthread_mutex_destroy(&mutex);
+
+    ret = pthread_condattr_init(&condattr);
+    if (ret == 0) {
+	ret = pthread_condattr_setpshared(&condattr, PTHREAD_PROCESS_SHARED);
+	condattrp = &condattr;
+    }
+
+    if (ret == 0)
+	ret = pthread_cond_init(&cond, condattrp);
+
+    if (condattrp != NULL)
+	(void)pthread_condattr_destroy(condattrp);
+    if (ret == 0)
+	(void) pthread_cond_destroy(&cond);
+    return ret;
+}
+#endif
+
 /*@-moduncon@*/ /* FIX: annotate db3 methods */
 static int db_init(dbiIndex dbi, const char * dbhome,
 		/*@null@*/ const char * dbfile,
@@ -258,37 +307,103 @@ static int db_init(dbiIndex dbi, const char * dbhome,
 
     rc = db_env_create(&dbenv, dbi->dbi_ecflags);
     rc = cvtdberr(dbi, "db_env_create", rc, _debug);
-    if (rc)
+    if (dbenv == NULL || rc)
 	goto errxit;
 
-    if (dbenv == NULL)
-	return 1;
-
   { int xx;
     /*@-noeffectuncon@*/ /* FIX: annotate db3 methods */
+
+ /* 4.1: dbenv->set_app_dispatch(???) */
+ /* 4.1: dbenv->set_alloc(???) */
+ /* 4.1: dbenv->set_data_dir(???) */
+ /* 4.1: dbenv->set_encrypt(???) */
+
     dbenv->set_errcall(dbenv, rpmdb->db_errcall);
     dbenv->set_errfile(dbenv, rpmdb->db_errfile);
     dbenv->set_errpfx(dbenv, rpmdb->db_errpfx);
     /*@=noeffectuncon@*/
+
+ /* 4.1: dbenv->set_feedback(???) */
+ /* 4.1: dbenv->set_flags(???) */
+
  /* dbenv->set_paniccall(???) */
-    xx = dbenv->set_verbose(dbenv, DB_VERB_CHKPOINT,
+
+    if ((dbi->dbi_ecflags & DB_CLIENT) && dbi->dbi_host) {
+	const char * home;
+	int retry = 0;
+
+	if ((home = strrchr(dbhome, '/')) != NULL)
+	    dbhome = ++home;
+
+	while (retry++ < 5) {
+/* XXX 3.3.4 change. */
+#if (DB_VERSION_MAJOR == 3 && DB_VERSION_MINOR == 3) || (DB_VERSION_MAJOR == 4)
+	    xx = dbenv->set_rpc_server(dbenv, NULL, dbi->dbi_host,
+		dbi->dbi_cl_timeout, dbi->dbi_sv_timeout, 0);
+	    xx = cvtdberr(dbi, "dbenv->set_server", xx, _debug);
+#else
+	    xx = dbenv->set_server(dbenv, dbi->dbi_host,
+		dbi->dbi_cl_timeout, dbi->dbi_sv_timeout, 0);
+	    xx = cvtdberr(dbi, "dbenv->set_server", xx, _debug);
+#endif
+	    if (!xx)
+		break;
+	    sleep(15);
+	}
+    } else {
+	xx = dbenv->set_verbose(dbenv, DB_VERB_CHKPOINT,
 		(dbi->dbi_verbose & DB_VERB_CHKPOINT));
-    xx = dbenv->set_verbose(dbenv, DB_VERB_DEADLOCK,
+	xx = dbenv->set_verbose(dbenv, DB_VERB_DEADLOCK,
 		(dbi->dbi_verbose & DB_VERB_DEADLOCK));
-    xx = dbenv->set_verbose(dbenv, DB_VERB_RECOVERY,
+	xx = dbenv->set_verbose(dbenv, DB_VERB_RECOVERY,
 		(dbi->dbi_verbose & DB_VERB_RECOVERY));
-    xx = dbenv->set_verbose(dbenv, DB_VERB_WAITSFOR,
+	xx = dbenv->set_verbose(dbenv, DB_VERB_WAITSFOR,
 		(dbi->dbi_verbose & DB_VERB_WAITSFOR));
- /* dbenv->set_lg_max(???) */
+
+	if (dbi->dbi_mmapsize) {
+	    xx = dbenv->set_mp_mmapsize(dbenv, dbi->dbi_mmapsize);
+	    xx = cvtdberr(dbi, "dbenv->set_mp_mmapsize", xx, _debug);
+	}
+	if (dbi->dbi_tmpdir) {
+	    const char * root;
+	    const char * tmpdir;
+
+	    root = (dbi->dbi_root ? dbi->dbi_root : rpmdb->db_root);
+	    if ((root[0] == '/' && root[1] == '\0') || rpmdb->db_chrootDone)
+		root = NULL;
+/*@-mods@*/
+	    tmpdir = rpmGenPath(root, dbi->dbi_tmpdir, NULL);
+/*@=mods@*/
+	    xx = dbenv->set_tmp_dir(dbenv, tmpdir);
+	    xx = cvtdberr(dbi, "dbenv->set_tmp_dir", xx, _debug);
+	    tmpdir = _free(tmpdir);
+	}
+    }
+
  /* dbenv->set_lk_conflicts(???) */
  /* dbenv->set_lk_detect(???) */
- /* dbenv->set_lk_max(???) */
-    xx = dbenv->set_mp_mmapsize(dbenv, dbi->dbi_mp_mmapsize);
-    xx = cvtdberr(dbi, "dbenv->set_mp_mmapsize", xx, _debug);
-    xx = dbenv->set_cachesize(dbenv, 0, dbi->dbi_mp_size, 0);
-    xx = cvtdberr(dbi, "dbenv->set_cachesize", xx, _debug);
+ /* 4.1: dbenv->set_lk_max_lockers(???) */
+ /* 4.1: dbenv->set_lk_max_locks(???) */
+ /* 4.1: dbenv->set_lk_max_objects(???) */
+
+ /* 4.1: dbenv->set_lg_bsize(???) */
+ /* 4.1: dbenv->set_lg_dir(???) */
+ /* 4.1: dbenv->set_lg_max(???) */
+ /* 4.1: dbenv->set_lg_regionmax(???) */
+
+    if (dbi->dbi_cachesize) {
+	xx = dbenv->set_cachesize(dbenv, 0, dbi->dbi_cachesize, 0);
+	xx = cvtdberr(dbi, "dbenv->set_cachesize", xx, _debug);
+    }
+
+ /* 4.1 dbenv->set_timeout(???) */
  /* dbenv->set_tx_max(???) */
+ /* 4.1: dbenv->set_tx_timestamp(???) */
  /* dbenv->set_tx_recover(???) */
+
+ /* dbenv->set_rep_transport(???) */
+ /* dbenv->set_rep_limit(???) */
+
     if (dbi->dbi_no_fsync) {
 #if (DB_VERSION_MAJOR == 3 && DB_VERSION_MINOR != 0) || (DB_VERSION_MAJOR == 4)
 	xx = db_env_set_func_fsync(db3_fsync_disable);
@@ -298,38 +413,10 @@ static int db_init(dbiIndex dbi, const char * dbhome,
 	xx = cvtdberr(dbi, "db_env_set_func_fsync", xx, _debug);
     }
 
-/* XXX 3.3.4 change. */
-#if (DB_VERSION_MAJOR == 3 && DB_VERSION_MINOR == 3) || (DB_VERSION_MAJOR == 4)
-    if ((dbi->dbi_ecflags & DB_CLIENT) && dbi->dbi_host) {
-	xx = dbenv->set_rpc_server(dbenv, NULL, dbi->dbi_host,
-		dbi->dbi_cl_timeout, dbi->dbi_sv_timeout, 0);
-	xx = cvtdberr(dbi, "dbenv->set_server", xx, _debug);
-    }
-#else
-    if ((dbi->dbi_ecflags & DB_CLIENT) && dbi->dbi_host) {
-	xx = dbenv->set_server(dbenv, dbi->dbi_host,
-		dbi->dbi_cl_timeout, dbi->dbi_sv_timeout, 0);
-	xx = cvtdberr(dbi, "dbenv->set_server", xx, _debug);
-    }
-#endif
     if (dbi->dbi_shmkey) {
 	xx = dbenv->set_shm_key(dbenv, dbi->dbi_shmkey);
 	xx = cvtdberr(dbi, "dbenv->set_shm_key", xx, _debug);
     }
-    if (dbi->dbi_tmpdir) {
-	const char * root;
-	const char * tmpdir;
-
-	root = (dbi->dbi_root ? dbi->dbi_root : rpmdb->db_root);
-	if ((root[0] == '/' && root[1] == '\0') || rpmdb->db_chrootDone)
-	    root = NULL;
-	/*@-mods@*/
-	tmpdir = rpmGenPath(root, dbi->dbi_tmpdir, NULL);
-	/*@=mods@*/
-	xx = dbenv->set_tmp_dir(dbenv, tmpdir);
-	xx = cvtdberr(dbi, "dbenv->set_tmp_dir", rc, _debug);
-	tmpdir = _free(tmpdir);
-    }
   }
 
 #if (DB_VERSION_MAJOR == 3 && DB_VERSION_MINOR != 0) || (DB_VERSION_MAJOR == 4)
@@ -366,7 +453,11 @@ static int db3sync(dbiIndex dbi, unsigned int flags)
     if (db != NULL)
 	rc = db->sync(db, flags);
     /* XXX DB_INCOMPLETE is returned occaisionally with multiple access. */
+#if (DB_VERSION_MAJOR == 4 && DB_VERSION_MINOR == 1)
+    _printit = _debug;
+#else
     _printit = (rc == DB_INCOMPLETE ? 0 : _debug);
+#endif
     rc = cvtdberr(dbi, "db->sync", rc, _printit);
     return rc;
 }
@@ -599,10 +690,17 @@ static int db3associate(dbiIndex dbi, dbiIndex dbisecondary,
 {
     DB * db = dbi->dbi_db;
     DB * secondary = dbisecondary->dbi_db;
+#if (DB_VERSION_MAJOR == 4 && DB_VERSION_MINOR == 1)
+    DB_TXN * txnid = NULL;
+#endif
     int rc;
 
 /*@-moduncon@*/ /* FIX: annotate db3 methods */
+#if (DB_VERSION_MAJOR == 4 && DB_VERSION_MINOR == 1)
+    rc = db->associate(db, txnid, secondary, callback, flags);
+#else
     rc = db->associate(db, secondary, callback, flags);
+#endif
 /*@=moduncon@*/
     rc = cvtdberr(dbi, "db->associate", rc, _debug);
     return rc;
@@ -637,6 +735,7 @@ static int db3close(/*@only@*/ dbiIndex dbi, /*@unused@*/ unsigned int flags)
     const char * dbfile;
     const char * dbsubfile;
     DB * db = dbi->dbi_db;
+    int _printit;
     int rc = 0, xx;
 
     flags = 0;	/* XXX unused */
@@ -672,7 +771,9 @@ static int db3close(/*@only@*/ dbiIndex dbi, /*@unused@*/ unsigned int flags)
 
     if (db) {
 	rc = db->close(db, 0);
-	rc = cvtdberr(dbi, "db->close", rc, _debug);
+	/* XXX ignore not found error messages. */
+	_printit = (rc == ENOENT ? 0 : _debug);
+	rc = cvtdberr(dbi, "db->close", rc, _printit);
 	db = dbi->dbi_db = NULL;
 
 	rpmMessage(RPMMESS_DEBUG, _("closed   db index       %s/%s\n"),
@@ -747,7 +848,9 @@ static int db3close(/*@only@*/ dbiIndex dbi, /*@unused@*/ unsigned int flags)
 			(dbfile ? dbfile : tagName(dbi->dbi_rpmtag)));
 
 		xx = db->close(db, 0);
-		xx = cvtdberr(dbi, "db->close", xx, _debug);
+		/* XXX ignore not found error messages. */
+		_printit = (xx == ENOENT ? 0 : _debug);
+		xx = cvtdberr(dbi, "db->close", xx, _printit);
 		db = NULL;
 		if (rc == 0 && xx) rc = xx;
 
@@ -789,6 +892,9 @@ static int db3open(rpmdb rpmdb, rpmTag rpmtag, dbiIndex * dbip)
 
     DB * db = NULL;
     DB_ENV * dbenv = NULL;
+#if (DB_VERSION_MAJOR == 4 && DB_VERSION_MINOR == 1)
+    DB_TXN * txnid = NULL;
+#endif
     u_int32_t oflags;
     int _printit;
 
@@ -870,6 +976,18 @@ static int db3open(rpmdb rpmdb, rpmTag rpmtag, dbiIndex * dbip)
      * Avoid incompatible DB_CREATE/DB_RDONLY flags on DBENV->open.
      */
     if (dbi->dbi_use_dbenv) {
+
+#if HAVE_LIBPTHREAD
+	if (rpmdb->db_dbenv == NULL) {
+	    /* Set DB_PRIVATE if posix mutexes are not shared. */
+	    xx = db3_pthread_nptl();
+	    if (xx) {
+		dbi->dbi_eflags |= DB_PRIVATE;
+		rpmMessage(RPMMESS_DEBUG, _("unshared posix mutexes found(%d), adding DB_PRIVATE, using fcntl lock\n"), xx);
+	    }
+	}
+#endif
+
 	if (access(dbhome, W_OK) == -1) {
 
 	    /* dbhome is unwritable, don't attempt DB_CREATE on DB->open ... */
@@ -981,18 +1099,7 @@ static int db3open(rpmdb rpmdb, rpmTag rpmtag, dbiIndex * dbip)
 	/*@=moduncon@*/
 	rc = cvtdberr(dbi, "db_create", rc, _debug);
 	if (rc == 0 && db != NULL) {
-	    if (rc == 0 && dbi->dbi_lorder) {
-		rc = db->set_lorder(db, dbi->dbi_lorder);
-		rc = cvtdberr(dbi, "db->set_lorder", rc, _debug);
-	    }
-	    if (rc == 0 && dbi->dbi_cachesize) {
-		rc = db->set_cachesize(db, 0, dbi->dbi_cachesize, 0);
-		rc = cvtdberr(dbi, "db->set_cachesize", rc, _debug);
-	    }
-	    if (rc == 0 && dbi->dbi_pagesize) {
-		rc = db->set_pagesize(db, dbi->dbi_pagesize);
-		rc = cvtdberr(dbi, "db->set_pagesize", rc, _debug);
-	    }
+
 /* XXX 3.3.4 change. */
 #if (DB_VERSION_MAJOR == 3 && DB_VERSION_MINOR == 3) || (DB_VERSION_MAJOR == 4)
 	    if (rc == 0 &&
@@ -1008,6 +1115,27 @@ static int db3open(rpmdb rpmdb, rpmTag rpmtag, dbiIndex * dbip)
 		rc = cvtdberr(dbi, "db->set_malloc", rc, _debug);
 	    }
 #endif
+
+/* 4.1: db->set_cache_priority(???) */
+	    if (rc == 0 && !dbi->dbi_use_dbenv && dbi->dbi_cachesize) {
+		rc = db->set_cachesize(db, 0, dbi->dbi_cachesize, 0);
+		rc = cvtdberr(dbi, "db->set_cachesize", rc, _debug);
+	    }
+/* 4.1: db->set_encrypt(???) */
+/* 4.1: db->set_errcall(dbenv, rpmdb->db_errcall); */
+/* 4.1: db->set_errfile(dbenv, rpmdb->db_errfile); */
+/* 4.1: db->set_errpfx(dbenv, rpmdb->db_errpfx); */
+ /* 4.1: db->set_feedback(???) */
+
+	    if (rc == 0 && dbi->dbi_lorder) {
+		rc = db->set_lorder(db, dbi->dbi_lorder);
+		rc = cvtdberr(dbi, "db->set_lorder", rc, _debug);
+	    }
+	    if (rc == 0 && dbi->dbi_pagesize) {
+		rc = db->set_pagesize(db, dbi->dbi_pagesize);
+		rc = cvtdberr(dbi, "db->set_pagesize", rc, _debug);
+	    }
+ /* 4.1: db->set_paniccall(???) */
 	    if (rc == 0 && oflags & DB_CREATE) {
 		switch(dbi->dbi_type) {
 		default:
@@ -1042,6 +1170,7 @@ static int db3open(rpmdb rpmdb, rpmTag rpmtag, dbiIndex * dbip)
 #endif
 		    break;
 		case DB_BTREE:
+/* 4.1: db->set_append_recno(???) */
 		    if (dbi->dbi_bt_flags) {
 			rc = db->set_flags(db, dbi->dbi_bt_flags);
 			rc = cvtdberr(dbi, "db->set_bt_flags", rc, _debug);
@@ -1073,6 +1202,7 @@ static int db3open(rpmdb rpmdb, rpmTag rpmtag, dbiIndex * dbip)
 		    break;
 		case DB_RECNO:
 		    if (dbi->dbi_re_delim) {
+/* 4.1: db->set_append_recno(???) */
 			rc = db->set_re_delim(db, dbi->dbi_re_delim);
 			rc = cvtdberr(dbi, "db->set_re_selim", rc, _debug);
 			if (rc) break;
@@ -1124,8 +1254,13 @@ static int db3open(rpmdb rpmdb, rpmTag rpmtag, dbiIndex * dbip)
 			? dbfullpath : dbfile;
 #endif
 
+#if (DB_VERSION_MAJOR == 4 && DB_VERSION_MINOR == 1)
+		rc = db->open(db, txnid, dbpath, dbsubfile,
+		    dbi->dbi_type, oflags, dbi->dbi_perms);
+#else
 		rc = db->open(db, dbpath, dbsubfile,
 		    dbi->dbi_type, oflags, dbi->dbi_perms);
+#endif
 
 		if (rc == 0 && dbi->dbi_type == DB_UNKNOWN) {
 #if (DB_VERSION_MAJOR == 3 && DB_VERSION_MINOR == 3 && DB_VERSION_PATCH == 11) \
@@ -1148,7 +1283,7 @@ static int db3open(rpmdb rpmdb, rpmTag rpmtag, dbiIndex * dbip)
 
 	    /*
 	     * Lock a file using fcntl(2). Traditionally this is Packages,
-	     * the file used * to store metadata of installed header(s),
+	     * the file used to store metadata of installed header(s),
 	     * as Packages is always opened, and should be opened first,
 	     * for any rpmdb access.
 	     *
@@ -1160,8 +1295,12 @@ static int db3open(rpmdb rpmdb, rpmTag rpmtag, dbiIndex * dbip)
 	     * the DBENV should provide it's own locking scheme. So try to
 	     * acquire a lock, but permit failures, as some other
 	     * DBENV player may already have acquired the lock.
+	     *
+	     * With NPTL posix mutexes, revert to fcntl lock on non-functioning
+	     * glibc/kernel combinations.
 	     */
 	    if (rc == 0 && dbi->dbi_lockdbfd &&
+		!((dbi->dbi_ecflags & DB_CLIENT) && dbi->dbi_host) &&
 		(!dbi->dbi_use_dbenv || _lockdbfd++ == 0))
 	    {
 		int fdno = -1;
@@ -1180,9 +1319,10 @@ static int db3open(rpmdb rpmdb, rpmTag rpmtag, dbiIndex * dbip)
 
 		    rc = fcntl(fdno, F_SETLK, (void *) &l);
 		    if (rc) {
-			/* Warning only if using CDB locking. */
+			/* Warning iff using non-private CDB locking. */
 			rc = ((dbi->dbi_use_dbenv &&
-				(dbi->dbi_eflags & DB_INIT_CDB))
+				(dbi->dbi_eflags & DB_INIT_CDB) &&
+				!(dbi->dbi_eflags & DB_PRIVATE))
 			    ? 0 : 1);
 			rpmError( (rc ? RPMERR_FLOCK : RPMWARN_FLOCK),
 				_("cannot get %s lock on %s/%s\n"),
