@@ -2,7 +2,7 @@
 #define H_HEADER
 
 /** \ingroup header
- * \file lib/header.h
+ * \file rpmdb/header.h
  *
  * An rpm header carries all information about a package. A header is
  * a collection of data elements called tags. Each tag has a data type,
@@ -50,6 +50,8 @@
  *	  to permit header data to be manipulated opaquely through vectors.
  *	- (rpm-4.0.3) Sanity checks on header data to limit #tags to 65K,
  *	  #bytes to 16Mb, and total metadata size to 32Mb added.
+ *	- with addition of tracking dependencies, the package version has been
+ *	  reverted back to 3.
  * .
  *
  * \par Development Issues
@@ -78,23 +80,12 @@
 /* RPM - Copyright (C) 1995-2001 Red Hat Software */
 
 #include <stdio.h>
-#include <rpmio.h>
+#include "rpmio.h"
 
 #ifdef __cplusplus
 extern "C" {
 #endif
 
-#if defined(__alpha__) || defined(__alpha)
-typedef long int int_64;
-typedef int int_32;
-typedef short int int_16;
-typedef char int_8;
-
-typedef unsigned int uint_32;
-typedef unsigned short uint_16;
-
-#else
-
 #if 0	/* XXX hpux needs -Ae in CFLAGS to grok this */
 typedef long long int int_64;
 #endif
@@ -102,9 +93,12 @@ typedef int int_32;
 typedef short int int_16;
 typedef char int_8;
 
+#if 0	/* XXX hpux needs -Ae in CFLAGS to grok this */
+typedef unsigned long long int uint_64;
+#endif
 typedef unsigned int uint_32;
 typedef unsigned short uint_16;
-#endif
+typedef unsigned char uint_8;
 
 /*@-redef@*/	/* LCL: no clue */
 /** \ingroup header
@@ -120,19 +114,20 @@ typedef int_32 *	hCNT_t;
 
 /** \ingroup header
  */
-typedef /*@abstract@*/ /*@refcounted@*/ struct headerToken * Header;
+typedef /*@abstract@*/ /*@refcounted@*/ struct headerToken_s * Header;
 
 /** \ingroup header
  */
-typedef /*@abstract@*/ struct headerIteratorS * HeaderIterator;
+typedef /*@abstract@*/ struct headerIterator_s * HeaderIterator;
 
 /** \ingroup header
  * Associate tag names with numeric values.
  */
 typedef /*@abstract@*/ struct headerTagTableEntry_s * headerTagTableEntry;
 struct headerTagTableEntry_s {
-/*@observer@*/ /*@null@*/ const char * name;	/*!< Tag name. */
-    int val;					/*!< Tag numeric value. */
+/*@observer@*/ /*@null@*/
+    const char * name;		/*!< Tag name. */
+    int val;			/*!< Tag numeric value. */
 };
 
 /** \ingroup header
@@ -153,7 +148,7 @@ enum headerSprintfExtensionType {
  * @param data		tag value
  * @param formatPrefix
  * @param padding
- * @param element
+ * @param element	RPM_BIN_TYPE: no. bytes of data
  * @return		formatted string
  */
 typedef /*only@*/ char * (*headerTagFormatFunction)(int_32 type,
@@ -166,10 +161,10 @@ typedef /*only@*/ char * (*headerTagFormatFunction)(int_32 type,
  * This is allowed to fail, which indicates the tag doesn't exist.
  *
  * @param h		header
- * @retval type		address of tag type
- * @retval data		address of tag value pointer
- * @retval count	address of no. of data items
- * @retval freedata	address of data-was-malloc'ed indicator
+ * @retval *type	tag type
+ * @retval *data	tag value
+ * @retval *count	no. of data items
+ * @retval *freedata	data-was-malloc'ed indicator
  * @return		0 on success
  */
 typedef int (*headerTagTagFunction) (Header h,
@@ -283,12 +278,16 @@ typedef union hRET_s {
 /*@-typeuse -fielduse@*/
 typedef struct HE_s {
     int_32 tag;
-/*@null@*/ hTYP_t typ;
+/*@null@*/
+    hTYP_t typ;
     union {
-/*@null@*/ hPTR_t * ptr;
-/*@null@*/ hRET_t * ret;
+/*@null@*/
+	hPTR_t * ptr;
+/*@null@*/
+	hRET_t * ret;
     } u;
-/*@null@*/ hCNT_t cnt;
+/*@null@*/
+    hCNT_t cnt;
 } * HE_t;
 /*@=typeuse =fielduse@*/
 
@@ -717,7 +716,7 @@ void * headerFreeData( /*@only@*/ /*@null@*/ const void * data, rpmTagType type)
 }
 
 #if !defined(__HEADER_PROTOTYPES__)
-#include <hdrinline.h>
+#include "hdrinline.h"
 #endif
 
 #ifdef __cplusplus
