@@ -616,6 +616,50 @@ dbiIndexSet dbiFreeIndexSet(dbiIndexSet set) {
     return set;
 }
 
+typedef struct miRE_s {
+    rpmTag		tag;		/*!< header tag */
+    rpmMireMode		mode;		/*!< pattern match mode */
+/*@only@*/ const char *	pattern;	/*!< pattern string */
+    int			notmatch;	/*!< like "grep -v" */
+/*@only@*/ regex_t *	preg;		/*!< regex compiled pattern buffer */
+    int			cflags;		/*!< regcomp(3) flags */
+    int			eflags;		/*!< regexec(3) flags */
+    int			fnflags;	/*!< fnmatch(3) flags */
+} * miRE;
+
+struct _rpmdbMatchIterator {
+/*@dependent@*/ /*@null@*/
+    rpmdbMatchIterator	mi_next;
+/*@only@*/
+    const void *	mi_keyp;
+    size_t		mi_keylen;
+/*@refcounted@*/
+    rpmdb		mi_db;
+    rpmTag		mi_rpmtag;
+    dbiIndexSet		mi_set;
+    DBC *		mi_dbc;
+    DBT			mi_key;
+    DBT			mi_data;
+    int			mi_setx;
+/*@refcounted@*/ /*@null@*/
+    Header		mi_h;
+    int			mi_sorted;
+    int			mi_cflags;
+    int			mi_modified;
+    unsigned int	mi_prevoffset;
+    unsigned int	mi_offset;
+    unsigned int	mi_filenum;
+    int			mi_nre;
+/*@only@*/ /*@null@*/
+    miRE		mi_re;
+/*@null@*/
+    rpmts		mi_ts;
+/*@null@*/
+    rpmRC (*mi_hdrchk) (rpmts ts, const void * uh, size_t uc, const char ** msg)
+	/*@modifies ts, *msg @*/;
+
+};
+
 /**
  */
 /*@unchecked@*/
@@ -631,14 +675,15 @@ static void handler(int signum);
 static struct sigtbl_s {
     int signum;
     int active;
-    struct sigaction act;
+    void (*handler) (int signum);
     struct sigaction oact;
 } satbl[] = {
-    { SIGHUP,	0, { {handler} } },
-    { SIGINT,	0, { {handler} } },
-    { SIGTERM,	0, { {handler} } },
-    { SIGQUIT,	0, { {handler} } },
-    { -1,	0, { {NULL}    } },
+    { SIGHUP,	0, handler },
+    { SIGINT,	0, handler },
+    { SIGTERM,	0, handler },
+    { SIGQUIT,	0, handler },
+    { SIGPIPE,	0, handler },
+    { -1,	0, NULL },
 };
 /*@=fullinitblock@*/
 
@@ -669,8 +714,9 @@ static int enableSignals(void)
 	/*@globals caught, satbl, fileSystem @*/
 	/*@modifies caught, satbl, fileSystem @*/
 {
-    struct sigtbl_s * tbl;
     sigset_t newMask, oldMask;
+    struct sigtbl_s * tbl;
+    struct sigaction act;
     int rc;
 
     (void) sigfillset(&newMask);		/* block all signals */
@@ -680,7 +726,9 @@ static int enableSignals(void)
 	if (tbl->active++ > 0)
 	    continue;
 	(void) sigdelset(&caught, tbl->signum);
-	rc = sigaction(tbl->signum, &tbl->act, &tbl->oact);
+	memset(&act, 0, sizeof(act));
+	act.sa_handler = tbl->handler;
+	rc = sigaction(tbl->signum, &act, &tbl->oact);
 	if (rc) break;
     }
     return sigprocmask(SIG_SETMASK, &oldMask, NULL);
@@ -689,13 +737,18 @@ static int enableSignals(void)
 /*@unchecked@*/
 static rpmdb rpmdbRock;
 
+/*@unchecked@*/ /*@exposed@*/ /*@null@*/
+static rpmdbMatchIterator rpmmiRock;
+
 int rpmdbCheckSignals(void)
-	/*@globals rpmdbRock, satbl @*/
-	/*@modifies rpmdbRock @*/
+	/*@globals rpmdbRock, rpmmiRock, satbl @*/
+	/*@modifies rpmdbRock, rpmmiRock @*/
 {
     struct sigtbl_s * tbl;
     sigset_t newMask, oldMask;
-    int terminate = 0;
+    static int terminate = 0;
+
+    if (terminate) return 0;
 
     (void) sigfillset(&newMask);		/* block all signals */
     (void) sigprocmask(SIG_BLOCK, &newMask, &oldMask);
@@ -708,7 +761,18 @@ int rpmdbCheckSignals(void)
 
     if (terminate) {
 	rpmdb db;
-	rpmMessage(RPMMESS_WARNING, "Exiting on signal ...\n");
+	rpmdbMatchIterator mi;
+
+	rpmMessage(RPMMESS_DEBUG, "Exiting on signal ...\n");
+
+/*@-branchstate@*/
+	while ((mi = rpmmiRock) != NULL) {
+/*@i@*/	    rpmmiRock = mi->mi_next;
+	    mi->mi_next = NULL;
+/*@i@*/	    mi = rpmdbFreeIterator(mi);
+	}
+/*@=branchstate@*/
+
 /*@-newreftrans@*/
 	while ((db = rpmdbRock) != NULL) {
 /*@i@*/	    rpmdbRock = db->db_next;
@@ -1278,9 +1342,9 @@ if (rc == 0)
 
 	    prevoff = offset;
 	    i++;
-	    offset = dbiIndexRecordOffset(allMatches, i);
-	} while (i < allMatches->count && 
-		(i == 0 || offset == prevoff));
+	    if (i < allMatches->count)
+		offset = dbiIndexRecordOffset(allMatches, i);
+	} while (i < allMatches->count && offset == prevoff);
 
 	baseNames = hfd(baseNames, bnt);
 	dirNames = hfd(dirNames, dnt);
@@ -1553,48 +1617,6 @@ static rpmRC dbiFindByLabel(dbiIndex dbi, DBC * dbcursor, DBT * key, DBT * data,
     /*@=nullstate@*/
 }
 
-typedef struct miRE_s {
-    rpmTag		tag;		/*!< header tag */
-    rpmMireMode		mode;		/*!< pattern match mode */
-/*@only@*/ const char *	pattern;	/*!< pattern string */
-    int			notmatch;	/*!< like "grep -v" */
-/*@only@*/ regex_t *	preg;		/*!< regex compiled pattern buffer */
-    int			cflags;		/*!< regcomp(3) flags */
-    int			eflags;		/*!< regexec(3) flags */
-    int			fnflags;	/*!< fnmatch(3) flags */
-} * miRE;
-
-struct _rpmdbMatchIterator {
-/*@only@*/
-    const void *	mi_keyp;
-    size_t		mi_keylen;
-/*@refcounted@*/
-    rpmdb		mi_db;
-    rpmTag		mi_rpmtag;
-    dbiIndexSet		mi_set;
-    DBC *		mi_dbc;
-    DBT			mi_key;
-    DBT			mi_data;
-    int			mi_setx;
-/*@refcounted@*/ /*@null@*/
-    Header		mi_h;
-    int			mi_sorted;
-    int			mi_cflags;
-    int			mi_modified;
-    unsigned int	mi_prevoffset;
-    unsigned int	mi_offset;
-    unsigned int	mi_filenum;
-    int			mi_nre;
-/*@only@*/ /*@null@*/
-    miRE		mi_re;
-/*@null@*/
-    rpmts		mi_ts;
-/*@null@*/
-    rpmRC (*mi_hdrchk) (rpmts ts, const void * uh, size_t uc, const char ** msg)
-	/*@modifies ts, *msg @*/;
-
-};
-
 /**
  * Rewrite a header into packages (if necessary) and free the header.
  *   Note: this is called from a markReplacedFiles iteration, and *must*
@@ -1660,7 +1682,10 @@ static int miFreeHeader(rpmdbMatchIterator mi, dbiIndex dbi)
 }
 
 rpmdbMatchIterator rpmdbFreeIterator(rpmdbMatchIterator mi)
+	/*@globals rpmmiRock @*/
+	/*@modifies rpmmiRock @*/
 {
+    rpmdbMatchIterator * prev, next;
     dbiIndex dbi;
     int xx;
     int i;
@@ -1668,6 +1693,14 @@ rpmdbMatchIterator rpmdbFreeIterator(rpmdbMatchIterator mi)
     if (mi == NULL)
 	return NULL;
 
+    prev = &rpmmiRock;
+    while ((next = *prev) != NULL && next != mi)
+	prev = &next->mi_next;
+    if (next) {
+/*@i@*/	*prev = next->mi_next;
+	next->mi_next = NULL;
+    }
+
     dbi = dbiOpen(mi->mi_db, RPMDBI_PACKAGES, 0);
     if (dbi == NULL)	/* XXX can't happen */
 	return NULL;
@@ -1694,6 +1727,7 @@ rpmdbMatchIterator rpmdbFreeIterator(rpmdbMatchIterator mi)
     mi->mi_set = dbiFreeIndexSet(mi->mi_set);
     mi->mi_keyp = _free(mi->mi_keyp);
     mi->mi_db = rpmdbUnlink(mi->mi_db, "matchIterator");
+
     mi = _free(mi);
 
     (void) rpmdbCheckSignals();
@@ -2422,7 +2456,9 @@ int rpmdbAppendIterator(rpmdbMatchIterator mi, const int * hdrNums, int nHdrNums
 }
 
 rpmdbMatchIterator rpmdbInitIterator(rpmdb db, rpmTag rpmtag,
-	const void * keyp, size_t keylen)
+		const void * keyp, size_t keylen)
+	/*@globals rpmmiRock @*/
+	/*@modifies rpmmiRock @*/
 {
     rpmdbMatchIterator mi;
     DBT * key;
@@ -2448,6 +2484,9 @@ rpmdbMatchIterator rpmdbInitIterator(rpmdb db, rpmTag rpmtag,
 	return NULL;
 
     mi = xcalloc(1, sizeof(*mi));
+    mi->mi_next = rpmmiRock;
+    rpmmiRock = mi;
+
     key = &mi->mi_key;
     data = &mi->mi_data;
 
@@ -2492,6 +2531,8 @@ if (rc == 0)
 	}
 	if (rc)	{	/* error/not found */
 	    set = dbiFreeIndexSet(set);
+	    rpmmiRock = mi->mi_next;
+	    mi->mi_next = NULL;
 	    mi = _free(mi);
 	    return NULL;
 	}
@@ -2533,9 +2574,7 @@ if (rc == 0)
     mi->mi_ts = NULL;
     mi->mi_hdrchk = NULL;
 
-    /*@-nullret@*/ /* FIX: mi->mi_key.data may be NULL */
-    return mi;
-    /*@=nullret@*/
+/*@i@*/ return mi;
 }
 
 /* XXX psm.c */
@@ -3275,7 +3314,9 @@ static struct skipDir_s {
     _skip("/usr/share/zoneinfo"),
     _skip("/usr/share/locale"),
     _skip("/usr/share/i18n"),
+    _skip("/usr/share/doc"),
     _skip("/usr/lib/locale"),
+    _skip("/usr/src/debug"),
     { 0, NULL }
 };
 
@@ -3808,10 +3849,6 @@ int rpmdbRebuild(const char * prefix, rpmts ts,
 
     }
 
-    if (!nocleanup) {
-	olddb->db_remove_env = 1;
-	newdb->db_remove_env = 1;
-    }
     xx = rpmdbClose(olddb);
     xx = rpmdbClose(newdb);
 
