@@ -16,17 +16,6 @@
 #if HAVE_NETINET_IN_SYSTM_H
 # include <sys/types.h>
 
-#if defined(__LCLINT__)
-/*@-redef@*/ /* FIX: rpmdb/db3.c also declares */
-typedef unsigned int u_int32_t;
-typedef unsigned short u_int16_t;
-typedef unsigned char u_int8_t;
-/*@-incondefs@*/        /* LCLint 3.0.0.15 */
-typedef int int32_t;
-/*@=incondefs@*/
-/*@=redef@*/
-#endif
-
 # include <netinet/in_systm.h>
 #endif
 
@@ -34,7 +23,8 @@ typedef int int32_t;
 #define	_USE_LIBIO	1
 #endif
 
-#if !defined(HAVE_HERRNO) && defined(__hpux) /* XXX HP-UX w/o -D_XOPEN_SOURCE needs */
+/* XXX HP-UX w/o -D_XOPEN_SOURCE needs */
+#if !defined(HAVE_HERRNO) && (defined(__hpux) || defined(__LCLINT__))
 /*@unchecked@*/
 extern int h_errno;
 #endif
@@ -67,15 +57,20 @@ static int inet_aton(const char *cp, struct in_addr *inp)
 #include <rpmio_internal.h>
 #undef	fdFileno
 #undef	fdOpen
+#define	fdOpen	__fdOpen
 #undef	fdRead
+#define	fdRead	__fdRead
 #undef	fdWrite
+#define	fdWrite	__fdWrite
 #undef	fdClose
+#define	fdClose	__fdClose
 
 #include "ugid.h"
 #include "rpmmessages.h"
 
 #include "debug.h"
 
+/*@access FILE @*/	/* XXX to permit comparison/conversion with void *. */
 /*@access urlinfo @*/
 /*@access FDSTAT_t @*/
 
@@ -138,7 +133,6 @@ _free(/*@only@*/ /*@null@*/ /*@out@*/ const void * p)
 /* =============================================================== */
 
 /*@-boundswrite@*/
-/*@-modfilesys@*/
 static /*@observer@*/ const char * fdbg(/*@null@*/ FD_t fd)
 	/*@*/
 {
@@ -176,8 +170,6 @@ static /*@observer@*/ const char * fdbg(/*@null@*/ FD_t fd)
 	    sprintf(be, "FD %d fp %p", fps->fdno, fps->fp);
 	} else if (fps->io == ufdio) {
 	    sprintf(be, "UFD %d fp %p", fps->fdno, fps->fp);
-	} else if (fps->io == fadio) {
-	    sprintf(be, "FAD %d fp %p", fps->fdno, fps->fp);
 	} else if (fps->io == gzdio) {
 	    sprintf(be, "GZD %p fdno %d", fps->fp, fps->fdno);
 #if HAVE_BZLIB_H
@@ -199,7 +191,6 @@ static /*@observer@*/ const char * fdbg(/*@null@*/ FD_t fd)
     }
     return buf;
 }
-/*@=modfilesys@*/
 /*@=boundswrite@*/
 
 /* =============================================================== */
@@ -237,9 +228,7 @@ FD_t fdDup(int fdno)
 	return NULL;
     fd = fdNew("open (fdDup)");
     fdSetFdno(fd, nfdno);
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==> fdDup(%d) fd %p %s\n", fdno, (fd ? fd : NULL), fdbg(fd)));
-/*@=modfilesys@*/
     /*@-refcounttrans@*/ return fd; /*@=refcounttrans@*/
 }
 
@@ -263,16 +252,13 @@ FILE *fdFdopen(void * cookie, const char *fmode)
     fdno = fdFileno(fd);
     if (fdno < 0) return NULL;
     fp = fdopen(fdno, fmode);
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==> fdFdopen(%p,\"%s\") fdno %d -> fp %p fdno %d\n", cookie, fmode, fdno, fp, fileno(fp)));
-/*@=modfilesys@*/
     fd = fdFree(fd, "open (fdFdopen)");
     return fp;
 }
 #endif
 
 /* =============================================================== */
-/*@-modfilesys@*/
 /*@-mustmod@*/ /* FIX: cookie is modified */
 static inline /*@null@*/ FD_t XfdLink(void * cookie, const char * msg,
 		const char * file, unsigned line)
@@ -291,9 +277,7 @@ DBGREFS(fd, (stderr, "--> fd  %p ++ %d %s at %s:%u %s\n", fd, fd->nrefs, msg, fi
     return fd;
 }
 /*@=mustmod@*/
-/*@=modfilesys@*/
 
-/*@-modfilesys@*/
 static inline /*@null@*/ FD_t XfdFree( /*@killref@*/ FD_t fd, const char *msg,
 		const char *file, unsigned line)
 	/*@modifies fd @*/
@@ -320,7 +304,6 @@ DBGREFS(fd, (stderr, "--> fd  %p -- %d %s at %s:%u %s\n", fd, fd->nrefs, msg, fi
     }
     return NULL;
 }
-/*@=modfilesys@*/
 
 static inline /*@null@*/ FD_t XfdNew(const char * msg,
 		const char * file, unsigned line)
@@ -337,9 +320,7 @@ static inline /*@null@*/ FD_t XfdNew(const char * msg,
     fd->nfps = 0;
     memset(fd->fps, 0, sizeof(fd->fps));
 
-    /*@-assignexpose@*/
     fd->fps[0].io = fdio;
-    /*@=assignexpose@*/
     fd->fps[0].fp = NULL;
     fd->fps[0].fdno = -1;
 
@@ -365,9 +346,11 @@ static inline /*@null@*/ FD_t XfdNew(const char * msg,
     return XfdLink(fd, msg, file, line);
 }
 
-/*@-redef@*/	/* FIX: legacy API should be made static */
-ssize_t fdRead(void * cookie, /*@out@*/ char * buf, size_t count)
-/*@=redef@*/
+static ssize_t fdRead(void * cookie, /*@out@*/ char * buf, size_t count)
+	/*@globals errno, fileSystem, internalState @*/
+	/*@modifies *buf, errno, fileSystem, internalState @*/
+	/*@requires maxSet(buf) >= (count - 1) @*/
+	/*@ensures maxRead(buf) == result @*/
 {
     FD_t fd = c2f(cookie);
     ssize_t rc;
@@ -382,16 +365,14 @@ ssize_t fdRead(void * cookie, /*@out@*/ char * buf, size_t count)
 
     if (fd->ndigests && rc > 0) fdUpdateDigests(fd, buf, rc);
 
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==>\tfdRead(%p,%p,%ld) rc %ld %s\n", cookie, buf, (long)count, (long)rc, fdbg(fd)));
-/*@=modfilesys@*/
 
     return rc;
 }
 
-/*@-redef@*/	/* FIX: legacy API should be made static */
-ssize_t fdWrite(void * cookie, const char * buf, size_t count)
-/*@=redef@*/
+static ssize_t fdWrite(void * cookie, const char * buf, size_t count)
+	/*@globals errno, fileSystem, internalState @*/
+	/*@modifies errno, fileSystem, internalState @*/
 {
     FD_t fd = c2f(cookie);
     int fdno = fdFileno(fd);
@@ -423,9 +404,7 @@ ssize_t fdWrite(void * cookie, const char * buf, size_t count)
 	if (ec == -1)	fd->syserrno = errno;
     }
 
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==>\tfdWrite(%p,%p,%ld) rc %ld %s\n", cookie, buf, (long)count, (long)rc, fdbg(fd)));
-/*@=modfilesys@*/
 
     return rc;
 }
@@ -447,16 +426,14 @@ static inline int fdSeek(void * cookie, _libio_pos_t pos, int whence)
     rc = lseek(fdFileno(fd), p, whence);
     fdstat_exit(fd, FDSTAT_SEEK, rc);
 
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==>\tfdSeek(%p,%ld,%d) rc %lx %s\n", cookie, (long)p, whence, (unsigned long)rc, fdbg(fd)));
-/*@=modfilesys@*/
 
     return rc;
 }
 
-/*@-redef@*/	/* FIX: legacy API should be made static */
-int fdClose( /*@only@*/ void * cookie)
-/*@=redef@*/
+static int fdClose( /*@only@*/ void * cookie)
+	/*@globals errno, fileSystem, systemState, internalState @*/
+	/*@modifies errno, fileSystem, systemState, internalState @*/
 {
     FD_t fd;
     int fdno;
@@ -472,17 +449,15 @@ int fdClose( /*@only@*/ void * cookie)
     rc = ((fdno >= 0) ? close(fdno) : -2);
     fdstat_exit(fd, FDSTAT_CLOSE, rc);
 
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==>\tfdClose(%p) rc %lx %s\n", (fd ? fd : NULL), (unsigned long)rc, fdbg(fd)));
-/*@=modfilesys@*/
 
     fd = fdFree(fd, "open (fdClose)");
     return rc;
 }
 
-/*@-redef@*/	/* FIX: legacy API should be made static */
-/*@null@*/ FD_t fdOpen(const char *path, int flags, mode_t mode)
-/*@=redef@*/
+static /*@null@*/ FD_t fdOpen(const char *path, int flags, mode_t mode)
+	/*@globals errno, fileSystem, internalState @*/
+	/*@modifies errno, fileSystem, internalState @*/
 {
     FD_t fd;
     int fdno;
@@ -496,9 +471,7 @@ DBGIO(fd, (stderr, "==>\tfdClose(%p) rc %lx %s\n", (fd ? fd : NULL), (unsigned l
     fd = fdNew("open (fdOpen)");
     fdSetFdno(fd, fdno);
     fd->flags = flags;
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==>\tfdOpen(\"%s\",%x,0%o) %s\n", path, (unsigned)flags, (unsigned)mode, fdbg(fd)));
-/*@=modfilesys@*/
     /*@-refcounttrans@*/ return fd; /*@=refcounttrans@*/
 }
 
@@ -753,6 +726,14 @@ static int getHostAddress(const char * host, /*@out@*/ struct in_addr * address)
 	/*@globals errno @*/
 	/*@modifies *address, errno @*/
 {
+#if 0	/* XXX workaround nss_foo module hand-off using valgrind. */
+    if (!strcmp(host, "localhost")) {
+	/*@-unrecog -moduncon @*/
+	if (!inet_aton("127.0.0.1", address))
+	    return FTPERR_BAD_HOST_ADDR;
+	/*@=unrecog =moduncon @*/
+    } else
+#endif
     if (xisdigit(host[0])) {
 	/*@-unrecog -moduncon @*/
 	if (!inet_aton(host, address))
@@ -1348,10 +1329,8 @@ int ufdCopy(FD_t sfd, FD_t tfd)
 	}
     }
 
-/*@-modfilesys@*/
     DBGIO(sfd, (stderr, "++ copied %d bytes: %s\n", itemsCopied,
 	ftpStrerror(rc)));
-/*@=modfilesys@*/
 
     if (urlNotify) {
 /*@-boundsread@*/
@@ -1473,9 +1452,7 @@ static int ftpAbort(urlinfo u, FD_t data)
     }
     ctrl = u->ctrl;
 
-/*@-modfilesys@*/
     DBGIO(0, (stderr, "-> ABOR\n"));
-/*@=modfilesys@*/
 
 /*@-usereleased -compdef@*/
     if (send(fdFileno(ctrl), ipbuf, sizeof(ipbuf), MSG_OOB) != sizeof(ipbuf)) {
@@ -2135,9 +2112,7 @@ fprintf(stderr, "*** ufdOpen(%s,0x%x,0%o)\n", url, (unsigned)flags, (unsigned)mo
 	(void) ufdClose(fd);
 	return NULL;
     }
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==>\tufdOpen(\"%s\",%x,0%o) %s\n", url, (unsigned)flags, (unsigned)mode, fdbg(fd)));
-/*@=modfilesys@*/
     return fd;
 }
 
@@ -2182,15 +2157,13 @@ static /*@null@*/ FD_t gzdOpen(const char * path, const char * fmode)
 	/*@modifies fileSystem @*/
 {
     FD_t fd;
-    gzFile *gzfile;
+    gzFile gzfile;
     if ((gzfile = gzopen(path, fmode)) == NULL)
 	return NULL;
     fd = fdNew("open (gzdOpen)");
     fdPop(fd); fdPush(fd, gzdio, gzfile, -1);
     
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==>\tgzdOpen(\"%s\", \"%s\") fd %p %s\n", path, fmode, (fd ? fd : NULL), fdbg(fd)));
-/*@=modfilesys@*/
     return fdLink(fd, "gzdOpen");
 }
 
@@ -2201,7 +2174,7 @@ static /*@null@*/ FD_t gzdFdopen(void * cookie, const char *fmode)
 {
     FD_t fd = c2f(cookie);
     int fdno;
-    gzFile *gzfile;
+    gzFile gzfile;
 
     if (fmode == NULL) return NULL;
     fdno = fdFileno(fd);
@@ -2221,7 +2194,7 @@ static int gzdFlush(FD_t fd)
 	/*@globals fileSystem @*/
 	/*@modifies fileSystem @*/
 {
-    gzFile *gzfile;
+    gzFile gzfile;
     gzfile = gzdFileno(fd);
     if (gzfile == NULL) return -2;
     return gzflush(gzfile, Z_SYNC_FLUSH);	/* XXX W2DO? */
@@ -2235,7 +2208,7 @@ static ssize_t gzdRead(void * cookie, /*@out@*/ char * buf, size_t count)
 	/*@modifies *buf, fileSystem, internalState @*/
 {
     FD_t fd = c2f(cookie);
-    gzFile *gzfile;
+    gzFile gzfile;
     ssize_t rc;
 
     if (fd == NULL || fd->bytesRemain == 0) return 0;	/* XXX simulate EOF */
@@ -2246,9 +2219,7 @@ static ssize_t gzdRead(void * cookie, /*@out@*/ char * buf, size_t count)
     fdstat_enter(fd, FDSTAT_READ);
     /*@-compdef@*/ /* LCL: *buf is undefined */
     rc = gzread(gzfile, buf, count);
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==>\tgzdRead(%p,%p,%u) rc %lx %s\n", cookie, buf, (unsigned)count, (unsigned long)rc, fdbg(fd)));
-/*@=modfilesys@*/
     /*@=compdef@*/
     if (rc < 0) {
 	int zerror = 0;
@@ -2272,7 +2243,7 @@ static ssize_t gzdWrite(void * cookie, const char * buf, size_t count)
 	/*@modifies fileSystem, internalState @*/
 {
     FD_t fd = c2f(cookie);
-    gzFile *gzfile;
+    gzFile gzfile;
     ssize_t rc;
 
     if (fd == NULL || fd->bytesRemain == 0) return 0;	/* XXX simulate EOF */
@@ -2284,9 +2255,7 @@ static ssize_t gzdWrite(void * cookie, const char * buf, size_t count)
 
     fdstat_enter(fd, FDSTAT_WRITE);
     rc = gzwrite(gzfile, (void *)buf, count);
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==>\tgzdWrite(%p,%p,%u) rc %lx %s\n", cookie, buf, (unsigned)count, (unsigned long)rc, fdbg(fd)));
-/*@=modfilesys@*/
     if (rc < 0) {
 	int zerror = 0;
 	fd->errcookie = gzerror(gzfile, &zerror);
@@ -2313,7 +2282,7 @@ static inline int gzdSeek(void * cookie, _libio_pos_t pos, int whence)
     int rc;
 #if HAVE_GZSEEK
     FD_t fd = c2f(cookie);
-    gzFile *gzfile;
+    gzFile gzfile;
 
     if (fd == NULL) return -2;
     assert(fd->bytesRemain == -1);	/* XXX FIXME */
@@ -2323,9 +2292,7 @@ static inline int gzdSeek(void * cookie, _libio_pos_t pos, int whence)
 
     fdstat_enter(fd, FDSTAT_SEEK);
     rc = gzseek(gzfile, p, whence);
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==>\tgzdSeek(%p,%ld,%d) rc %lx %s\n", cookie, (long)p, whence, (unsigned long)rc, fdbg(fd)));
-/*@=modfilesys@*/
     if (rc < 0) {
 	int zerror = 0;
 	fd->errcookie = gzerror(gzfile, &zerror);
@@ -2347,7 +2314,7 @@ static int gzdClose( /*@only@*/ void * cookie)
 	/*@modifies fileSystem, internalState @*/
 {
     FD_t fd = c2f(cookie);
-    gzFile *gzfile;
+    gzFile gzfile;
     int rc;
 
     gzfile = gzdFileno(fd);
@@ -2361,13 +2328,9 @@ static int gzdClose( /*@only@*/ void * cookie)
     /* XXX TODO: preserve fd if errors */
 
     if (fd) {
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==>\tgzdClose(%p) zerror %d %s\n", cookie, rc, fdbg(fd)));
-/*@=modfilesys@*/
 	if (rc < 0) {
-	    /*@-usereleased@*/
-	    fd->errcookie = gzerror(gzfile, &rc);
-	    /*@=usereleased@*/
+	    fd->errcookie = "gzclose error";
 	    if (rc == Z_ERRNO) {
 		fd->syserrno = errno;
 		fd->errcookie = strerror(fd->syserrno);
@@ -2377,9 +2340,7 @@ DBGIO(fd, (stderr, "==>\tgzdClose(%p) zerror %d %s\n", cookie, rc, fdbg(fd)));
 	}
     }
 
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==>\tgzdClose(%p) rc %lx %s\n", cookie, (unsigned long)rc, fdbg(fd)));
-/*@=modfilesys@*/
 
     if (_rpmio_debug || rpmIsDebug()) fdstat_print(fd, "GZDIO", stderr);
     /*@-branchstate@*/
@@ -2579,9 +2540,7 @@ static int bzdClose( /*@only@*/ void * cookie)
 	}
     }
 
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==>\tbzdClose(%p) rc %lx %s\n", cookie, (unsigned long)rc, fdbg(fd)));
-/*@=modfilesys@*/
 
     if (_rpmio_debug || rpmIsDebug()) fdstat_print(fd, "BZDIO", stderr);
     /*@-branchstate@*/
@@ -2644,9 +2603,7 @@ size_t Fread(void *buf, size_t size, size_t nmemb, FD_t fd) {
     int rc;
 
     FDSANE(fd);
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==> Fread(%p,%u,%u,%p) %s\n", buf, (unsigned)size, (unsigned)nmemb, (fd ? fd : NULL), fdbg(fd)));
-/*@=modfilesys@*/
 
     if (fdGetIo(fd) == fpio) {
 	/*@+voidabstract -nullpass@*/
@@ -2669,9 +2626,7 @@ size_t Fwrite(const void *buf, size_t size, size_t nmemb, FD_t fd)
     int rc;
 
     FDSANE(fd);
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==> Fwrite(%p,%u,%u,%p) %s\n", buf, (unsigned)size, (unsigned)nmemb, (fd ? fd : NULL), fdbg(fd)));
-/*@=modfilesys@*/
 
     if (fdGetIo(fd) == fpio) {
 /*@-boundsread@*/
@@ -2702,9 +2657,7 @@ int Fseek(FD_t fd, _libio_off_t offset, int whence) {
     long int rc;
 
     FDSANE(fd);
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==> Fseek(%p,%ld,%d) %s\n", fd, (long)offset, whence, fdbg(fd)));
-/*@=modfilesys@*/
 
     if (fdGetIo(fd) == fpio) {
 	FILE *fp;
@@ -2729,9 +2682,7 @@ int Fclose(FD_t fd)
     int rc = 0, ec = 0;
 
     FDSANE(fd);
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==> Fclose(%p) %s\n", (fd ? fd : NULL), fdbg(fd)));
-/*@=modfilesys@*/
 
     fd = fdLink(fd, "Fclose");
     /*@-branchstate@*/
@@ -2922,8 +2873,6 @@ fprintf(stderr, "*** Fdopen(%p,%s) %s\n", fd, fmode, fdbg(fd));
 #endif
 	} else if (!strcmp(end, "ufdio")) {
 	    iof = ufdio;
-	} else if (!strcmp(end, "fadio")) {
-	    iof = fadio;
 	} else if (!strcmp(end, "fpio")) {
 	    iof = fpio;
 	    if (noLibio) {
@@ -2967,9 +2916,7 @@ fprintf(stderr, "*** Fdopen fpio fp %p\n", (void *)fp);
 	    ciof.seek = iof->seek;
 	    ciof.close = iof->close;
 	    fp = fopencookie(fd, stdio, ciof);
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==> fopencookie(%p,\"%s\",*%p) returns fp %p\n", fd, stdio, iof, fp));
-/*@=modfilesys@*/
 	}
 #endif
 
@@ -2986,9 +2933,7 @@ DBGIO(fd, (stderr, "==> fopencookie(%p,\"%s\",*%p) returns fp %p\n", fd, stdio,
 	/*@=branchstate@*/
     }
 
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==> Fdopen(%p,\"%s\") returns fd %p %s\n", ofd, fmode, (fd ? fd : NULL), fdbg(fd)));
-/*@=modfilesys@*/
     /*@-refcounttrans -retalias@*/ return fd; /*@=refcounttrans =retalias@*/
 }
 /*@=boundswrite@*/
@@ -3018,14 +2963,6 @@ fprintf(stderr, "*** Fopen fdio path %s fmode %s\n", path, fmode);
 	    if (fd) (void) fdClose(fd);
 	    return NULL;
 	}
-    } else if (!strcmp(end, "fadio")) {
-if (_rpmio_debug)
-fprintf(stderr, "*** Fopen fadio path %s fmode %s\n", path, fmode);
-	fd = fadio->_open(path, flags, perms);
-	if (fdFileno(fd) < 0) {
-	    /*@-refcounttrans@*/ (void) fdClose(fd); /*@=refcounttrans@*/
-	    return NULL;
-	}
     } else {
 	FILE *fp;
 	int fdno;
@@ -3122,9 +3059,7 @@ int Ferror(FD_t fd)
 	if (rc == 0 && ec)
 	    rc = ec;
     }
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==> Ferror(%p) rc %d %s\n", fd, rc, fdbg(fd)));
-/*@=modfilesys@*/
     return rc;
 }
 
@@ -3137,9 +3072,7 @@ int Fileno(FD_t fd)
 	rc = fd->fps[i].fdno;
 /*@=boundsread@*/
     }
-/*@-modfilesys@*/
 DBGIO(fd, (stderr, "==> Fileno(%p) rc %d %s\n", (fd ? fd : NULL), rc, fdbg(fd)));
-/*@=modfilesys@*/
     return rc;
 }
 
